(**
 * The First-Order LoopW Proof Checker compiler.
 *
 * The compiler calls several modules in order to
 * perform the requested tasks.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure Compiler :> COMPILER = struct

(**
 * The type of compiling parameters.
 *)
type compilerParams = 
	{ (** the path of the source file which contains the imperative program ".loop". *)
	  sourceFile : string,
	  (** the source file name prefix. *)
	  fileNamePrefix : string,
	  (** the verbosity level (0: nothing, ..., 3: very verbose). *)
	  verbLevel : int,
	  (** asks the compiler to write the untyped program in a file ".cs". *)
	  printUntyped : bool,
	  (** asks the compiler to write the translated functional program in a file ".sml". *)
	  printFun : bool
	}

(**
 * prints a message according to its importance and the requested verbosity.
 *
 * @params params level mess
 * @param params the compiling parameters containing the requested verbosity level.
 * @param level the message verbosity level.
 * @param mess the message to be printed.
 *)
fun print_verb (params : compilerParams) level mess =
    if (#verbLevel params) >= level then print mess else ()

(**
 * prints a string list to a file.
 *
 * @params name strList
 * @param name the file name.
 * @param strList the string list.
 *)
fun writeFile name strList = let
    val file = TextIO.openOut (name)
    val rec aux = fn
	[] => TextIO.closeOut file
      | s::q => (TextIO.output (file, s ^ "\n"); aux q)
in
    aux strList
end

(**
 * prints the compiling parameters values.
 *
 * @params params
 * @param params the compiling parameters.
 *)
fun printParams (params : compilerParams) =
	if (#verbLevel params) >= 3
	then
		let val () = print ("\n== Compilation options:")
			val () = print ("\n== source file: " ^ (#sourceFile params))
			val () = print ("\n== source file name prefix: " ^ (#fileNamePrefix params))
			val () = print ("\n== verbosity level: " ^ (Int.toString (#verbLevel params)))
			val () = print ("\n== untyped source print: " ^ (if #printUntyped params then "yes" else "no"))
			val () = print ("\n== functional prog source print: " ^ (if #printFun params then "yes" else "no"))
		in () end
	else ()

(**
 * performs type checking and on-demand functional translation.
 *
 * @params params
 * @param params the compiling parameters.
 *)
fun compile (params : compilerParams) = let
	val () = print_verb params 3 ("\n==(PC)== Starting Compilation  =====\n==")
	val () = printParams params

	(* Parsing source code *)
	val prog = Parser.parseFile (#sourceFile params)

	val () = (* if the user asks for the untyped source code *)
	    if (#printUntyped params) then
		let val () = print_verb params 3 "\n--(PC)-- Writing down untyped source code -----\n\n"
			val () = writeFile ((#fileNamePrefix params) ^ ".pcs")
				       (ProofPrettyPrinter.stringOfProgram_Untyped false false prog)
		in () end
	    else ()

	val () = print_verb params 3 "\n--(PC)-- Proof Checking Source Code: "
	val checked = ProofChecker.checkClosedProg prog
	val () = print_verb params 3 ((if checked then "ok" else "error") ^ " -----\n\n")

	val () = print_verb params 3 "\n--(PC)-- Translating Source Code: "
	val smlProg = FunTranslation.translateProgram prog
	val () = print_verb params 3 "ok -----\n\n"

	val () = (* if the user asks for the functional source code *)
	    if (#printFun params) then
		let val () = print_verb params 3 "\n--(PC)-- Writing down functional source code (.sml) -----\n\n"
			val () = writeFile ((#fileNamePrefix params) ^ ".sml")
				       (FunPrettyPrinter.stringOfTerm smlProg)
		in () end
	    else ()
	    
	val () = print_verb params 3 "\n--(PC)-- Type Checking Functional Source Code: "
	val checked = (FunTypeChecker.typeCheck [] smlProg)
	handle FunTypeChecker.TypeError (G, (t, ty)) =>
		raise Fail ("\nFunTypeError:\n\n["
					^ (FunPrettyPrinter.stringOfEnv G) ^ "]\n\n"
					^ (String.concatWith "\n" (FunPrettyPrinter.stringOfTerm (t, ty))))
	val () = print_verb params 3 ((if checked then "ok" else "error") ^ " -----\n\n")

    in print_verb params 3 "\n==(PC)== Compilation succeeded =====\n" end
    
end

