(**
 * The First-Order LoopW Proof Checker functional pretty-printer.
 *
 * The functional pretty-printer converts functional terms into string.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure FunPrettyPrinter :> FUN_PRETTY_PRINTER = struct

open FunAst
open PrettyPrinterKeywords
open BasePrettyPrinter

(**
 * appends a string at the end of the last string of a string list.
 *
 * @params l s
 * @param l the string list [s1, ..., sn].
 * @param s the string to append.
 *
 * @return [s1, ..., sn ^ s] (or [] if l = []).
 *)
fun append_last l s =
	case l of
	  [] => []
	| [(s1, s2)] => [(s1, s2 ^ s)]
	| s'::q => s'::(append_last q s)

(**
 * converts formulas/types abstract syntax into string.
 *
 * @params t
 * @param t the type.
 *
 * @return the string representation of t.
 *)
fun stringOfType t =
	case t of
	  TBot => kwdBot false
	| TNat (n) =>  (kwd false "nat") ^ "(" ^ (stringOfTerm false n) ^ ")"
	| TEqual (n, m) => (stringOfTerm false n) ^ " = " ^ (stringOfTerm false m)
	| TTuple (i, tl) =>
	(if null i then "" else "Exists " ^ (String.concatWith "," (List.map stringOfIdent i)) ^ ".")
	^ "(" ^ (String.concatWith " & " (List.map stringOfType tl)) ^ ")"
	| TFun (i, t1, t2) =>
	(if null i then "" else "Forall " ^ (String.concatWith "," (List.map stringOfIdent i)) ^ ".")
	^ "(" ^ (stringOfType t1) ^ " => " ^ (stringOfType t2) ^ ")"
	| TVar (i) => stringOfIdent i

(**
 * converts typed identifiers into string.
 *
 * @params (id, t)
 * @param id the identifier.
 * @param t the type.
 *
 * @return the string representation of (id, t).
 *)
fun stringOfTypedIdent (id, t) = (stringOfIdent id) ^ " (* : " ^ (stringOfType t) ^ " *)"

(**
 * converts typed identifier lists into string.
 *
 * @params il
 * @param il the typed identifier list.
 *
 * @return the string representation of il.
 *)
fun stringOfTypedIdentList il = String.concatWith ", " (List.map stringOfTypedIdent il)

(**
 * builds a string representation of a typing environment.
 *
 * @params env
 * @param env the environment.
 *
 * @return the string representation of env.
 *)
fun stringOfEnv env = String.concatWith ",\n " (List.map stringOfTypedIdent env)

(* Lemma counter *)
val (init, next) = let val c = ref 0 in (fn () => c := 0, fn () => (c := !c + 1; !c)) end

(**
 * builds a compact string representation of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the string representation of proof obligations and the compact string representation of t.
 *)
fun stringOfTerm_compact (t, typ) : (string list * string) =
	case t of
	  Id (id) => ([], stringOfTypedIdent (id, typ))
	| Tuple (subst, tl) =>
	let val r = List.map stringOfTerm_compact tl
	in
		(List.concat (List.map #1 r),
		(if List.length tl = 1 then "" else "(") ^ (String.concatWith ", " (List.map #2 r))
		^ (if List.length tl = 1 then "" else ")")
		^ (if null subst then ""
			else " (* [" ^ (String.concatWith ", " (List.map (stringOfTerm false) subst)) ^ "] *)"))
	end
	
	| Z => ([], "0")
	
	| S (t) =>
	let val rec aux =
		(fn (Z, _) => 0
		  | (S(n), _) => (aux n) + 1
		  | _ => raise Fail "FunPrettyPrinter.stringOfTerm_compact.aux")
	in
		([], (Int.toString (aux (S t, typ))))
	end
	
	| Succ (t) => let val r = stringOfTerm_compact t in (#1 r, "Succ(" ^ (#2 r) ^ ")") end
	
	| Pred (t) => let val r = stringOfTerm_compact t in (#1 r, "Pred(" ^ (#2 r) ^ ")") end
	
	| App (t1, subst, t2) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2
	in
		((#1 r1) @ (#1 r2),
		(#2 r1) ^ "(" ^ (#2 r2) ^ ")"
		^ (if null subst then ""
			else " (* [" ^ (String.concatWith ", " (List.map (stringOfTerm false) subst)) ^ "] *)"))
	end
	
	| Abs (idl, t) =>
	let val r = stringOfTerm_compact t
	in
	  (#1 r,
	  "fn (" ^ (stringOfTypedIdentList idl)
	   ^ ")" ^ " => " ^ (#2 r))
	end
	
	| Rec (t1, t2, t3) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2
		val r3 = stringOfTerm_compact t3
	in
		((#1 r1) @ (#1 r2) @ (#1 r3),
		"Rec (" ^ (#2 r1) ^ ", " ^ (#2 r2) ^ "," ^ (#2 r3) ^ ")")
	end
	
	| LetIn (idl, t1, t2) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2 in
	  ((#1 r1) @ (#1 r2),
	   "let val " ^ (if List.length idl = 1 then "" else "(")
	   ^ (stringOfTypedIdentList idl)
	   ^ (if List.length idl = 1 then "" else ")") ^ " = " ^ (#2 r1)
	   ^ " in " ^ (#2 r2) ^ " end")
	end
	
	| Callcc (id, t) =>
	let val r = stringOfTerm_compact t
	in
		(#1 r,
		"callcc (fn " ^ (stringOfIdent id) ^ " => "
		^ (#2 r) ^ ")")
	end
	
	| Throw (t1, subst, t2) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2
	in
		((#1 r1) @ (#1 r2),
		"throw " ^ (#2 r1) ^ " " ^ (#2 r2)
		^ (if null subst then ""
			else " (* [" ^ (String.concatWith ", " (List.map (stringOfTerm false) subst)) ^ "] *)"))
	end
	
	| Subst (t1, typi, id, t2) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2
	in
		((#1 r1) @ (#1 r2),
		(#2 r1)
		^ " (* :> {" ^ (stringOfIdent id) ^ "/" ^ (stringOfType typi) ^ "}[" ^ (stringOfType (#2 t2)) ^ (#2 r2) ^ "] *)")
	end
	
	| Lemma (tl, t) =>
	let val n = Int.toString (next ())
	in
		([n ^ ": " ^ (String.concatWith ", " (List.map stringOfType tl))
		^ " |- " ^ (stringOfType t)],
		" by #" ^ n)
	end

(**
 * builds a expanded string representation of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the string representation of proof obligations and the expanded string representation of t.
 *)
fun stringOfTerm_expanded tab (tt, typ) : (string list * (string * string) list) =
	case tt of
	  Tuple (subst, tl) =>
	let val (po, sl) =
		if List.length tl = 1
		then
			let val (po, r) = stringOfTerm_expanded tab (List.hd tl)
			in
				if (null subst) then (po, r)
				else
					(po, (stringOfTab tab,
							" (* [" ^ (String.concatWith ", " (List.map (stringOfTerm false) subst)) ^ "] *)")
						:: r)
			end
		else
			let val r = stringOfTerm_compact (tt, typ)
			in
				(#1 r, [(stringOfTab tab, #2 r)])
			end
	in
		(po, (#1 (List.hd sl), (#2 (List.hd sl))) :: (List.tl sl))
	end
	
	| Abs (idl, t) =>
	let val r = stringOfTerm_expanded (tab + 1) t
	in
		((#1 r),
		(stringOfTab tab,
		"fn (" ^ (stringOfTypedIdentList idl) ^ ")"
		^ " =>") :: (#2 r))
	end
	
	| Rec (t1, t2, (Abs (idl1, (Abs (idl2, t3), typ2)), typ1)) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2
		val r3 = stringOfTerm_expanded (tab + 1) t3
	in
		((#1 r1) @ (#1 r2) @ (#1 r3),
		(stringOfTab tab,
		"Rec (" ^ (#2 r1) ^ ", " ^ (#2 r2) ^ ", "
		^ "fn " ^ (if List.length idl1 = 1 then "" else "(")
		^ (stringOfTypedIdentList idl1)
		^ (if List.length idl1 = 1 then "" else ")")
		^ " => fn (" ^ (stringOfTypedIdentList idl2) ^ ")" ^ " =>")
	   :: (append_last (#2 r3) " )"))
	end
	
	| Rec (t1, t2, t3) =>
	let val r1 = stringOfTerm_compact t1
		val r2 = stringOfTerm_compact t2
		val r3 = stringOfTerm_expanded (tab + 1) t3
	in
		((#1 r1) @ (#1 r2) @ (#1 r3),
		(stringOfTab tab,
		"Rec (" ^ (#2 r1) ^ ", " ^ (#2 r2) ^ ",")
		:: (append_last (#2 r3) " )"))
	end
	
	| LetIn (idl, t1, (LetIn (idl', t1', t2'), typ2)) =>
	let val r1 = stringOfTerm_expanded (tab + 1) t1
		val r2 = stringOfTerm_expanded tab (LetIn (idl', t1', t2'), typ2)
		val s2 = (stringOfTab tab, "      " ^ (String.extract (#2 (List.hd (#2 r2)), 4, NONE)))
					:: (List.tl (#2 r2))
	in
		((#1 r1) @ (#1 r2),
		(stringOfTab tab,
		"let val " ^ (if List.length idl = 1 then "" else "(")
		^ (stringOfTypedIdentList idl)
		^ (if List.length idl = 1 then " = " else ") = ") ^ (#2 (List.hd (#2 r1))))
		:: (List.tl (#2 r1))
		@ s2)
	end
	
	| LetIn (idl, t1, t2) =>
	let val r1 = stringOfTerm_expanded (tab + 1) t1
		val r2 = stringOfTerm_compact t2
	in
		((#1 r1) @ (#1 r2),
		(stringOfTab tab,
		"let val " ^ (if List.length idl = 1 then "" else "(")
		^ (stringOfTypedIdentList idl)
		^ (if List.length idl = 1 then " = " else ") = ") ^ (#2 (List.hd (#2 r1))))
	   :: (List.tl (#2 r1))
	   @ [(stringOfTab tab, "in " ^ (#2 r2) ^ " end")])
	end
	
	| Callcc (id, t) =>
	let val r = stringOfTerm_expanded (tab + 1) t
	in
		((#1 r),
		(stringOfTab tab,
		"callcc (fn " ^ (stringOfIdent id) ^ " =>")
	   :: (append_last (#2 r) ")"))
	end
	
	| Subst (t1, typi, id, t2) =>
	let val r1 = stringOfTerm_expanded tab t1
		val r2 = stringOfTerm_compact t2
	in
		((#1 r1) @ (#1 r2),
		append_last (#2 r1)
			(" (* :> {" ^ (stringOfIdent id) ^ "/" ^ (stringOfType typi) ^ "}[" ^ (stringOfType (#2 t2)) ^ (#2 r2) ^ "] *)"))
	end

	| _ => let val r = stringOfTerm_compact (tt, typ) in (#1 r, [(stringOfTab tab, #2 r)]) end

(**
 * builds a string representation of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the string representation of t.
 *)
fun stringOfTerm t =
	let val () = init ()
		val (pos, ss) = stringOfTerm_expanded 0 t
		val s = List.map (fn (a, b) => a ^ b) ss
	in
		s @ ("" :: List.map (fn s => "(* " ^ s ^ " *)") pos)
	end

end

