(**
 * The First-Order LoopW Proof Checker translator into functional syntax.
 *
 * The functional translator converts imperative programs into functional terms.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure FunTranslation :> FUN_TRANSLATION = struct

open ProofAst
open ProofAstUtils

(**
 * builds a typed singleton tuple from a term and a type.
 *
 * @params tterm typ
 * @param tterm the term.
 * @param typ the type.
 *
 * @return the tuple (tterm):(typ).
 *)
fun mkTuple tterm typ = (FunAst.Tuple ([], [tterm]), FunAst.TTuple ([], [typ]))

(**
 * orders an environment according to a given variable list.
 *
 * @params xl omega
 * @param xl the variable list [x1, ..., xn].
 * @param omega the environment [xi:ti]*.
 *
 * @throws Fail if length mismatch.
 *
 * @return the ordered environment [x1:t1, ..., xn:tn].
 *)
fun order xl omega =
    List.map (fn i => case ListAssoc.assoc i omega of
			  NONE => raise Fail "FunTranslation.order: list mismatch"
			| SOME t => (i, t)) xl

(**
 * orders a substitution according to a given variable list.
 *
 * @params xl subst
 * @param xl the variable list [x1, ..., xn].
 * @param subst the substitution [xi <- ti]*.
 *
 * @throws Fail if length mismatch.
 *
 * @return the ordered substitution [x1 <- t1, ..., xn <- tn].
 *)
fun orderSubst xl subst =
    List.map (fn i => case ListAssoc.assoc i subst of
			  NONE => raise Fail "FunTranslation.orderSubst: list mismatch"
			| SOME t => (i, t)) xl

(**
 * translates imperative types into functional types.
 *
 * @params t
 * @param t the imperative type.
 *
 * @return the functional type t*.
 *)
fun translateType t =
    case t of
		TBot => FunAst.TBot
      | TNat (t) => FunAst.TNat (t)
      | TEqual (n, m) => FunAst.TEqual (n, m)
      | TProc (i, inl, j, outl) =>
    	FunAst.TFun (i, FunAst.TTuple ([], List.map translateType inl),
    					FunAst.TTuple (j, List.map translateType outl))
      | TVar (i) => FunAst.TVar (i)

(**
 * translates an imperative sequence type into a function tuple type.
 *
 * @params (j, om)
 * @param j the identifier list of existential quantification.
 * @param om the formula list.
 *
 * @return the functional type Exist j.om* (TTuple (j, om* )).
 *)
fun omega2tuple (j, om) =
    FunAst.TTuple (j, List.map (fn (_,t) => translateType t) om)

(**
 * translate an imperative value into a functional typed term.
 *
 * @params (v, tau)
 * @param v the imperative value.
 * @param tau the type of v.
 *
 * @return the functional term v*.
 *)
fun translateValue (v, tau) =
    case v of
		Star => (FunAst.Tuple ([], []), translateType tau)

      |	Int (n) =>
    	let val rec succ_n = (* Builds (S^n(0), nat(s^n(0))) from n *)
    		fn n =>
    			if n = 0 then (FunAst.Z, FunAst.TNat (FZero))
    			else (case succ_n (n - 1) of
    					(tt, FunAst.TNat (a)) => (FunAst.S (tt, FunAst.TNat (a)), FunAst.TNat (s a))
    				  | _ => raise Fail "FunTranslation.translateValue.succ_n: impossible")
    	in
	    	succ_n n
	    end

      | Proc (i, inl, j, outl, (s, (k, indl))) =>
		(FunAst.Abs (List.map (fn (x,t) => (x, translateType t)) inl,
					translateSequence s (List.map #1 indl)),
		translateType tau)
		
	  | ValRegion (v, _) => translateValue (v, tau)

(**
 * translate an imperative expression into a functional typed term.
 *
 * @params (e, tau)
 * @param e the imperative expression.
 * @param tau the type of e.
 *
 * @return the functional term e*.
 *)
and translateExpression (e, tau) =
    case e of
		Id (id) => (FunAst.Id (id), translateType tau)
      | Val (v) => translateValue v
      | ESubst (e, t_i, i, e', tnm) =>
		(FunAst.Subst (translateExpression e,
						translateType t_i, i,
						translateExpression e'),
		translateType tau)
      | Lemma (fl, f) => (FunAst.Lemma (List.map translateType fl, translateType f), translateType tau)
      | ExpRegion (e, _) => translateExpression (e, tau)

(**
 * translate an imperative sequence into a functional typed term.
 *
 * @params (seq, (j, omega)) z
 * @param seq the imperative sequence.
 * @param j the identifier list of existential quantification.
 * @param omega the existentialy quantified formula list typing seq.
 * @param z the output variable vector.
 *
 * @return the functional term (s)*_z.
 *)
and translateSequence (seq, (j, omega)) z =
    let val out = omega2tuple (j, (order z omega)) in
	case seq of
	    Empty (subst) =>
	    let val omega' = List.map (fn (i,t)
					  => (i, translateType
						     (Terms.substituteList subst t)))
				      (order z omega) (* Builds the instantiated type *)
		val subst' = orderSubst j subst (* Orders the substitution list (to match the ident list in the type) *)
	    in
		(FunAst.Tuple (List.map #2 subst',
			       List.map (fn (x,t) => (FunAst.Id x, t)) omega'), out)
	    end

	  | Var (id, t, e, s) =>
	    (FunAst.LetIn ([(id, translateType t)], mkTuple (translateExpression e) (translateType t),
			   translateSequence s z), out)

	  | Cst (id, t, e, s) =>
	    (FunAst.LetIn ([(id, translateType t)], mkTuple (translateExpression e) (translateType t),
			   translateSequence s z), out)

	  | SSubst (s, omega_i, i, e, tnm) =>
	    (FunAst.Subst (translateSequence s z,
			   omega2tuple ([], List.filter (fn (x,t) => Utils.listMember x z) omega_i),
			   i, translateExpression e), out)

	  | Comm (Affect (id, t, e), s) =>
	    (FunAst.LetIn ([(id, translateType t)], mkTuple (translateExpression e) (translateType t),
			   translateSequence s z), out)

	  | Comm (Inc (id, t), s') =>
	    let val st = translateType (case t of TNat (x) => TNat (s x) | _ => raise Fail "Inc...") in
		(FunAst.LetIn ([(id, st)],
			       mkTuple (FunAst.Succ (FunAst.Id (id), translateType t), st) st,
			       translateSequence s' z), out)
	    end

	  | Comm (Dec (id, t), s) =>
	    let val st = translateType (case t of TNat (x) => TNat (p x) | _ => raise Fail "Inc...") in
		(FunAst.LetIn ([(id, st)],
			       mkTuple (FunAst.Pred (FunAst.Id (id), translateType t), st) st,
			       translateSequence s z), out)
	    end

	  | Comm (ProcCall (e, t, etl, subst, itl), s) =>
	    (case translateExpression e of
		 (trans_e, FunAst.TFun (i, sigma, FunAst.TTuple(k, tau))) =>
		 (FunAst.LetIn (List.map (fn (x,t) => (x, translateType t)) itl,
				(FunAst.App ((trans_e, FunAst.TFun (i, sigma, FunAst.TTuple(k, tau))), (* e : ALLi.(sigma => EXk.tau) *)
					     List.map #2 (order i subst), (* the term list for instantiation *)
					     (FunAst.Tuple ([], List.map (fn (e,t) => translateExpression e) etl),
					      omega2tuple ([], etl))),
				 omega2tuple (k, itl)),
				translateSequence s z), out)
	       | _ => raise Fail "FunTranslation.translateSequence: ProcCall and no fun type")

	  | Comm (Block (s1, (k, itl)), s2) =>
	    (FunAst.LetIn (List.map (fn (x,t) => (x, translateType t)) itl,
			   translateSequence s1 (List.map #1 itl),
			   translateSequence s2 z), out)

	  | Comm (For (id, lid, e, t, (s1, (k, itl))), s2) =>
	    let val outl = (List.map (fn (x, tau) =>
					 (x, Terms.substitute lid (case t of TNat (x) => x
									    | _ => raise Fail "For...") tau))
				     itl) (* phi[n/lid] with e : nat(n) *)
		val zerol = (List.map (fn (x, tau) =>
					  (x, Terms.substitute lid FZero tau))
				     itl) (* phi[0/lid] *)
		val delta_tuple = omega2tuple (k, itl)
		val delta_tuple' =
		    omega2tuple
			(k, (List.map (fn (x, t) =>
					  (x, Terms.substitute lid (s (FId (lid))) t))
				      itl)) (* phi[s(lid)/lid] *)
	    in
		(FunAst.LetIn
		     (List.map (fn (x,t) => (x, translateType t)) outl,
		      (FunAst.Rec (translateExpression e, (* e : Nat(n) *)
				   (FunAst.Tuple ([], List.map (fn (x,t) => (FunAst.Id x, translateType t)) zerol),
				    omega2tuple (k, zerol)), (* z : phi[0/lid] *)
				   (FunAst.Abs ([(id, FunAst.TNat (FId (lid)))], (* fn id : Nat(lid) => *)
						(FunAst.Abs (* fn z : phi => *)
						     (List.map (fn (x,t) => (x, translateType t)) itl,
						      translateSequence s1 (List.map #1 itl)), (* s1* : phi[s(lid)/lid] *)
						 FunAst.TFun (k, delta_tuple, delta_tuple'))),
				    FunAst.TFun ([lid], FunAst.TTuple ([], [FunAst.TNat (FId (lid))]),
						 FunAst.TFun (k, delta_tuple, delta_tuple')))),
		       omega2tuple (k, outl)), (* : phi[n/lid] *)
		      translateSequence s2 z), out)
	    end

	  | Comm (Jump (e, t, etl, subst, itl), s) =>
	    (case translateExpression e of
		 (trans_e, FunAst.TFun (i, sigma, FunAst.TTuple ([], [FunAst.TBot]))) =>
		 (FunAst.LetIn (List.map (fn (x,t) => (x, translateType t)) itl,
				(FunAst.Throw ((trans_e,
						FunAst.TFun (i, sigma, FunAst.TTuple ([], [FunAst.TBot]))),
					       List.map #2 (order i subst), (* the term list for instantiation *)
					       (FunAst.Tuple ([], List.map (fn (e,t) => translateExpression e) etl),
						omega2tuple ([], etl))),
				 omega2tuple ([], itl)),
				translateSequence s z), out)
	       | _ => raise Fail "FunTranslation.translateSequence: Throw and no neg type")

	  | Comm (Label (id, t, (s1, (k, itl))), s2) =>
	    (FunAst.LetIn (List.map (fn (x,t) => (x, translateType t)) itl,
			   (FunAst.Callcc (id, translateSequence s1 (List.map #1 itl)),
			    omega2tuple (k, itl)),
			   translateSequence s2 z), out)

	  | Comm (CommRegion (c, _), s) => translateSequence (Comm (c, s), (j, omega)) z
	  | SeqRegion (s, _) => translateSequence (s, (j, omega)) z
    end

(**
 * translate a complete imperative program into a functional typed term.
 *
 * @params p
 * @param p the imperative program.
 *
 * @return the functional term p*.
 *)
fun translateProgram p = translateSequence p []

end

