structure FunTypeChecker = struct

    open ProofAst
    open FunAst

    fun s t = FFun (("s",0), [t])
    fun p t = FFun (("p",0), [t])

    fun mem l x = List.exists (fn y => x = y) l
    fun subset l l' = List.all (mem l') l

    infix -- @@ ##
    fun l -- l' = List.filter (not o (mem l')) l

    (* If l and l' are sets the union l @@ l' is also a set *)
    fun l @@ l' = (l -- l') @ l'

    (* If l and l' are sets the intersection l ## l' is also a set *)
    fun l ## l' = List.filter (mem l') l

    fun union l = List.foldr op@@ [] l

    fun vars t =
        case t of
            FId x => [x]
          | FZero => []
          | FFun (i, l) => [i] @@ (union (map vars l))

    fun freeVars ty =
        case ty of
            TBot => []
          | TNat n => vars n
          | TEqual (n, m) => (vars n) @@ (vars n)
          | TTuple (il, tyl) => (union (map freeVars tyl)) -- il
          | TFun (il, ty1, ty2) => ((freeVars ty1) @@ (freeVars ty2)) -- il
          | TVar (i) => []

    fun subst_term i m t =
        case t of
            FId x => if x = i then m else FId x
          | FZero => FZero
          | FFun (j, l) =>
            if j = i then raise Fail "Function symbol substitution not supported"
            else FFun (j, (map (subst_term i m) l))

    fun subst i m ty =
        case ty of
            TBot => TBot
          | TNat n => TNat (subst_term i m n)
          | TEqual (n1, n2) => TEqual ((subst_term i m n1), (subst_term i m n2))
          | TTuple (il, tyl) =>
            if null (il ## (vars m))
            then TTuple (il, map (subst i m) tyl)
            else raise Fail "alpha conversion required"
          | TFun (il, ty1, ty2) =>
            if null (il ## (vars m))
            then TFun (il, (subst i m ty1), (subst i m ty2))
            else raise Fail "alpha conversion required"
          | TVar (i) => TVar (i)

    fun subst_list [] [] ty = ty
      | subst_list (i::il) (m::ml) ty = subst i m (subst_list il ml ty)
      | subst_list _ _ _ = raise Fail "unmatched substitution list"

    fun alphaEqual (ty1, ty2) =
    	case (ty1, ty2) of
    		(TBot, TBot) => true
    	  | (TNat (n), TNat (m)) => n = m
    	  | (TEqual (n1, n2), TEqual (m1, m2)) => (n1 = m1) andalso (n2 = m2)
    	  | (TVar (i), TVar (j)) => i = j
    	  | (TTuple (il1, tyl1), TTuple (il2, tyl2)) =>
    		(List.length il1 = List.length il2)
    		andalso (List.length tyl1 = List.length tyl2)
    		andalso
    		let val jl = List.map Terms.alpha il1
    			val tyl1' = List.map (subst_list il1 (List.map (fn i => FId (i)) jl)) tyl1
    			val tyl2' = List.map (subst_list il2 (List.map (fn i => FId (i)) jl)) tyl2
    		in
    			List.all alphaEqual (ListPair.zip (tyl1', tyl2'))
    		end
    	  | (TFun (il1, ty1, oty1), TFun (il2, ty2, oty2)) =>
    		(List.length il1 = List.length il2)
    		andalso
    		let val jl = List.map Terms.alpha il1
    			val ty1' = subst_list il1 (List.map (fn i => FId (i)) jl) ty1
    			val oty1' = subst_list il1 (List.map (fn i => FId (i)) jl) oty1
    			val ty2' = subst_list il2 (List.map (fn i => FId (i)) jl) ty2
    			val oty2' = subst_list il2 (List.map (fn i => FId (i)) jl) oty2
    		in
    			(alphaEqual (ty1', ty2')) andalso (alphaEqual (oty1', oty2'))
    		end
    	  | _ => false

    exception TypeError of (ident * expType) list * typed_term

    fun typeCheckVar [] (x, ty) = false
      | typeCheckVar ((x', ty')::tl) (x, ty) =
        if x' = x then alphaEqual (ty', ty) else typeCheckVar tl (x, ty)

    fun typeCheck G (t, ty) =
        if typeCheck' G (t, ty) then true else raise (TypeError (G, (t, ty)))
    and typeCheck' G (t, ty) =
        case t of
            Id x
            => typeCheckVar G (x, ty)
          | Tuple ([], [])
            => (case ty of
                    TEqual(n, m) => true
                  | TTuple([], []) => true
                  | _ => false)
          | Tuple (nl, tl)
            => (case ty of
                    TTuple(il, tyl) =>
                    List.all (typeCheck G) tl
                    andalso List.all alphaEqual (ListPair.zip (map #2 tl, List.map (subst_list il nl) tyl))
                  | _ => false)
          | Z
            => ty = TNat(FZero)
          | S tt
            => (case tt of
                    (_, TNat(m)) => typeCheck G tt andalso ty = TNat(s(m))
                  | _ => false)
          | Succ tt
            => (case tt of
                    (_, TNat(m)) => typeCheck G tt andalso ty = TNat(s(m))
                  | _ => false)
          | Pred tt
            => (case tt of
                    (_, TNat(m)) => typeCheck G tt andalso ty = TNat(p(m))
                  | _ => false)
          | App ((t1,ty1), nl, (t2,ty2))
            => (case ty1 of
                    TFun(il, sigma, tau) =>
                    typeCheck G (t1, ty1) andalso typeCheck G (t2, ty2)
                    andalso (alphaEqual (subst_list il nl sigma, ty2) orelse raise Fail ("OOOO\n\n" ^ (FunPrettyPrinter.stringOfType sigma) ^ "\n\n//\n\n" ^ (FunPrettyPrinter.stringOfType ty2)))
                    andalso alphaEqual (subst_list il nl tau, ty)
                  | _ => false)
          | Abs (til, (t1, ty1))
            => (case ty of
                    TFun (il, sigma, tau) =>
                    typeCheck (til @ G) (t1, ty1)
                    andalso alphaEqual (sigma, TTuple([], map #2 til))
                    andalso alphaEqual (tau, ty1)
                    andalso null (il ## (union (map (freeVars o #2) G)))
                  | _ => false)
          | Rec ((t1, ty1), (t2, ty2), (t3, ty3))
            => (case (ty1, ty3) of
                    (TNat(m), TFun([i], TTuple([], [TNat(j)]), TFun([], tau, tau'))) =>
                    (FId i) = j
                    andalso typeCheck G (t1, ty1)
                    andalso typeCheck G (t2, ty2)
                    andalso typeCheck G (t3, ty3)
                    andalso alphaEqual (ty2, subst i FZero tau)
                    andalso alphaEqual (tau', subst i (s (FId i)) tau)
                    andalso alphaEqual (ty, subst i m tau)
                  | _ => false)
          | LetIn (til, (t1, ty1), (u, tau))
            => (case ty1 of
                    TTuple (il, tyl) =>
                    typeCheck G (t1, ty1)
                    andalso typeCheck (til @ G) (u, tau)
                    andalso List.all alphaEqual (ListPair.zip (map #2 til, tyl))
                    andalso null (il ## (union (map freeVars (tau::(map #2 G)))))
                  | _ => false)
          | Callcc (k, (t1, ty1))
            => (case ty1 of
                    TTuple (il, tyl) =>
                    alphaEqual (ty1, ty)
                    andalso typeCheck ((k, TFun (il, TTuple ([], tyl), TTuple ([], [TBot])))::G) (t1, ty1)
                  | _ => false)
          | Throw ((t1, ty1), nl, (t2, ty2))
            => (case ty1 of
                    TFun (il, tau, TTuple ([], [TBot])) =>
                    typeCheck G (t1, ty1)
                    andalso typeCheck G (t2, ty2)
                    andalso alphaEqual (subst_list il nl tau, ty2)
                  | _ => false)
          | Subst ((t1, ty1), tau, i, (u, eq))
            => (case eq of
                    TEqual(n, m) =>
                    typeCheck G (t1, ty1) andalso typeCheck G (u, eq)
                    andalso alphaEqual (subst i n tau, ty1) andalso alphaEqual (subst i m tau, ty)
                  | _ => false)
          | Lemma (tl, ty1)
            => subset tl (map #2 G) andalso alphaEqual (ty, ty1)

end

