(* -*- sml-lex -*- *)
structure T = Tokens

type pos = int
type svalue = T.svalue
type ('a, 'b) token = ('a, 'b) T.token
type lexresult = (svalue, pos) token

type lexarg =
     ((pos * pos -> lexresult) -> bool) *
     ((pos * pos -> lexresult) -> bool) *
     (string * (pos * pos -> lexresult)) list *
     (pos * pos -> string -> unit)

type arg = lexarg

(* keywords *)
fun idToken keywords (t, p) = let
    val t' = String.map Char.toLower t
in
    case List.find (fn (id, _) => id = t') keywords of
	NONE => T.Id (t, p, p + size t)
      | SOME (_, tok) => tok (p, p + size t)
end

(* regions *)
val current = ref ""
val (line, linePos) = (ref 1, ref [1])

fun initLexer name = (current := name; line := 1; linePos := [1])

fun convert (p:int, a::rest, n) =
    if a < p then (n, (p - a)) else convert (p, rest, n-1)
  | convert _ = (0, 0)

fun locate' (pos1, pos2) = let
    val (l1, _) = convert (pos1, !linePos, !line)
    val (l2, _) = convert (pos2, !linePos, !line)
in (!current, l1, l2) end

fun locate (pos1, pos2) = let
    val (l1, c1) = convert (pos1, !linePos, !line)
    val (l2, c2) = convert (pos2, !linePos, !line)
in (!current, (l1, c1), (l2, c2)) end

(* new lines *)
fun newline pos = (line := !line + 1; linePos := pos :: !linePos)

(* comments *)
val depth = ref 0
fun enterC () = depth := !depth + 1
fun leaveC () = let val d = !depth - 1 in depth := d; d = 0 end

(* string litterals *)
val curstring = ref ([] : char list)
val startpos = ref 0
val instring = ref false
fun newS sp = (startpos := sp; curstring := []; instring := true)
fun addS c = curstring := c :: !curstring
fun getS ep = (instring := false; (implode (rev (!curstring)), !startpos, ep))

fun eof (_, _, _, lexerror) = let
    val pos = hd (!linePos)
in
    if !depth > 0 then
	lexerror (pos, pos) "unexpected EOF in comment"
    else if !instring then
	lexerror (pos, pos) "unexpected EOF in string literal"
    else ();
    T.EOF (pos, pos)
end %%

    %s COMMENT1 COMMENT2 STRING;
    %header (functor SyntaxLexFun (structure Tokens: Syntax_TOKENS));
    %arg (blockComment, lineComment, kwds, lexerror);

    alphanum=[a-zA-Z][a-zA-Z0-9_]*; letter=[a-zA-Z]; digit=[0-9];
    exp=[eE][+-]?{digit}+; frac={digit}+"."{digit}+; spc=[\ \t\r\f];
    float={frac}{exp}?|{digit}+{exp};

    %%
    <COMMENT1>"/*"            => (enterC (); continue());
    <COMMENT1>"*/"            => (if leaveC () then YYBEGIN INITIAL else ();
				  continue());
    <COMMENT1>"//".*"\n"      => (newline (yypos + size yytext - 1);
				  continue());
    <COMMENT1>"\n"            => (newline yypos; continue());
    <COMMENT1>.               => (continue());
    <COMMENT2>"(*"            => (enterC (); continue());
    <COMMENT2>"*)"            => (if leaveC () then YYBEGIN INITIAL else ();
				  continue());
    <COMMENT2>"//".*"\n"      => (newline (yypos + size yytext - 1);
				  continue());
    <COMMENT2>"\n"            => (newline yypos; continue());
    <COMMENT2>.               => (continue());
    <INITIAL>"/*"             => (if blockComment T.SLASHSTAR then
				      (YYBEGIN COMMENT1; enterC (); continue())
				  else T.SLASHSTAR(yypos, yypos + 2));
    <INITIAL>"(*"             => (if blockComment T.LPARSTAR then
				      (YYBEGIN COMMENT2; enterC (); continue())
				  else T.LPARSTAR (yypos, yypos + 2));
    <INITIAL>"*)"             => (T.STARRPAR (yypos, yypos + 2));
    <INITIAL>"*/"             => (T.STARSLASH (yypos, yypos + 2));
    <INITIAL>"//".*"\n"       => (if lineComment T.DBLSLASH then
				      (newline (yypos + size yytext - 1);
				       continue())
				  else T.DBLSLASH (yypos, yypos + 2));
    <INITIAL>"--"[^#\n].*"\n" => (if lineComment T.DBLMINUS then
				      (newline (yypos + size yytext - 1);
				       continue())
				  else T.DBLMINUS (yypos, yypos + 2));
    <INITIAL>"--\n"           => (if lineComment T.DBLMINUS then
				      (newline (yypos + size yytext - 1);
				       continue())
				  else T.DBLMINUS (yypos, yypos + 2));
    <INITIAL>"--#"            => (T.DBLMINUSSHARP (yypos, yypos + 3));
    <INITIAL>{alphanum}       => (idToken kwds (yytext, yypos));
    <INITIAL>"||"             => (T.DBLBAR (yypos, yypos + 2));
    <INITIAL>"&&"             => (T.DBLAMPERSAND (yypos, yypos + 2));
    <INITIAL>"|"              => (T.BAR (yypos, yypos + 1));
    <INITIAL>"&"              => (T.AMPERSAND (yypos, yypos + 1));
    <INITIAL>"<"              => (T.LESSTHAN (yypos, yypos + 1));
    <INITIAL>">"              => (T.GREATERTHAN (yypos, yypos + 1));
    <INITIAL>">="             => (T.GREATEREQ (yypos, yypos + 2));
    <INITIAL>"<="             => (T.LESSEQ (yypos, yypos + 2));
    <INITIAL>"=>"             => (T.DBLRIGHTARROW (yypos, yypos + 1));
    <INITIAL>"->"             => (T.RIGHTARROW (yypos, yypos + 1));
    <INITIAL>"<-"             => (T.LEFTARROW (yypos, yypos + 1));
    <INITIAL>"!="             => (T.NOTEQUALTO (yypos, yypos + 2));
    <INITIAL>"/="             => (T.DIFFER (yypos, yypos + 2));
    <INITIAL>"/"              => (T.SLASH (yypos, yypos + 1));
    <INITIAL>"\\"             => (T.BACKSLASH (yypos, yypos + 1));
    <INITIAL>"=="             => (T.EQUALTO (yypos, yypos + 2));
    <INITIAL>":="             => (T.ASSIGN (yypos, yypos + 2));
    <INITIAL>"="              => (T.EQUAL (yypos, yypos + 1));
    <INITIAL>"*"              => (T.STAR (yypos, yypos + 1));
    <INITIAL>"-"              => (T.MINUS (yypos, yypos + 1));
    <INITIAL>"+"              => (T.PLUS (yypos, yypos + 1));
    <INITIAL>".."             => (T.DBLDOT (yypos, yypos + 1));
    <INITIAL>"."              => (T.DOT (yypos, yypos + 1));
    <INITIAL>"{"              => (T.LCURL (yypos, yypos + 1));
    <INITIAL>"}"              => (T.RCURL (yypos, yypos + 1));
    <INITIAL>"["{spc}*"]"     => (T.LRSQUARE (yypos, yypos + size yytext));
    <INITIAL>"["              => (T.LSQUARE (yypos, yypos + 1));
    <INITIAL>"]"              => (T.RSQUARE (yypos, yypos + 1));
    <INITIAL>"("              => (T.LPAR (yypos, yypos + 1));
    <INITIAL>")"              => (T.RPAR (yypos, yypos + 1));
    <INITIAL>";"              => (T.SEMI (yypos, yypos + 1));
    <INITIAL>":"              => (T.COLON (yypos, yypos + 1));
    <INITIAL>"::"             => (T.DBLCOLON (yypos, yypos + 2));
    <INITIAL>":>"             => (T.CAST (yypos, yypos + 2));
    <INITIAL>"|>"             => (T.SEQTYPE (yypos, yypos + 2));
    <INITIAL>"|-"             => (T.VDASH (yypos, yypos + 2));
    <INITIAL>","              => (T.COMMA (yypos, yypos + 1));
    <INITIAL>"!"              => (T.EXCLAM (yypos, yypos + 1));
    <INITIAL>"?"              => (T.QUESTION (yypos, yypos + 1));
    <INITIAL>"#"              => (T.SHARP (yypos, yypos + 1));
    <INITIAL>"`"              => (T.BACKQUOTE (yypos, yypos + 1));
    <INITIAL>"'"              => (T.QUOTE (yypos, yypos + 1));
    <INITIAL>"%"              => (T.PERCENT (yypos, yypos + 1));
    <INITIAL>"$"              => (T.DOLLAR (yypos, yypos + 1));
    <INITIAL>"^"              => (T.CARRET (yypos, yypos + 1));
    <INITIAL>"_"              => (T.UNDERSCORE (yypos, yypos + 1));
    <INITIAL>"~"              => (T.TILDE (yypos, yypos + 1));
    <INITIAL>{digit}+         => (T.Integer
				      (case Int.fromString yytext of
					   SOME n => n
					 | NONE =>
					   (lexerror
						(yypos, yypos + size yytext)
						("illegal number"); 0),
				       yypos, yypos + size yytext));
    <INITIAL>{float}          => (T.Real
				      (case Real.fromString yytext of
					   SOME r => r
					 | NONE =>
					   (lexerror
						(yypos, yypos + size yytext)
						("illegal real"); 0.0),
				       yypos, yypos + size yytext));
    <INITIAL>\n               => (newline yypos; continue());
    <INITIAL>{spc}            => (continue());
    <INITIAL>"\""             => (YYBEGIN STRING; newS yypos; continue());
    <STRING>"\\n"             => (addS #"\n"; continue());
    <STRING>"\\t"             => (addS #"\t"; continue());
    <STRING>"\\\""            => (addS #"\""; continue());
    <STRING>"\\\\"            => (addS #"\\"; continue());
    <STRING>"\\".             => (lexerror
				      (yypos, yypos+1)
				      ("illegal escape character in STRING "
				       ^ yytext);
				  continue());
    <STRING>"\""              => (YYBEGIN INITIAL; T.String (getS yypos));
    <STRING>\n                => (lexerror (yypos, yypos+1)
					   "illegal linebreak in STRING";
				  continue());
    <STRING>.                 => (addS (String.sub (yytext, 0)); continue());
    <INITIAL>.                => (lexerror (yypos, yypos)
					   ("illegal character " ^ yytext);
				  continue());
