(**
 * Utility functions that deals with association lists.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
signature LIST_ASSOC = sig

(**
 * returns the list of keys.
 *
 * @params l
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return the list of keys [x1, ..., xn].
 *)
val getKeyList : (''a * ''b) list -> ''a list

(**
 * returns the list of values.
 *
 * @params l
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return the list of keys [v1, ..., vn].
 *)
val getValueList : (''a * ''b) list -> ''b list

(**
 * test if a key belongs to the association list.
 *
 * @params x l
 * @param x the key to test.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return true if and only if x is one of x1, ..., xn.
 *)
val memberKey : ''a -> (''a * ''b) list -> bool

(**
 * test if an association belongs to the association list.
 *
 * @params (x, v) l
 * @param x the key to test.
 * @param v the value to test.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return true if and only if (x, v) is one of (x1,v1), ..., (xn, vn).
 *)
val assoc : ''a -> (''a * ''b) list -> ''b option

(**
 * returns all the elements of the association for which the key belongs to
 * the given key list.
 *
 * @params kl l
 * @param kl the key list.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (xi,vi) for all i such that xi belongs to kl.
 *)
val filterKey : ''a list -> (''a * ''b) list -> (''a * ''b) list

(**
 * returns the association list without any association matching the given key.
 *
 * @params k l
 * @param k the key.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (xi,vi) for all i such that xi <> k.
 *)
val removeKey : ''a -> (''a * ''b) list -> (''a * ''b) list

(**
 * returns the association list without any association matching the given key list.
 *
 * @params kl l
 * @param kl the key list.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (xi,vi) for all i such that xi does not belongs to kl.
 *)
val removeAllKey : ''a list -> (''a * ''b) list -> (''a * ''b) list

(**
 * adds a new association, ensuring that only one association has this key.
 *
 * @params (x, v) l
 * @param xl the key.
 * @param v the value.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (x,v)::l' where l' is l without all associations having key k.
 *)
val add : (''a * ''b) -> (''a * ''b) list -> (''a * ''b) list

(**
 * adds a list of new association, ensuring that only one association has the added keys.
 *
 * @params l1 l2
 * @param l1 the association list [(x1,v1), ..., (xn, vn)].
 * @param l2 the association list [(y1,w1), ..., (ym, wm)].
 *
 * @return [(x1,v1), ..., (xn, vn)]@l' where l' is l without all associations having
 * key belonging to x1, ..., xn.
 *)
val addAll : (''a * ''b) list -> (''a * ''b) list -> (''a * ''b) list

(**
 * unite to association lists. Same key must have the same values.
 *
 * @params l1 l2
 * @param l1 the association list [(x1,v1), ..., (xn, vn)].
 * @param l2 the association list [(y1,w1), ..., (ym, wm)].
 *
 * @return l1'@l2 (where l1' is l1 without all associations having
 * key belonging to y1, ..., yn), or NONE if the same key appears with
 * different values.
 *)
val union : (''a * ''b) list -> (''a * ''b) list -> (''a * ''b) list option

(**
 * unite many association lists, flattening them. Same key must have the same values.
 *
 * @params ll
 * @param ll the association list list.
 *
 * @return the flattened union (as in union), or NONE if the same key appears with
 * different values.
 *)
val unionFlat : (''a * ''b) list list -> (''a * ''b) list option

(**
 * test if all the keys occurs in the association list.
 *
 * @params kl l
 * @param kl the key list.
 * @param l the association list.
 *
 * @return true if and only if all key in kl belong to l.
 *)
val existsAllKey : ''a list -> (''a * ''b) list -> bool

(**
 * test if the first association list is a sub-list (without ordering consideration)
 * of the second list.
 *
 * @params l1 l2
 * @param l1 the first association list.
 * @param l2 the second association list.
 *
 * @return true if and only if all associations in l1 belongs to l2.
 *)
val areAllIn : (''a * ''b) list -> (''a * ''b) list -> bool

(**
 * partitions an association list in two list, according to the keys bonlonging to
 * the given key list.
 *
 * @params k1 l
 * @param kl the key list.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return ([(xi,vi), ..., (xk, vk)], [(xj,vj), ..., (xl, vl)]) where
 * (xi, ..., xk) belongs to kl, and (xj, ..., xl) does not belongs to kl.
 *)
val partitionWithKeyList : ''a list -> (''a * ''b) list -> (((''a * ''b) list) * ((''a * ''b) list))

end

