(**
 * Utility functions that deals with association lists.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ListAssoc :> LIST_ASSOC = struct

(**
 * returns the list of keys.
 *
 * @params l
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return the list of keys [x1, ..., xn].
 *)
fun getKeyList l = List.map (fn (x,y) => x) l

(**
 * returns the list of values.
 *
 * @params l
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return the list of keys [v1, ..., vn].
 *)
fun getValueList l = List.map (fn (x,y) => y) l

(**
 * test if a key belongs to the association list.
 *
 * @params x l
 * @param x the key to test.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return true if and only if x is one of x1, ..., xn.
 *)
fun memberKey x l =
    List.exists (fn (a,b) => x = a) l

(**
 * test if an association belongs to the association list.
 *
 * @params (x, v) l
 * @param x the key to test.
 * @param v the value to test.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return true if and only if (x, v) is one of (x1,v1), ..., (xn, vn).
 *)
fun isAssoc (x, v) l =
    List.exists (fn (a,b) => (x = a) andalso (v = b)) l

(**
 * return the associated value of a key, or NONE.
 *
 * @params x l
 * @param x the key to test.
 * @param v the value to test.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return true if and only if (x, v) is one of (x1,v1), ..., (xn, vn).
 *)
fun assoc x l =
    case List.find (fn (a,b) => x = a) l of
        NONE => NONE
      | SOME (a,b) => SOME b

(**
 * returns all the elements of the association for which the key belongs to
 * the given key list.
 *
 * @params kl l
 * @param kl the key list.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (xi,vi) for all i such that xi belongs to kl.
 *)
fun filterKey xl l =
    List.filter (fn (a,b) => Utils.listMember a xl) l

(**
 * returns the association list without any association matching the given key.
 *
 * @params k l
 * @param k the key.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (xi,vi) for all i such that xi <> k.
 *)
fun removeKey x l =
    List.filter (fn (a,b) => x <> a) l

(**
 * returns the association list without any association matching the given key list.
 *
 * @params kl l
 * @param kl the key list.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (xi,vi) for all i such that xi does not belongs to kl.
 *)
fun removeAllKey xl l =
    List.filter (fn (a,b) => not (Utils.listMember a xl)) l

(**
 * adds a new association, ensuring that only one association has this key.
 *
 * @params (x, v) l
 * @param xl the key.
 * @param v the value.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return (x,v)::l' where l' is l without all associations having key k.
 *)
fun add (x,v) l = (x,v)::(removeKey x l)

(**
 * adds a list of new association, ensuring that only one association has the added keys.
 *
 * @params l1 l2
 * @param l1 the association list [(x1,v1), ..., (xn, vn)].
 * @param l2 the association list [(y1,w1), ..., (ym, wm)].
 *
 * @return [(x1,v1), ..., (xn, vn)]@l2' where l2' is l2 without all associations having
 * key belonging to x1, ..., xn.
 *)
fun addAll l1 l2 =
    l1 @ (removeAllKey (getKeyList l1) l2)

(**
 * unite to association lists. Same key must have the same values.
 *
 * @params l1 l2
 * @param l1 the association list [(x1,v1), ..., (xn, vn)].
 * @param l2 the association list [(y1,w1), ..., (ym, wm)].
 *
 * @return l1'@l2 (where l1' is l1 without all associations having
 * key belonging to y1, ..., yn), or NONE if the same key appears with
 * different values.
 *)
fun union l1 = fn
    [] => SOME l1
  | (x,t)::q =>
    if not (memberKey x l1) then
        (case union l1 q of
             NONE => NONE
           | SOME l => SOME ((x,t)::l))
    else if isAssoc (x,t) l1 then union l1 q
    else NONE

(**
 * unite many association lists, flattening them. Same key must have the same values.
 *
 * @params ll
 * @param ll the association list list.
 *
 * @return the flattened union (as in union), or NONE if the same key appears with
 * different values.
 *)
fun unionFlat ll =
    case ll of
        [] => SOME []
      | l1::q =>
        (case unionFlat q of
             NONE => NONE
           | SOME l2 => union l1 l2)

(**
 * test if all the keys occurs in the association list.
 *
 * @params kl l
 * @param kl the key list.
 * @param l the association list.
 *
 * @return true if and only if all key in kl belongs to l.
 *)
fun existsAllKey xl l =
    List.all (fn x => memberKey x l) xl

(**
 * test if the first association list is a sub-list (without ordering consideration)
 * of the second list.
 *
 * @params l1 l2
 * @param l1 the first association list.
 * @param l2 the second association list.
 *
 * @return true if and only if all associations in l1 belongs to l2.
 *)
fun areAllIn l1 l2 =
    List.all (fn xv => isAssoc xv l2) l1

(**
 * partitions an association list in two list, according to the keys bonlonging to
 * the given key list.
 *
 * @params k1 l
 * @param kl the key list.
 * @param l the association list [(x1,v1), ..., (xn, vn)].
 *
 * @return ([(xi,vi), ..., (xk, vk)], [(xj,vj), ..., (xl, vl)]) where
 * (xi, ..., xk) belongs to kl, and (xj, ..., xl) does not belongs to kl.
 *)
fun partitionWithKeyList xl l =
    List.partition (fn (a,b) => Utils.listMember a xl) l

end

