(**
 * The First-Order LoopW Proof Checker proof abstract syntax.
 *
 * The proof abstract syntax is used to represent annotated LoopW programs, i.e. iCND proofs.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ProofAst = struct

open Error

(** identifiers as string (numeroted for alpha-conversion). *)
type ident = string * int

(** terms in the logic. *)
datatype term =
	(** identifier. *)
	  FId of ident
	(** zero: Z *)
	| FZero
	(** n-ary function (s and p will be instances of this): f(tl) *)
	| FFun of ident * term list

(** first-order formulas. *)
datatype formula =
	(** nat predicate: nat(t) corresponds to "t in nat". *)
	  TNat of term
	(** procedures: proc(il, fl, jl, hl) corresponds to "forall il.(fl => exists jl.hl)". *)
	| TProc of ident list * formula list * ident list * formula list
	(** equality predicate over terms: t = t'. *)
	| TEqual of term * term
	(** absurd: _|_. *)
	| TBot
	(** formula meta-variable (only used in input programs). *)
	| TVar of ident

(** commands. *)
datatype command =
	(** block: { s }{i}x:tau *)
	  Block of block
	(** for-loop: for I : nat(i) := e : tau { s }{j}x:tau[i]. *)
	| For of ident * ident * expression * formula * block
	(** assignment: x : tau := e. *)
	| Affect of ident * formula * expression
	(** incrementation: inc(x : tau). *)
	| Inc of ident * formula
	(** decrementation: dec(x : tau). *)
	| Dec of ident * formula
	(** procedure call: p:tau(etl [subst], itl). *)
	| ProcCall of
		(** p : proc(tau[i], sig) *)
		expression * formula
		(** (e1 : tau1[u], ..., en : taun[u]; *)
		* (expression * formula) list
		(** [u/i] *)
		* (ident * term) list
		(** r1 : sig1, ..., rp : sigp) *)
		* (ident * formula) list
	(** label definition: k : {i}~tau :{ s }{i}x:tau. *)
	| Label of ident * formula * block
	(** jump to a label: jump(k : ~tau [subst], etl)x:tau. *)
	| Jump of
		(** k : ~tau *)
		expression * formula
		(* (e1 : tau1[u], ..., en : taun[u]; *)
		* (expression * formula) list
		(* [u/i] *)
		* (ident * term) list
		(* r1 : sig1, ..., rp : sigp *)
		* (ident * formula) list
	| CommRegion of command * region

(** untyped sequences. *)
and i_sequence =
	(** empty sequence : [n/j] "substitution used for existential introduction". *)
	  Empty of (ident * term) list
	(** composed command: c; s. *)
	| Comm of command * sequence
	(** constant declaration: cst x : tau = e; s. *)
	| Cst of ident * formula * expression * sequence
	(** variable declaration: var x : tau := e; s. *)
	| Var of ident * formula * expression * sequence
	(** sequence substitution: s :> {i/({j}Omega)[i]} [e : (n = m)]. *)
	| SSubst of
		(** s *)
		sequence
		(** ({j}Omega)[i] *)
		* (ident * formula) list
		(** i *)
		* ident
		(** e : (n = m) *)
		* expression * formula
	| SeqRegion of i_sequence * region

(** untyped expressions. *)
and i_expression =
	(** identifier. *)
	  Id of ident
	(** value. *)
	| Val of value
	(** expression substitution: e :> {i/tau[i]} [e' : (n = m)]. *)
	| ESubst of
		(** e *)
		expression
		(** tau[i] *)
		* formula
		(** i *)
		* ident
		(** e' : (n = m) *)
		* expression * formula
	(** lemma (equation over term in a context): G |- n = m. *)
	| Lemma of formula list * formula
	| ExpRegion of i_expression * region

(** untyped values. *)
and i_value =
	(** null value: *. *)
	  Star
	(** natural number: n. *)
	| Int of int
	(** anonymous procedure declaration: proc({il} in xfl; {jl} out yfl){ s }{jl}yfl. *)
	| Proc of
		(** {il} in x:fl *)
		ident list * (ident * formula) list
		(** {jl} out y:fl' *)
		* ident list * (ident * formula) list
		(** { s }{jl}yfl *)
		* block
	| ValRegion of i_value * region

withtype
(** typed sequences s |> {i}x:tau. *)
	sequence = i_sequence * (ident list * (ident * formula) list)
(** typed expressions e : tau. *)
and expression = i_expression * formula
(** typed values v : tau. *)
and value = i_value * formula
(** blocks { s |> {i'}x':tau' }{i}x:tau. *)
and block = (i_sequence * (ident list * (ident * formula) list))
		 * (ident list * (ident * formula) list)

(** programs *)
type program = sequence

end

