(**
 * The First-Order LoopW Proof Checker proof checking.
 *
 * The proof checker verify that the given program respresents a correct proof.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ProofChecker :> PROOF_CHECKER = struct

open ProofAst
open ProofAstUtils
open ProofEnvironment
open Terms

open BasePrettyPrinter

(**
 * displays an error message with the current proof checking environment and the source code region.
 *
 * @params gamma omega (j, omega') seq reg msg
 * @param gamma the constant proof checking environment.
 * @param omega the variable proof checking environment.
 * @param j the out type existential quantified variables.
 * @param omega' the out type.
 * @param seq the sequence that generates the error.
 * @param k the seq out type existential quantified variables.
 * @param delta the seq out type.
 * @param reg the region.
 * @param msg the error message.
 *)
fun typeFailureSeq gamma omega (j, omega') (seq, (k, delta)) reg msg =
	Error.fail reg ("\nGamma: [" ^ (stringOfTypedIdentList false false false gamma) ^ "]\nOmega: ["
		^ (stringOfTypedIdentList false false false omega) ^ "]\nj: {"
		^ (stringOfIdentList j) ^ "}\nOmega': ["
		^ (stringOfTypedIdentList false false false omega') ^ "]\nk: {"
		^ (stringOfIdentList k) ^ "}\nDelta: ["
		^ (stringOfTypedIdentList false false false delta) ^ "]\n\n"
		(*^ (ProofPrettyPrinter.stringOfSequence seq ^ "\n")*)
		^ msg)

fun typeFailureInSeq gamma omega (j, omega') reg msg =
	Error.fail reg ("\nGamma: [" ^ (stringOfTypedIdentList false false false gamma) ^ "]\nOmega: ["
		^ (stringOfTypedIdentList false false false omega) ^ "]\nj: {"
		^ (stringOfIdentList j) ^ "}\nOmega': ["
		^ (stringOfTypedIdentList false false false omega') ^ "]\n\n"
		^ msg)

fun typeFailureExp gamma omega typ exp reg msg =
	Error.fail reg ("\n\nGamma: [" ^ (stringOfTypedIdentList false false false gamma) ^ "]\nOmega: ["
		^ (stringOfTypedIdentList false false false omega) ^ "]\nExpr type: "
		^ (stringOfType false false false typ) ^ "\n\n"
		^ (ProofPrettyPrinter.stringOfExpression (exp, typ)) ^ "\n\n"
		^ msg)

(* Operations on environment *)
infix +! ++

(**
 * check that the given sequence represents a correct proof.
 *
 * @params gamma omega (j, omega') reg seq
 * @param gamma the constant proof checking environment.
 * @param omega the variable proof checking environment.
 * @param j the out type existential quantified variables.
 * @param omega' the out type.
 * @param reg the region.
 * @param seq the sequence to check.
 *
 * @return true if and only if seq is a correct proof.
 *)
fun checkSequence gamma omega (j, omega') reg seq =
	let val (iseq, (k, delta)) = seq
	in (* Checks whether the out type is equal to the sequence type annotation *)
		((envEquals delta omega') andalso (Utils.listEquals k j)
		 andalso (checkInSequence gamma omega (k, delta) reg iseq)) (* Checks the internal sequence *)
		orelse typeFailureSeq gamma omega (j, omega') seq reg "ProofChecker.checkSequence: out type and annotation mismatch"
	end

and checkInSequence gamma omega (j, omega') reg = fn
	(* Rule T.EMPTY *)
	  Empty (subst) =>
		(* We apply the substitution to build the instantiated type *)
		let val out_subst = substituteAllList subst omega'
			(* We restrict omega to omega' *)
			val delta = List.filter (fn (x,t) => ListAssoc.memberKey x omega') omega
		in
			((Utils.listEquals j (List.map #1 subst)) (* No alpha-equivalence ! Variables must be identical *)
			andalso (envEquals out_subst delta)) (* The instantiated type must be equal to omega *)
			orelse typeFailureInSeq gamma omega (j, omega') reg
					"ProofChecker.checkSequence: T.EMPTY"
		end

	(* Rule T.CST *)
	| Cst (id, typ, exp, seq) =>
		((checkExp gamma omega typ reg exp)
		andalso (checkSequence (gamma +! (id, typ)) omega (j, omega') reg seq))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.CST"

	(* Rule T.VAR *)
	| Var (id, typ, exp, seq) =>
		((checkExp gamma omega typ reg exp)
		andalso (checkSequence gamma (omega +! (id, typ)) (j, omega') reg seq))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.VAR "

	(* Rule T.BLOCK *)
	| Comm (Block (seq1, (k, x_tau_list)), seq2) =>
		(case split omega (dom x_tau_list) of (* We try to split omega according to the out list *)
		  NONE => false (* Failure *)
		| SOME (x_sigma_list, omega'') =>
		(checkSequence gamma x_sigma_list (k, x_tau_list) reg seq1)
		andalso (List.all (fn id => not (Utils.listMember id ((fv_formulas (img gamma)) (* k not-in FV(G,W) *)
																@ (fv_formulas (img omega))))
									andalso ((Utils.listMember id j) (* k in j or k not-in FV(W')*)
									orelse (not (Utils.listMember id (fv_formulas (img omega'))))))
						k)
		andalso (checkSequence gamma (omega'' ++ x_tau_list) (j, omega') reg seq2))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.BLOCK"

	(* Rule T.INC *)
	| Comm (Inc (id, typ), seq) =>
		(case getVarType id omega of
		  SOME (TNat (t)) =>
			(typ = TNat (t)) andalso
			checkSequence gamma (omega +! (id, TNat (s(t)))) (j, omega') reg seq
		| _ => false)
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.INC"

	(* Rule T.DEC *)
	| Comm (Dec (id, typ), seq) =>
		(case getVarType id omega of
		  SOME (TNat (t)) =>
			(typ = TNat (t)) andalso
			checkSequence gamma (omega +! (id, TNat (p(t)))) (j, omega') reg seq
		| _ => false)
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.DEC"

	(* Rule T.ASSIGN *)
	| Comm (Affect (id, typ, exp), seq) =>
		(case getVarType id omega of
		  NONE => false
		| _ => (checkExp gamma omega typ reg exp)
				andalso (checkSequence gamma (omega +! (id, typ)) (j, omega') reg seq))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.ASSIGN"

	(* Rule T.FOR *)
	| Comm (For (id, lid, exp, typ, (seq1, (k, sigma_i))), seq2) =>
		(case (typ, split omega (dom sigma_i)) of (* We try to split omega according to the out list *)
		  (TNat (n), SOME (sigma_0, omega'')) =>
			(envEquals sigma_0 (substituteAll lid FZero sigma_i)) (* sigma[0] must be obtained *)
			andalso (checkExp gamma omega typ reg exp)
			andalso (List.length k = 0) (* k must be empty: no existential quantification in recursion *)
			andalso (checkSequence (gamma +! (id, TNat (FId (lid)))) sigma_i (* sigma[i] |- sigma[s(i)] *)
						([], (substituteAll lid (s(FId (lid))) sigma_i)) reg seq1)
			andalso (checkSequence gamma (omega'' ++ (substituteAll lid n sigma_i)) (* sigma[n] *)
						(j, omega') reg seq2)
			andalso not (Utils.listMember lid (fv_formulas (img gamma))) (* i not_in FV(G) *)
		| _ => false)
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.FOR"

	(* Rule T.CALL *)
	| Comm (ProcCall (exp, typ, exp_typ_list, subst, id_typ_list), seq) =>
		((checkExp gamma omega typ reg exp)
		andalso List.all (fn (e, t) => checkExp gamma omega t reg e) exp_typ_list
		(* We split both omega and omega' to check type equalities *)
		andalso
			(case (typ, split omega (dom id_typ_list)) of
			  (TProc (i, tau_list, k, sigma_list), SOME (_, omega'')) =>
				(Utils.listEquals i (List.map #1 subst)) (* No alpha-equivalence ! Variables must be identical *)
				andalso
				(let val tau_subst = List.map (substituteList subst) tau_list (* tau[n/i] *)
					 val typ_list = List.map #2 exp_typ_list
				 in (* exp_list types = tau[n/i] *)
				 	(ListPair.all (fn (t1, t2) => alphaEqual(t1, t2)) (tau_subst, typ_list))
				 	orelse typeFailureInSeq (List.map (fn t => (("_", 0), t)) tau_subst)
				 			(List.map (fn t => (("_", 0), t)) typ_list)
				 			(j, omega') reg "ProofChecker.checkSequence: T.CALL !!!"
				 end)
				andalso (List.length id_typ_list = List.length sigma_list)
				andalso (ListPair.all (fn (t1, t2) => (substituteList subst t1) = t2)
							(sigma_list, List.map #2 id_typ_list)) (* id_list types = sigma[n/i] *)
				andalso (checkSequence gamma (omega'' ++ id_typ_list) (j, omega') reg seq)
				andalso (List.all (fn id => not (Utils.listMember id ((fv_formulas (img gamma)) (* k not-in FV(G,W) *)
																		@ (fv_formulas (img omega))))
											andalso ((Utils.listMember id j) (* k in j or k not-in FV(W') *)
											orelse (not (Utils.listMember id (fv_formulas (img omega'))))))
								k)
			| _ => false))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.CALL"

	(* Rule T.SUBST-II *)
	| SSubst (seq, delta_i, lid, exp, typ) =>
		((checkExp gamma omega typ reg exp)
		andalso (case typ of
				  TEqual (n, m) =>
				  (checkSequence gamma omega (j, (substituteAll lid n delta_i)) reg seq) (* delta[n/i] *)
				  andalso (envEquals (substituteAll lid m delta_i) omega') (* delta[m/i] *)
				| _ => false))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.SUBST-II"

	(* Rule T.LABEL *)
	| Comm (Label (id, typ, (seq1, (k, id_typ_list))), seq2) =>
		(case (typ, split omega (dom id_typ_list)) of
		  (TProc (k', sigma, [], [TBot]), SOME (x_tau, omega'')) =>
		  (Utils.listEquals sigma (img id_typ_list))
		  andalso (Utils.listEquals k k') (* No alpha-equivalence ! Variables must be identical *)
		  andalso (checkSequence (gamma +! (id, typ)) x_tau (k, id_typ_list) reg seq1)
		  andalso (checkSequence gamma (omega'' ++ id_typ_list) (j, omega') reg seq2)
		| _ => false)
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.LABEL"

	(* Rule T.JUMP *)
	| Comm (Jump (exp, typ, exp_typ_list, subst, id_typ_list), seq) =>
		((checkExp gamma omega typ reg exp)
		andalso List.all (fn (e, t) => checkExp gamma omega t reg e) exp_typ_list
		andalso (case (typ, split omega (dom id_typ_list)) of
				  (TProc (k, sigma, [], [TBot]), SOME (_, omega'')) =>
				  (Utils.listEquals k (List.map #1 subst)) (* No alpha-equivalence ! Variables must be identical *)
				  andalso
				  	(let val sigma_subst = List.map (substituteList subst) sigma (* sigma[n/i] *)
				  		val typ_list = List.map #2 exp_typ_list
				  	in (* We check whether exp_list types = sigma[n/i] *)
				  		(ListPair.all (fn (t1, t2) => alphaEqual(t1, t2)) (sigma_subst, typ_list))
				  		orelse typeFailureInSeq (List.map (fn t => (("_", 0), t)) sigma_subst)
				  					(List.map (fn t => (("_", 0), t)) typ_list)
				  					(j, omega') reg "ProofChecker.checkSequence: T.JUMP !!!"
				  	end)
				  andalso (checkSequence gamma (omega'' ++ id_typ_list) (j, omega') reg seq)
				| _ => false))
		orelse typeFailureInSeq gamma omega (j, omega') reg
				"ProofChecker.checkSequence: T.JUMP"

	| Comm (CommRegion (c, reg), seq) => checkInSequence gamma omega (j, omega') reg (Comm (c, seq))
	| SeqRegion (seq, reg) => checkInSequence gamma omega (j, omega') reg seq

(**
 * check that the given expression represents a correct proof.
 *
 * @params gamma omega typ reg (expression, tau)
 * @param gamma the constant proof checking environment.
 * @param omega the variable proof checking environment.
 * @param typ the requiered type.
 * @param reg the region.
 * @param expression the expression to check.
 * @param tau the given type.
 *
 * @return true if and only if expression is a correct proof of formula tau.
 *)
and checkExp gamma omega typ reg (expression, tau) =
	(tau = typ) andalso
	case expression of
	(* Rule T.ENV *)
	  Id (id) =>
	  (case getVarType id (gamma ++ omega) of
	    NONE => false
	  | SOME t => t = typ)
	  orelse typeFailureExp gamma omega typ expression reg
	  			"ProofChecker.checkExp: T.ENV"

	| Val (v) => checkValue gamma omega typ reg v

	(* Rule T.SUBST-I *)
	| ESubst (exp1, typ1, lid, exp2, typ2) =>
	((checkExp gamma omega typ2 reg exp2)
	andalso (case typ2 of
			  TEqual (n, m) =>
			  (checkExp gamma omega (substitute lid n typ1) reg exp1)
			  andalso (typ = (substitute lid m typ1))
			| _ => false))
	orelse typeFailureExp gamma omega typ expression reg
			"ProofChecker.checkExp: T.SUBST-I"

	(* Lemma *)
	| Lemma (typlist, typ') =>
	(let val hyp = img (gamma ++ omega)
	in
		(typ = typ') andalso
		(List.all (fn x => Utils.listMember x hyp) typlist)
	end)
	orelse typeFailureExp gamma omega typ expression reg
			"ProofChecker.checkExp: Lemma"

	| ExpRegion (exp, reg) => checkExp gamma omega typ reg (exp, tau)

(**
 * check that the given value represents a correct proof.
 *
 * @params v reg
 * @param v the value.
 *
 * @return true if and only if p is a correct proof.
 *)
and checkValue gamma omega typ reg (value, tau) =
	(tau = typ) andalso
	case value of
	(* Rule T.STAR *)
	  Star =>
	(case typ of TEqual (n, m) => true
				| _ => false)
	orelse typeFailureExp gamma omega typ (Val (value, tau)) reg
			"ProofChecker.checkValue: T.STAR"

	(* Rule T.NUM *)
	| Int (n) =>
	(typ = (TNat (mkNat n)))
	orelse typeFailureExp gamma omega typ (Val (value, tau)) reg
			"ProofChecker.checkValue: T.NUM"

	(* Rule T.PROC *)
	| Proc (i, in_list, j, out_list, (seq, (j', typid_list))) =>
	(let val sigmalist = img in_list
		 val taulist = img out_list
	 in
		 (typ = (TProc (i, sigmalist, j, taulist)))
		 andalso (Utils.listEquals j j') (* No alpha-equivalence ! Variables must be identical *)
		 andalso (envEquals out_list typid_list)
		 andalso (checkSequence (in_list ++ gamma) (List.map (fn (x,t) => (x, top)) typid_list)
					(j', typid_list) reg seq)
		 andalso (List.all (fn id => not (Utils.listMember id (fv_formulas (img gamma)))) i) (* i not-in FV(G) *)
	 end)
	orelse typeFailureExp gamma omega typ (Val (value, tau)) reg
				  "ProofChecker.checkValue: T.PROC"

	| ValRegion (v, reg) => checkValue gamma omega typ reg (v, tau)

(**
 * check that the given program represents a correct proof.
 *
 * @params p
 * @param p the program.
 *
 * @return true if and only if p is a correct proof.
 *)
fun checkClosedProg (p : program) =
	case p of
	  (SeqRegion (seq, reg), seqType) => checkSequence empty empty seqType reg p
	| _ => raise Fail "ProofChecker.checkClosedProg: not a SeqRegion"

end

