(**
 * The First-Order LoopW Proof Checker proof environment abstract type.
 *
 * The proof derivation pretty-printer proof environment abstract type
 * provides functions to deal with typing environment.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
signature PROOF_ENVIRONMENT = sig

(**
 * the typing environment.
 *)
type env = (ProofAst.ident * ProofAst.formula) list

(**
 * the empty environment.
 *)
val empty
	: env

(**
 * returns the type of an identifier in an environment.
 *
 * @params x env
 * @param x the variable.
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return SOME Ti if there exists i such that x = xi, or NONE.
 *)
val getVarType
	: ProofAst.ident -> env -> ProofAst.formula option

(**
 * removes an identifier typing in an environment.
 *
 * @params x env
 * @param x the identifier.
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return env without x:T for any T.
 *)
val -!
	: (env * ProofAst.ident) -> env

(**
 * adds an identifier typing in an environment. Ensures that x occurs only once
 * in the environment.
 *
 * @params env (x, t)
 * @param env the environment x1:T1, ..., xn:Tn.
 * @param x the identifier.
 * @param t the type of the identifier.
 *
 * @return x:t,(env without x:T for any T).
 *)
val +!
	: (env * (ProofAst.ident * ProofAst.formula)) -> env

(**
 * adds every typed identifiers of an environment at the end of
 * another environment. Ensures that each identifier occurs only once
 * in the resulting environment.
 *
 * @params env1 env2
 * @param env1 the first environment x1:T1, ..., xn:Tn.
 * @param env2 the second environment y1:U1, ..., ym:Um.
 *
 * @return x1:T1, ..., xn:Tn,(env2 without x1, ..., xn).
 *)
val ++
	: (env * env) -> env

(**
 * check that an identifier list is included in an environment,
 * returns the partition of it according to that list.
 *
 * @params env xl
 * @param env the environment x1:T1, ..., xn:Tn.
 * @param xl the identifier list y1, ..., ym.
 *
 * @return SOME((y1:U1, ..., ym:Um), (env without y1, ..., ym)) if
 * y1, ..., ym belongs to env, else NONE.
 *)
val split
	: env -> ProofAst.ident list -> (env * env) option

(**
 * returns the domain of the environment.
 *
 * @params env
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return x1, ..., xn.
 *)
val dom
	: env -> ProofAst.ident list

(**
 * returns the image of the environment.
 *
 * @params env
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return T1, ..., Tn.
 *)
val img
	: env -> ProofAst.formula list

(**
 * test if the enviroment are equals.
 *
 * @params env1 env2
 * @param env1 the first environment.
 * @param env2 the second environment.
 *
 * @return true if the contents of env1 and env2 are equals (independantly
 * of the ordering).
 *)
val envEquals
	: env -> env -> bool

end

