(**
 * The First-Order LoopW Proof Checker proof environment abstract type.
 *
 * The proof derivation pretty-printer proof environment abstract type
 * provides functions to deal with typing environment.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ProofEnvironment :> PROOF_ENVIRONMENT = struct

open ProofAst

infix -! +! ++

(**
 * the typing environment.
 *)
type env = (ident * formula) list

(**
 * the empty environment.
 *)
val empty = []

(**
 * returns the type of an identifier in an environment.
 *
 * @params x env
 * @param x the identifier.
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return SOME Ti if there exists i such that x = xi, or NONE.
 *)
fun getVarType x env = ListAssoc.assoc x env

(**
 * removes an identifier typing in an environment.
 *
 * @params x env
 * @param x the identifier.
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return env without x:T for any T.
 *)
fun env -! x = ListAssoc.removeKey x env

(**
 * adds an identifier typing in an environment. Ensures that x occurs only once
 * in the environment.
 *
 * @params (x, t) env
 * @param x the identifier.
 * @param t the type of the identifier.
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return x:t, (env without x:T for any T).
 *)
fun env +! (x, t) = ListAssoc.add (x, t) env

(**
 * adds every typed identifiers of an environment at the end of
 * another environment. Ensures that each identifier occurs only once
 * in the resulting environment.
 *
 * @params env1 env2
 * @param env1 the first environment x1:T1, ..., xn:Tn.
 * @param env2 the second environment y1:U1, ..., ym:Um.
 *
 * @return x1:T1, ..., xn:Tn, (env2 without x1, ..., xn).
 *)
fun env1 ++ env2 = ListAssoc.addAll env1 env2

(**
 * check that an identifier list is included in an environment,
 * returns the partition of it according to that list.
 *
 * @params env xl
 * @param env the environment x1:T1, ..., xn:Tn.
 * @param xl the identifier list y1, ..., ym.
 *
 * @return SOME((y1:U1, ..., ym:Um), (env without y1, ..., ym)) if
 * y1, ..., ym belongs to env, else NONE.
 *)
fun split env xl =
    if ListAssoc.existsAllKey xl env
    then SOME (ListAssoc.partitionWithKeyList xl env)
    else NONE

(**
 * returns the domain of the environment.
 *
 * @params env
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return x1, ..., xn.
 *)
fun dom env = ListAssoc.getKeyList env

(**
 * returns the image of the environment.
 *
 * @params env
 * @param env the environment x1:T1, ..., xn:Tn.
 *
 * @return T1, ..., Tn.
 *)
fun img env = ListAssoc.getValueList env

(**
 * test if the enviroment are equals.
 *
 * @params env1 env2
 * @param env1 the first environment.
 * @param env2 the second environment.
 *
 * @return true if the contents of env1 and env2 are equals (independantly
 * of the ordering).
 *)
fun envEquals env1 env2 = Utils.listEquals env1 env2

end

