(**
 * The First-Order LoopW Proof Checker proof syntax pretty-printer.
 *
 * The proof pretty-printer converts proof abstract syntax into string.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ProofPrettyPrinter :> PROOF_PRETTY_PRINTER = struct

open PrettyPrinterKeywords
open BasePrettyPrinter
open ProofAst

(**
 * converts a substitution into a string.
 *
 * @params ott subst
 * @param ott the ott flag indicates to display only subst terms individualy between [].
 * @param subst the subtitution as (id, term) list.
 *
 * @return the string representation of subst.
 *)
fun stringOfSubst ott subst =
	(if subst = [] then ""
	 else
	 	let val subst = List.rev subst in
	 		if ott
	 		then String.concat (List.map (fn (i,t) => "{" ^ (stringOfTerm ott t) ^ "}") subst)
	 		else ("[" ^ (String.concatWith ", "
	 					(List.map (fn (i,t) => (stringOfTerm ott t) ^ "/" ^ (stringOfIdent i)) subst))
	 			^ "]")
	 	end)

(**
 * converts a substitution with an environment to a existential quantification introduction string list for ott.
 *
 * @params tab subst (j, omega)
 * @param tab the tabulation level.
 * @param subst the subtitution as (id, term) list.
 * @param j the existential quantification identifier list.
 * @param omega the formula
 *
 * @return the string list representation of the existential introduction list.
 *)
fun stringOfExistentialIntro tab subst (j, omega) = (* j might be useless *)
	let val rec aux = fn om => fn
		  [] => []
		| (x, t)::q =>
		(aux (Terms.substituteAll x t om) q)
		@ [(stringOfTab tab) ^ "[ " ^ (stringOfTerm true t) ^ " (- "
		   ^ (stringOfQuantifiedTypedIdentList false false true (x::(List.map #1 q)) om)
		   ^ " ]"]
	in
		aux omega subst
	end

(**
 * converts a substitution with a type list to a negation existential quantification elimination string for ott.
 *
 * @params subst (j, omega)
 * @param subst the subtitution as (id, term) list.
 * @param j the existential quantification identifier list.
 * @param omega the formula
 *
 * @return the string representation of the negation existential quantification elimination.
 *)
fun stringOfNegExistElim subst (j, omega) = (* j might be useless *)
	let val rec aux = fn om => fn
		  [] => ""
		| (x, t)::q =>
		" <: {" ^ (stringOfIdent x) ^ "/"
		^ (stringOfQuantifiedTypeList false false true (List.map #1 q) om)
		^ "}{" ^ (stringOfTerm true t) ^ "}"
		^ (aux (List.map (Terms.substitute x t) om) q)
	in
		aux omega subst
	end

(**
 * converts command into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott c
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param c the command.
 *
 * @return the string representation of the command.
 *)
fun stringOfCommand_internal tab form tex typ ott = fn
    Block (seq, (j, idlist)) =>
    ((stringOfTab tab) ^ "{") ::
    ((stringOfSequence_internal (tab + 1) form tex typ ott seq)
	@ [(stringOfTab tab) ^ "}"
		^ (if typ then stringOfQuantifiedTypedIdentList form tex ott j idlist
				  else stringOfIdentList (List.map #1 idlist))
		^ ";" ^ (if ott then String.concat (List.map (fn id => "?" ^ (stringOfIdent id) ^ ".") j)
				 else "")])

  | For (id, lid, exp, etyp, (seq, (j, idlist))) =>
    ((stringOfTab tab) ^ (kwd tex "for") ^ " " ^ (stringOfIdent id)
     ^ (if typ then " : " ^ (stringOfType form tex ott (TNat (FId (lid)))) else "")
     ^ " := 0 " ^ (kwd tex "until") ^ " "
     ^ (List.hd (stringOfExpression_internal tab form tex typ ott exp))
     ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
     ^ " {")
    ::
	((stringOfSequence_internal (tab + 1) form tex typ ott seq)
	 @ [(stringOfTab tab) ^ "}"
		^ (if typ then stringOfQuantifiedTypedIdentList form tex ott j idlist
				  else stringOfIdentList (List.map #1 idlist))
		^ ";"])

  | Affect (id, etyp, exp) =>
    let val l = stringOfExpression_internal tab form tex typ ott exp in
		[(stringOfTab tab) ^ (stringOfIdent id)
		^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
		^ " := " ^ (List.hd l) ^ (if List.length l = 1 then ";" else "")]
		@ (if List.length l > 1 then List.tl l else [])
    end

  | Inc (id, etyp) =>
    [(stringOfTab tab) ^ (kwd tex "inc") ^ "(" ^ (stringOfIdent id)
     ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
     ^ ");"]

  | Dec (id, etyp) =>
    [(stringOfTab tab) ^ (kwd tex "dec") ^ "(" ^ (stringOfIdent id)
     ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
     ^ ");"]

  | ProcCall (exp, etyp, exp_typ_list, subst, id_list) =>
    [(stringOfTab tab)
     ^ (String.concatWith " | " (stringOfExpression_internal tab form tex typ ott exp))
     ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
     ^ (if typ then stringOfSubst ott subst else "")
     ^ "(" ^ (String.concat (stringOfTypedExpressionList tab form tex typ ott exp_typ_list))
     ^ "; "
     ^ (if (typ andalso (not ott))
     	then (stringOfTypedIdentList form tex ott id_list)
     	else (stringOfIdentList (List.map #1 id_list)))
     ^ ");" ^ (if ott then
     			(case etyp of
     				TProc(i, sigma, j, tau) => String.concat (List.map (fn id => "?" ^ (stringOfIdent id) ^ ".") j)
     			  | _ => raise Fail "ProofPrettyPrinter.sml: not a proc")
     			else "")]

  | Label (id, etyp, (seq, (j, idlist))) =>
    ((stringOfTab tab) ^ (stringOfIdent id)
     ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
     ^ " :{")
     ::
     ((stringOfSequence_internal (tab + 1) form tex typ ott seq)
	 @ [(stringOfTab tab) ^ "}"
		^ (if typ then stringOfQuantifiedTypedIdentList form tex ott j idlist
				  else stringOfIdentList (List.map #1 idlist))
		^ ";" ^ (if ott then String.concat (List.map (fn id => "?" ^ (stringOfIdent id) ^ ".") j)
				 else "")])

  | Jump (exp, etyp, exp_typ_list, subst, typid_list) =>
    [(stringOfTab tab) ^ (kwd tex "jump") ^ "("
     ^ (String.concatWith " | "  (stringOfExpression_internal tab form tex typ ott exp))
     ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott etyp) else "")
     ^ (if typ then if ott then stringOfNegExistElim subst
     		(case etyp of TProc(i, sigma, [], [TBot]) => (i, sigma)
     			| _ => raise Fail "ProofPrettyPrinter.sml: not a proc")
     		else stringOfSubst false subst
     	else "")
     ^ ", " ^ (String.concat (stringOfTypedExpressionList tab form tex typ ott exp_typ_list))
     ^ ")" ^ (if typ then stringOfQuantifiedTypedIdentList form tex ott [] typid_list
	      else stringOfIdentList (List.map #1 typid_list))
     ^ ";"]

  | CommRegion (c, _) => stringOfCommand_internal tab form tex typ ott c
 
(**
 * converts sequence into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott s
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param s the sequence.
 *
 * @return the string representation of the sequence.
 *)
and stringOfSequence_internal tab form tex typ ott (s, (j, omega)) =
	case s of
		Empty (subst) =>
		if typ
		then
			if ott then stringOfExistentialIntro tab subst (j, omega)
			else
			[(stringOfTab tab) ^
			 (stringOfSubst false (*ott*) subst) ^ " |> "
			 ^ (stringOfQuantifiedTypedIdentList form tex ott j omega)]
		else []

      | Comm (c, seq) =>
      	(stringOfCommand_internal tab form tex typ ott c)
      	@ (stringOfSequence_internal tab form tex typ ott seq)
      	@ (if (typ andalso (not ott))
      		then [(stringOfTab tab) ^ " |> " ^ (stringOfQuantifiedTypedIdentList form tex ott j omega)]
      		else [])

      | Cst (id, t, exp, seq) =>
      	let val l = stringOfExpression_internal tab form tex typ ott exp in
	    	[(stringOfTab tab) ^ (kwd tex "cst") ^ " " ^ (stringOfIdent id)
	    	^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott t) else "")
	    	^ " = " ^ (List.hd l) ^ (if List.length l = 1 then ";" else "")]
	     	@ (if List.length l > 1 then List.tl l else [])
	     	@ (stringOfSequence_internal tab form tex typ ott seq)
	     	@ (if (typ andalso (not ott))
      			then [(stringOfTab tab) ^ " |> " ^ (stringOfQuantifiedTypedIdentList form tex ott j omega)]
      			else [])
	    end

      | Var (id, t, exp, seq) =>
      	let val l = stringOfExpression_internal tab form tex typ ott exp in
	    	[(stringOfTab tab) ^ (kwd tex "var") ^ " " ^ (stringOfIdent id)
	    	^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott t) else "")
	    	^ " := " ^ (List.hd l) ^ (if List.length l = 1 then ";" else "")]
			@ (if List.length l > 1 then List.tl l else [])
			@ (stringOfSequence_internal tab form tex typ ott seq)
	     	@ (if (typ andalso (not ott))
      			then [(stringOfTab tab) ^ " |> " ^ (stringOfQuantifiedTypedIdentList form tex ott j omega)]
      			else [])
		end

      | SSubst (seq, omega_i, id, exp, t) =>
      	let val el = stringOfExpression_internal tab form tex typ ott exp
      		val seq_s = stringOfSequence_internal tab form tex typ ott seq
      	in
      		if not ott
      		then
	    		seq_s
	    		@ (if typ
	    			then [(stringOfTab tab) ^ "::>> " ^ (stringOfTypedIdentList form tex ott omega_i)
	    				^ "[" ^ (stringOfIdent id) ^ " <- "
	    				^ (String.concatWith " | " el) ^ " : " ^ (stringOfType form tex ott t)
	    				^ "] |> " ^ (stringOfQuantifiedTypedIdentList form tex ott j omega)]
	    			else [])
	    	else
	    		seq_s
	    		@ [(stringOfTab tab) ^ "( :> "
	    			^ (stringOfQuantifiedTypedIdentList form tex ott [id] omega_i)
	    			^ "[" ^ (case t of TEqual (n, m) => stringOfType form tex ott (TEqual (m, n))
	    								| x => stringOfType form tex ott x) ^ "])"]
	    end
	    
	  | SeqRegion (s, _) => stringOfSequence_internal tab form tex typ ott (s, (j, omega))

(**
 * converts expression into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott e
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param e the expression.
 *
 * @return the string representation of the expression.
 *)
and stringOfExpression_internal tab form tex typ ott (e, tau) =
	case e of
		Id (id) => [(stringOfIdent id)
			^ (if (typ andalso (not ott))
				then " : " ^ (stringOfType form tex ott tau) else "")]

      | Val (v) => stringOfValue_internal tab form tex typ ott v

      | ESubst (e, t_i, id, exp, neqm) =>
      	let val l = stringOfExpression_internal tab form tex typ ott exp
      		val e_s = stringOfExpression_internal tab form tex typ ott e
      		val subst_s = ":> "
	    			^ "{" ^ (stringOfIdent id) ^ "/" ^ (stringOfType form tex ott t_i)
	    			^ (if not ott then (String.concatWith " | " l) else "") ^ "}"
	    			^ "[" (*^ (case neqm of TEqual (n, m) => stringOfType form tex ott (TEqual (m, n))
	    								| x => stringOfType form tex ott x)*)
	    			^ (stringOfType form tex ott neqm)
	    			^ "]"
	    			^ (if not ott then " : " ^ (stringOfType form tex ott tau) else "")
      	in
      		if List.length e_s = 1
      		then [(List.hd e_s) ^ (if typ then " " ^ subst_s else "")]
      		else if typ then e_s @ [(stringOfTab tab) ^ subst_s] else e_s
	    end

      | Lemma (tl, t) =>
      	if (typ andalso (not ott))
      	then [(String.concatWith ", " (List.map (stringOfType form tex ott) tl))
      			^ " " ^ (kwdVdash tex) ^ " " ^ (stringOfType form tex ott t)]
      	else []
      	
      | ExpRegion (e, _) => stringOfExpression_internal tab form tex typ ott (e, tau)

(**
 * converts a pair expression * type into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott e
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param e the expression.
 *
 * @return the string representation of the expression.
 *)
and stringOfTypedExpression_internal tab form tex typ ott (e,t) =
	[(String.concatWith " | " (stringOfExpression_internal tab form tex typ ott e))
	 ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott t) else "")]

(**
 * converts a list of expression into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott exp_list
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param exp_list the expression list.
 *
 * @return the string representation of the expression list.
 *)
and stringOfExpressionList tab form tex typ ott exp_list =
    Utils.listConcatWith [", "] (List.map (stringOfExpression_internal tab form tex typ ott) exp_list)

(**
 * converts a list of typed expression into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott exp_list
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param exp_list the typed expression list.
 *
 * @return the string representation of the typed expression list.
 *)
and stringOfTypedExpressionList tab form tex typ ott exp_typ_list =
    Utils.listConcatWith [", "] (List.map (stringOfTypedExpression_internal tab form tex typ ott) exp_typ_list)

(**
 * converts value into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ ott v
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param ott flag indicating whether the output must respect ott/twelf specific syntax.
 * @param v the value.
 *
 * @return the string representation of the value.
 *)
and stringOfValue_internal tab form tex typ ott (v, tau) =
    case v of
		Star => ["*" ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott tau) else "")]

      | Int (n) => [(Int.toString n) ^ (if (typ andalso (not ott))
      									then " : " ^ (stringOfType form tex ott tau) else "")]

      | Proc (i, in_list, j, out_list, (seq, (jj, idlist))) =>
      	(if not ott then
      		((kwd tex "proc") ^ "(" ^ (stringOfParamList form tex typ i (kwd tex "in") in_list)
      		 ^ "; " ^ (stringOfParamList form tex typ j (kwd tex "out") out_list)
      		 ^ ")" ^ (if (typ andalso (not ott)) then " : " ^ (stringOfType form tex ott tau) else "")
      		 ^ " {")
      	else
      		((kwd tex "proc") ^ " "
      		 ^ (stringOfQuantifiedTypedIdentList form tex ott i in_list)
      		 ^ " " ^ (kwd tex "out") ^ " "
      		 ^ (stringOfQuantifiedTypedIdentList form tex ott j out_list)
      		 ^ " {"))
      	:: ((stringOfSequence_internal (tab + 1) form tex typ ott seq)
      	@ [(stringOfTab tab) ^ "};"])
      	
      | ValRegion (v, _) => stringOfValue_internal tab form tex typ ott (v, tau)

(**
 * converts program into string accordingly to the formula/tex demand.
 *
 * @params form tex p
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param p the program.
 *
 * @return the string representation of p.
 *)
fun stringOfProgram form tex p =
	stringOfSequence_internal 0 form tex true false p

(**
 * converts program into string, generating ott-oriented type and syntax.
 *
 * @params p
 * @param p the program.
 *
 * @return the string representation of p.
 *)
fun stringOfProgram_Ott p =
	stringOfSequence_internal 0 false false true true p

(**
 * converts program into string accordingly to the formula/tex demand,
 * without any type information.
 *
 * @params form tex p
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param p the program.
 *
 * @return the string representation of p.
 *)
fun stringOfProgram_Untyped form tex p =
	stringOfSequence_internal 0 form tex false false p

(**
 * converts expression into string.
 *
 * @params e
 * @param e the expression.
 *
 * @return the string representation of e.
 *)
fun stringOfExpression e =
	String.concatWith " | " (stringOfExpression_internal 0 false false true false e)

end

