(**
 * The First-Order LoopW Proof Checker formula/term substitution and alpha-equivalence.
 *
 * Provides functions in order to calculate free variables, perform term substitution
 * into formulas/terms, and identify formulas up to alpha-equivalence.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
signature TERMS = sig

(**
 * gives a new fresh variable with the given identifier.
 *
 * @params i
 * @param i the identifier.
 *
 * @return a fresh identifier.
 *)
val alpha
	: ProofAst.ident -> ProofAst.ident

(**
 * calculates the free variables of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the list of free identifiers.
 *)
val fv_term
	: ProofAst.term -> ProofAst.ident list

(**
 * calculates the free variables of a formula list.
 *
 * @params fl
 * @param fl the formula list.
 *
 * @return the list of free identifiers.
 *)
val fv_formulas
	: ProofAst.formula list -> ProofAst.ident list

(**
 * performs the substitution of a term into a formula.
 *
 * @params x t f
 * @param x the identifier to be substituted.
 * @param t the term to substitute to x.
 * @param f the formula in which to perform the substitution.
 *
 * @return the new formula f[t/x].
 *)
val substitute
	: ProofAst.ident -> ProofAst.term -> ProofAst.formula -> ProofAst.formula

(**
 * performs a list of substitutions into a formula.
 *
 * @params xtlist f
 * @param xtlist the list of substitutions to perform.
 * @param f the formula in which to perform the substitution.
 *
 * @return the new formula f[t1/x1, ..., tn/xn].
 *)
val substituteList
	: (ProofAst.ident * ProofAst.term) list
		-> ProofAst.formula -> ProofAst.formula

(**
 * performs a substitution into a named formula list.
 *
 * @params x t yflist
 * @param x the identifier to be substituted.
 * @param t the term to substitute to x.
 * @param yflist the named formula list in which to perform the substitution.
 *
 * @return the new formula [y,f][t/x].
 *)
val substituteAll
	: ProofAst.ident -> ProofAst.term -> (ProofAst.ident * ProofAst.formula) list
		-> (ProofAst.ident * ProofAst.formula) list

(**
 * performs a list of substitutions into a named formula list.
 *
 * @params xtlist yflist
 * @param xtlist the list of substitutions to perform.
 * @param yflist the named formula list in which to perform the substitution.
 *
 * @return the new formula [y,f][t1/x1, ..., tn/xn].
 *)
val substituteAllList
	: (ProofAst.ident * ProofAst.term) list
		-> (ProofAst.ident * ProofAst.formula) list
		-> (ProofAst.ident * ProofAst.formula) list

(**
 * converts the proc types in order to have the same bound variables.
 *
 * @params (f1, f2)
 * @param f1 the first formula proc({i} sigma; out {j} tau).
 * @param f2 the second formula proc({i'} sigma'; out {j'} tau').
 *
 * @return (f1', f2') where f1 = proc({i''} sigma[i''/i]; out {j''} tau[j''/j]),
 * and f1 = proc({i''} sigma'[i''/i']; out {j''} tau'[j''/j']).
 *)
val alphaConvProc
	: (ProofAst.ident list * ProofAst.formula list * ProofAst.ident list * ProofAst.formula list)
	* (ProofAst.ident list * ProofAst.formula list * ProofAst.ident list * ProofAst.formula list)
	-> (ProofAst.ident list * ProofAst.formula list * ProofAst.ident list * ProofAst.formula list)
	* (ProofAst.ident list * ProofAst.formula list * ProofAst.ident list * ProofAst.formula list)

(**
 * check the equality of a pair of formulas up to alpha-equivalence.
 *
 * @params (f1, f2)
 * @param f1 the first formula to compare.
 * @param f2 the second formula to compare.
 *
 * @return true if and only if f1 = f2 up to alpha-equivalence.
 *)
val alphaEqual
	: ProofAst.formula * ProofAst.formula -> bool

end

