(**
 * Utility functions that deals with lists.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure Utils :> UTILS = struct

(**
 * check if the given element belongs to the list.
 *
 * @params e l
 * @param e the element.
 * @param l the list.
 *
 * @return true if e belongs to l, else false.
 *)
fun listMember x l =
    List.exists (fn y => y = x) l

(**
 * merges the two list without duplicating identical elements.
 *
 * @params l1 l2
 * @param l1 the first list.
 * @param l2 the second list.
 *
 * @return l1 followed by (l2 - l1).
 *)
fun listFusion l1 l2 =
    case l1 of
		[] => l2
      | e::q => e::(List.filter (fn x => x <> e) l2)

(**
 * merges the two list without duplicating identical elements.
 *
 * @params l1 l2
 * @param l1 the first list.
 * @param l2 the second list.
 *
 * @return l1 followed by (l2 - l1).
 *)
fun listFusionPair (l1, l2) = listFusion l1 l2

(**
 * check if the given element belongs to the list, and removes it.
 *
 * @params e l
 * @param e the element.
 * @param l the list.
 *
 * @return ( (true if e belongs to l, else false) , (l without e) ).
 *)
fun listMemberRemove x l =
	(case l of
		[] => (false, [])
	  | (y::q) =>
	    if x = y then (true, q) else
	    (case listMemberRemove x q of
	    	(false, l2) => (false, (y::l2))
	      | (true, l2) => (true, (y::l2))))

(**
 * check if the two lists contains the same elements (not considering their
 * position in the list).
 *
 * @params l1 l2
 * @param l1 the first list.
 * @param l2 the second list.
 *
 * @return true if and only if l1 and l2 contains the same elements.
 *)
fun listEquals l1 l2 =
    (case l1 of
	 	[] => l2 = []
      | (t::q) =>
	 (case listMemberRemove t l2 of
	      (false, _) => false
	    | (true, new_l2) => listEquals q new_l2))

(**
 * flattens a list of list, inserting a separator between each inner lists.
 *
 * @params sep l
 * @param sep the separator (itself a list).
 * @param l the list list [l1, l2, ..., ln].
 *
 * @return l1 @ sep @ l2 @ ... @ ln.
 *)
fun listConcatWith lsep = fn
    [] => []
  | [l] => l
  | l::q => l @ lsep @ (listConcatWith lsep q)

end

