(**
 * The First-Order LoopW Type Inference base pretty-printer.
 *
 * The base pretty-printer converts terms, formulas, and ident abstract syntax into string.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure BasePrettyPrinter :> BASE_PRETTY_PRINTER = struct

open PrettyPrinterKeywords
open ProofAst

(**
 * builds a string with spaces for tabular display.
 *
 * @params n
 * @param n the number of tabs.
 *
 * @return the string tab.
 *)
fun stringOfTab n =
	if n = 0
	then ""
	else "    " ^ (stringOfTab (n - 1))

(**
 * converts identifiers into string.
 *
 * @params i
 * @param i the identifier.
 *
 * @return the string representation of i.
 *)
fun stringOfIdent (x,n) =
	x ^ (if n > 0 then "_" ^ (Int.toString n) else "")

(**
 * converts identifier lists into string.
 *
 * @params il
 * @param il the identifier list.
 *
 * @return the string representation of il.
 *)
fun stringOfIdentList idl =
	String.concatWith ", " (List.map stringOfIdent idl)

(**
 * converts a term into an integer if it is an unary natural number into an integer, returning NONE if not.
 *
 * @params t
 * @param t the term.
 *
 * @return an integer Option.
 *)
fun getNat t =
	case t of
		FZero => SOME (0)
      | FFun (("s", _), [t]) => (case getNat t of NONE => NONE | SOME n => SOME (n + 1))
      | _ => NONE

(**
 * converts terms abstract syntax into string.
 *
 * @params t
 * @param t the term.
 *
 * @return the string representation of t.
 *)
fun stringOfTerm ott t =
	case t of
		FId(id) => stringOfIdent id
      | FZero => "0"
      | FFun(f,tl) =>
      	(case getNat (FFun(f,tl)) of
      		SOME n => Int.toString n
      	  | _ =>
      	  	(case (f, tl) of
      	  		(("+", _), t1::t2::[]) =>
      	  		if ott then "add(" ^ (stringOfTerm ott t1) ^ ", " ^ (stringOfTerm ott t2) ^ ")"
      	  		else "(" ^ (stringOfTerm ott t1) ^ " + " ^ (stringOfTerm ott t2) ^ ")"
      	  	  | (("-", _), t1::t2::[]) =>
      	  		if ott then "sub(" ^ (stringOfTerm ott t1) ^ ", " ^ (stringOfTerm ott t2) ^ ")"
      	  		else "(" ^ (stringOfTerm ott t1) ^ " - " ^ (stringOfTerm ott t2) ^ ")"
      	  	  | (("*", _), t1::t2::[]) =>
      	  		if ott then "mult(" ^ (stringOfTerm ott t1) ^ ", " ^ (stringOfTerm ott t2) ^ ")"
      	  		else "(" ^ (stringOfTerm ott t1) ^ " * " ^ (stringOfTerm ott t2) ^ ")"
      	  	  | (("s", _), t::[]) =>
      	  		if ott then "succ(" ^ (stringOfTerm ott t) ^ ")"
      	  		else "s(" ^ (stringOfTerm ott t) ^ ")"
      	  	  | _ => (stringOfIdent f) ^ "(" ^ (String.concatWith ", " (List.map (stringOfTerm ott) tl)) ^ ")"))

(**
 * converts procedural type into string accordingly to the formula/tex demand.
 *
 * @params form tex ott idlist1 typlist1 idlist2 typlist2
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the ott flag.
 * @param idlist1 the identifier list of universal quantification.
 * @param typlist1 the formula list on the left hand side of the implication.
 * @param idlist2 the identifier list of existential quantification.
 * @param typlist2 the formula list on the right hand side of the implication.
 *
 * @return the string representation of the procedural type.
 *)
fun stringOfProcType form tex ott idlist1 typlist1 idlist2 typlist2 =
	let val idl1 = stringOfIdentList idlist1
		val idl2 = stringOfIdentList idlist2
		val tyl1 = String.concatWith
					(if form then kwdAnd tex else ", ")
					(List.map (stringOfType form tex ott) typlist1)
		val tyl2 = String.concatWith
					(if form then kwdAnd tex else ", ")
					(List.map (stringOfType form tex ott) typlist2)
	in
		if form
		then (* forall idl1.(tyl1 => exists idl2.tyl2) *)
			(if List.null idlist1 then "" else (kwdForall tex) ^ idl1 ^ ".")
			^ "(" ^ tyl1 ^ (kwdImplies tex)
			^ (if List.null idlist2 then "" else (kwdExists tex) ^ idl2 ^ ".")
			^ tyl2 ^ ")"
		else if ott
		then (* proc {id1}...{idn} ([tyl1] out {id1'}...{idm'} [tyl2]) *)
			(kwd tex "proc") ^ " " ^ (List.foldr (fn (id, s) => "{" ^ (stringOfIdent id) ^ "}" ^ s) "" idlist1)
			^ "([" ^ tyl1 ^ "] " ^ (kwd tex "out") ^ " "
			^ (List.foldr (fn (id, s) => "{" ^ (stringOfIdent id) ^ "}" ^ s) "" idlist2)
			^ "[" ^ tyl2 ^ "])"
		else (* proc({idl1} in tyl1; {idl2} out tyl2) *)
			(kwd tex "proc") ^ "(" ^ (if List.null idlist1 then "" else "{" ^ idl1 ^ "} ")
			^ (kwd tex "in") ^ " " ^ tyl1 ^ "; "
			^ (if List.null idlist2 then "" else "{" ^ idl2 ^ "} ")
			^ (kwd tex "out") ^ " " ^ tyl2 ^ ")"
	end

(**
 * converts negation type into string accordingly to the formula/tex demand.
 *
 * @params form tex ott idlist typlist
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the tex flag.
 * @param idlist the identifier list of universal quantification.
 * @param typlist the formula list.
 *
 * @return the string representation of the negation type.
 *)
and stringOfNegType form tex ott idlist typlist =
	let val idl = stringOfIdentList idlist
		val tyl = String.concatWith
					(if form then kwdAnd tex else ", ")
					(List.map (stringOfType form tex ott) typlist)
	in
		if form
		then (* forall idl.~tyl *)
			(if List.null idlist then "" else (kwdForall tex) ^ idl ^ ".")
			^ (kwdNeg tex)
			^ (if List.length typlist = 0 then "()"
				else if List.length typlist > 1 then "(" ^ tyl ^ ")" else tyl)
		else (* {idl}~tyl *)
			(if List.null idlist then "" else "{" ^ idl ^ "}")
			^ (kwdNeg tex)
			^ (if List.length typlist = 0 then "()"
				else if List.length typlist > 1 then "(" ^ tyl ^ ")" else tyl)
	end

(**
 * converts formulas/types abstract syntax into string accordingly to the formula/tex demand.
 *
 * @params form tex f
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the tex flag.
 * @param f the formula.
 *
 * @return the string representation of f.
 *)
and stringOfType form tex ott f =
	case f of
		TNat(t) => (kwd tex "nat") ^ "(" ^ (stringOfTerm ott t) ^ ")"
	  | TProc(i, t_in, [], [TBot]) => stringOfNegType form tex ott i t_in
      | TProc(i, t_in, j, t_out) => stringOfProcType form tex ott i t_in j t_out
      (*| TEqual(FZero, FZero) => kwdTop tex*)
      | TEqual(t1,t2) => "(" ^ (stringOfTerm ott t1) ^ " = " ^ (stringOfTerm ott t2) ^ ")"
      | TBot => kwdBot tex
      | TVar (id) => stringOfIdent id

(**
 * converts quantified type list into string.
 *
 * @params form tex ott i il
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the tex flag.
 * @param i the quantifier identifier list.
 * @param il the type list.
 *
 * @return the string representation of i il.
 *)
fun stringOfQuantifiedTypeList form tex ott i il =
	let val idl = if List.null i then ""
					else if not ott then "{" ^ (stringOfIdentList i) ^ "}"
							else (List.foldr (fn (id, s) => "{" ^ (stringOfIdent id) ^ "}" ^ s) "" i)
		val tyl = String.concatWith ", " (List.map (stringOfType form tex ott) il)
	in
		idl	^ "[" ^ tyl ^ "]"
	end

(**
 * converts typed identifiers into string.
 *
 * @params form tex ott i
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the tex flag.
 * @param i the identifier.
 *
 * @return the string representation of i.
 *)
fun stringOfTypedIdent form tex ott (x, t) =
	(stringOfIdent x) ^ ":" ^ (stringOfType form tex ott t)

(**
 * converts typed identifier lists into string.
 *
 * @params form tex ott il
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the tex flag.
 * @param il the typed identifier list.
 *
 * @return the string representation of il.
 *)
fun stringOfTypedIdentList form tex ott il =
	String.concatWith
		(if form then kwdAnd tex else ", ")
		(List.map (stringOfTypedIdent form tex ott) il)

(**
 * converts quantified typed identifier lists into string.
 *
 * @params form tex ott i il
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param ott the tex flag.
 * @param i the quantifier identifier list.
 * @param il the typed identifier list.
 *
 * @return the string representation of i il.
 *)
fun stringOfQuantifiedTypedIdentList form tex ott i il =
	let val idl = if List.null i then ""
					else if not ott then "{" ^ (stringOfIdentList i) ^ "}"
							else (List.foldr (fn (id, s) => "{" ^ (stringOfIdent id) ^ "}" ^ s) "" i)
		val tyl = stringOfTypedIdentList form tex ott il
	in
		idl	^ "[" ^ tyl ^ "]"
	end

(**
 * converts parameter list into string accordingly to the formula/tex demand.
 *
 * @params form tex displayType idlist mode typidlist
 * @param form the formula-disply (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param displayType flag indicating whether types have to be displayed.
 * @param idlist the identifier/quantification list.
 * @param mode the mode in or out.
 * @param typidlist the typed identifier list of formal parameters.
 *
 * @return the string representation of the parameter list.
 *)
fun stringOfParamList form tex displayType idlist mode typidlist =
	(if not (displayType) then ""
     else if List.null idlist then "" else "{" ^ (stringOfIdentList idlist) ^ "} ")
    ^ (if List.null typidlist then ""
       else mode ^ " " ^ (if displayType
       						then stringOfTypedIdentList form tex false typidlist
       						else stringOfIdentList (List.map #1 typidlist)))

end

