(**
 * The First-Order LoopW Type Inference compiler.
 *
 * The compiler calls several modules in order to
 * perform the requested tasks.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure Compiler :> COMPILER = struct

(**
 * The type of compiling parameters.
 *)
type compilerParams = 
	{ (** the path of the source file which contains the imperative program ".loop". *)
	  sourceFile : string,
	  (** the source file name prefix. *)
	  fileNamePrefix : string,
	  (** the verbosity level (0: nothing, ..., 3: very verbose). *)
	  verbLevel : int,
	  (** asks the compiler to write the untyped program in a file ".cs". *)
	  printUntyped : bool,
	  (** asks the compiler to write the typed program in a file ".rev.loop". *)
	  printTyped : bool,
	  (** asks the compiler to write the complete infered proof in a file ".proof". *)
	  printProof : bool,
	  (** asks the compiler to write the types as formulas. *)
	  formula : bool,
	  (** asks the compiler to write the output for further tex compilation. *)
	  tex : bool,
	  (** asks the compiler to write the complete infered proof in a file ".loop.ott" for ott/twelf compilation. *)
	  ott : bool
	}

(**
 * prints a message according to its importance and the requested verbosity.
 *
 * @params params level mess
 * @param params the compiling parameters containing the requested verbosity level.
 * @param level the message verbosity level.
 * @param mess the message to be printed.
 *)
fun print_verb (params : compilerParams) level mess =
    if (#verbLevel params) >= level then print mess else ()

(**
 * prints a string list to a file.
 *
 * @params name strList
 * @param name the file name.
 * @param strList the string list.
 *)
fun writeFile name strList = let
    val file = TextIO.openOut (name)
    val rec aux = fn
	[] => TextIO.closeOut file
      | s::q => (TextIO.output (file, s ^ "\n"); aux q)
in
    aux strList
end

(**
 * prints the compiling parameters values.
 *
 * @params params
 * @param params the compiling parameters.
 *)
fun printParams (params : compilerParams) =
	if (#verbLevel params) >= 3
	then
		let val () = print ("\n== Compilation options:")
			val () = print ("\n== source file: " ^ (#sourceFile params))
			val () = print ("\n== source file name prefix: " ^ (#fileNamePrefix params))
			val () = print ("\n== verbosity level: " ^ (Int.toString (#verbLevel params)))
			val () = print ("\n== untyped source print: " ^ (if #printUntyped params then "yes" else "no"))
			val () = print ("\n== typed source print: " ^ (if #printTyped params then "yes" else "no"))
			val () = print ("\n== infered proof print: " ^ (if #printProof params then "yes" else "no"))
			val () = print ("\n== formulas: " ^ (if #formula params then "yes" else "no"))
			val () = print ("\n== tex: " ^ (if #tex params then "yes" else "no"))
			val () = print ("\n== ott: " ^ (if #ott params then "yes" else "no"))
			val () = print ("\n==\n\n")
		in () end
	else ()

(**
 * performs type inference.
 *
 * @params params
 * @param params the compiling parameters.
 *)
fun compile (params : compilerParams) = let
	val () = print_verb params 3 ("\n===== Starting Compilation  =====\n==")
	val () = printParams params

	(* Parsing source code *)
	val prog = Parser.parseFile (#sourceFile params)

	val () = (* if the user asks for the untyped source code *)
	    if (#printUntyped params) then
		let val () = print_verb params 3 "\n----- Writing down untyped source code -----\n\n"
			val () = writeFile ((#fileNamePrefix params) ^ ".cs")
				       (CorePrettyPrinter.stringOfProgram_Untyped (#formula params) (#tex params) prog)
		in () end
	    else ()

	val () = (* if the user asks for the typed source code *)
	    if (#printTyped params) then
		let val () = print_verb params 3 "\n----- Writing down typed source code -----\n\n"
			val () = writeFile ((#fileNamePrefix params) ^ ".rev.loop")
				       (CorePrettyPrinter.stringOfProgram (#formula params) (#tex params) prog)
		in () end
	    else ()

	val () = print_verb params 3 "\n----- Infering Proof Code from Source Code: "
	val proofSeq = ProofInfer.proofInfer prog
	val () = print_verb params 3 "ok -----\n\n"

	val () = print_verb params 3 "\n----- Writing down Typing and Proof Obligations (.typ) -----\n\n"
	val () = writeFile ((#fileNamePrefix params) ^ ".typ")
					(ProofDerivPrettyPrinter.stringOfTyping
						(#formula params) (#tex params)
						(CorePrettyPrinter.stringOfProgram_Untyped (#formula params) (#tex params) prog)
						proofSeq)

	val () = (* if the user asks for the complete proof *)
	    if (#printProof params) then
		let val () = print_verb params 3 "\n----- Writing down complete proof code (.proof) -----\n\n"
			val () = writeFile ((#fileNamePrefix params) ^ ".proof")
				       (ProofPrettyPrinter.stringOfProgram (#formula params) (#tex params) proofSeq)
		in () end
	    else ()
	    
	val () = (* if the user asks for the complete proof for ott/twelf compilation *)
	    if (#ott params) then
		let val () = print_verb params 3
						"\n----- Writing down complete proof code for ott/twelf compilation (.ott) -----\n\n"
			val () = writeFile ((#fileNamePrefix params) ^ ".ott")
				       (ProofPrettyPrinter.stringOfProgram_Ott proofSeq)
		in () end
	    else ()
	    
    in print_verb params 3 "\n===== Compilation succeeded =====\n" end
    
end

