(**
 * The First-Order LoopW Type Inference core abstract syntax.
 *
 * The core abstract syntax is used to represent input LoopW programs.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure CoreAst = struct

open Error

(** identifiers as in proof abstract syntax. *)
type ident = ProofAst.ident

(** terms as in proof abstract syntax. *)
type term = ProofAst.term

(** formulas as in proof abstract syntax. *)
type expType = ProofAst.formula 

(** typed identifiers. *)
type typedIdent = ident * expType

(** commands. *)
datatype command =
	(** block: { s }{i}x:tau *)
	Block of block
	(** for-loop: for I := e { s }{j}x:tau. *)
  | For of ident * expression * block
  	(** assignment: x := e. *)
  | Affect of ident * expression
  	(** incrementation: inc(x). *)
  | Inc of ident
  	(** decrementation: dec(x). *)
  | Dec of ident
   	(** procedure call: p(el, il). *)
  | ProcCall of expression * expression list * ident list
  	(** label definition: k:{ s }{j}x:tau. *)
  | Label of ident * block
   	(** jump to a label: jump(k,el)x:tau. *)
  | Jump of expression * expression list * typedIdent list
  | (** region used for error display. *) CommRegion of command * region

(** sequences. *)
and sequence =
	(** empty sequence. *)
	Empty
	(** composed command: c; s. *)
  | Comm of command * sequence
  	(** constant declaration: cst x = e; s. *)
  | Cst of ident * expression * sequence
  	(** variable declaration: var x := e; s. *)
  | Var of ident * expression * sequence
  | (** region used for error display. *) SeqRegion of sequence * region

(** expressions. *)
and expression =
	(** identifier. *)
	Id of ident
	(** value. *)
  | Val of value
  	(** cast to a given type: e :> tau. *)
  | Cast of expression * expType
  | (** region used for error display. *) ExpRegion of expression * region

(** values. *)
and value =
	(** null value: *. *)
	Star
	(** natural number: n. *)
  | Int of int
 	(** anonymous procedure declaration: proc({il} in xfl; {jl} out yfl){ s }{jl}yfl. *)
  | Proc of ident list * typedIdent list
  	* ident list * typedIdent list * block
  | (** region used for error display. *) ValRegion of value * region

(** blocks. *)
withtype block = sequence * (ident list * typedIdent list)

(** programs. *)
type program = sequence

end

