(**
 * Utility function for the core syntax.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure CoreAstUtils :> CORE_AST_UTILS = struct

open CoreAst

(** True type *)
val top = ProofAst.TEqual (ProofAst.FZero, ProofAst.FZero)

(**
 * builds the successor of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the term s(t).
 *)
fun s t = ProofAst.FFun (("s", 0), [t])

(**
 * builds the predecessor of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the term p(t).
 *)
fun p t = ProofAst.FFun (("p", 0), [t])

(**
 * builds an unary natural number from an int.
 *
 * @params n
 * @param n the integer value.
 *
 * @return the unary natural number s^n(Z).
 *)
fun mkNat n =
	if n = 0
	then ProofAst.FZero
	else ProofAst.FFun (("s", 0), [mkNat (n - 1)])

(**
 * removes regions in sequences abstract syntax.
 *
 * @params s
 * @param s the sequence.
 *
 * @return the sequence without any region.
 *)
fun removeRegionSequence s =
    case s of
	Empty => Empty
      | Comm (c, s) => Comm (removeRegionCommand c,
			     removeRegionSequence s)
      | Cst (id, e, s) => Cst (id, removeRegionExpression e,
			       removeRegionSequence s)
      | Var (id, e, s) => Var (id, removeRegionExpression e,
			       removeRegionSequence s)
      | SeqRegion (s, _) => removeRegionSequence s

(**
 * removes regions in commands abstract syntax.
 *
 * @params c
 * @param c the expression.
 *
 * @return the command without any region.
 *)
and removeRegionCommand c =
    case c of
	Block (s, tidl) => Block (removeRegionSequence s, tidl)
      | For (id, e, (s, tidl)) => For (id, removeRegionExpression e,
				       (removeRegionSequence s, tidl))
      | Affect (id, e) => Affect (id, removeRegionExpression e)
      | Inc (id) => Inc (id)
      | Dec (id) => Dec (id)
      | ProcCall (e, el, idl) => ProcCall (removeRegionExpression e, 
					   List.map removeRegionExpression el, idl)
      | Label (id, (s, tidl)) => Label (id, (removeRegionSequence s, tidl))
      | Jump (e, el, idl) => Jump (removeRegionExpression e, 
				   List.map removeRegionExpression el, idl)
      | CommRegion (c, _) => removeRegionCommand c

(**
 * removes regions in expressions abstract syntax.
 *
 * @params e
 * @param e the expression.
 *
 * @return the expression without any region.
 *)
and removeRegionExpression e =
    case e of
	Id (id) => Id (id)
      | Val (v) => Val (removeRegionValue v)
      | Cast (e,t) => Cast (removeRegionExpression e, t)
      | ExpRegion (e, _) => removeRegionExpression e

(**
 * removes regions in values abstract syntax.
 *
 * @params v
 * @param v the value.
 *
 * @return the value without any region.
 *)
and removeRegionValue v =
    case v of
	Star => Star
      |	Int (n) => Int (n)
      | Proc (i, inl, j, outl, (s, outl2)) => Proc (i, inl, j, outl, (removeRegionSequence s, outl2))
      | ValRegion (v, _) => removeRegionValue v

(**
 * removes regions in programs abstract syntax.
 *
 * @params p
 * @param p the program.
 *
 * @return the program without any region.
 *)
fun removeRegionProgram p = removeRegionSequence p

end

