(**
 * The First-Order LoopW Type Inference core syntax pretty-printer.
 *
 * The core pretty-printer converts core abstract syntax into string.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure CorePrettyPrinter :> CORE_PRETTY_PRINTER = struct

open PrettyPrinterKeywords
open BasePrettyPrinter
open CoreAst

(**
 * converts block into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ annot b
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param annot flag indicating whether block annotation has to be displayed (for proc...).
 * @param b the block.
 *
 * @return the string representation of the block.
 *)
fun stringOfBlock_internal tab form tex typ annot (seq, (j, idlist)) =
    (stringOfSequence_internal (tab + 1) form tex typ seq)
    @ [(stringOfTab tab) ^ "}"
    	^ (if annot
    	   then (if typ
    		     then (if List.null j then "" else "{" ^ (stringOfIdentList j) ^ "}")
    		     		^ (stringOfTypedIdentList form tex false idlist)
    			 else stringOfIdentList (List.map #1 idlist))
    	   else "")
		^ ";"]

(**
 * converts command into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ c
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param c the command.
 *
 * @return the string representation of the command.
 *)
and stringOfCommand_internal tab form tex typ = fn
    Block blc =>
    ((stringOfTab tab) ^ "{") :: (stringOfBlock_internal tab form tex typ true blc)

  | For (id, exp, blc) =>
    ((stringOfTab tab) ^ (kwd tex "for") ^ " " ^ (stringOfIdent id) ^ " := 0 "
     ^ (kwd tex "until") ^ " "
     ^ (List.hd (stringOfExpression_internal tab form tex typ exp)) ^ " {")
    :: (stringOfBlock_internal tab form tex typ true blc)

  | Affect (id, exp) =>
    let val l = stringOfExpression_internal tab form tex typ exp in
		[(stringOfTab tab) ^ (stringOfIdent id) ^ " := "
		^ (List.hd l) ^ (if List.length l = 1 then ";" else "")]
		@ (if List.length l > 1 then List.tl l else [])
    end

  | Inc (id) =>
    [(stringOfTab tab) ^ (kwd tex "inc") ^ "(" ^ (stringOfIdent id) ^ ");"]

  | Dec (id) =>
    [(stringOfTab tab) ^ (kwd tex "dec")  ^ "(" ^ (stringOfIdent id) ^ ");"]

  | ProcCall (exp, exp_list, id_list) =>
    [(stringOfTab tab)
    ^ (List.hd (stringOfExpression_internal tab form tex typ exp))
    ^ "(" ^ (String.concat (stringOfExpressionList tab form tex typ exp_list))
    ^ "; " ^ (stringOfIdentList id_list) ^ ");"]

  | Label (id, blc) =>
    ((stringOfTab tab) ^ (stringOfIdent id) ^ ": {") :: (stringOfBlock_internal tab form tex typ true blc)

  | Jump (exp, exp_list, typid_list) =>
    [(stringOfTab tab) ^ (kwd tex "jump") ^ "("
    ^ (List.hd (stringOfExpression_internal tab form tex typ exp))
    ^ ", " ^ (String.concat (stringOfExpressionList tab form tex typ exp_list))
    ^ ")" ^ (if typ then stringOfTypedIdentList form tex false typid_list
    			else stringOfIdentList (List.map #1 typid_list)) ^ ";"]

  | CommRegion (c, _) => stringOfCommand_internal tab form tex typ c

(**
 * converts sequence into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ s
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param s the sequence.
 *
 * @return the string representation of the sequence.
 *)
and stringOfSequence_internal tab form tex typ = fn
    Empty => []

  | Comm (c, seq) =>
    (stringOfCommand_internal tab form tex typ c)
    @ stringOfSequence_internal tab form tex typ seq

  | Cst (id, exp, seq) =>
    let val l = stringOfExpression_internal tab form tex typ exp in
	[(stringOfTab tab) ^ (kwd tex "cst") ^ " " ^ (stringOfIdent id) ^ " = "
	 ^ (List.hd l) ^ (if List.length l = 1 then ";" else "")]
	@ (if List.length l > 1 then List.tl l else [])
	@ (stringOfSequence_internal tab form tex typ seq)
    end

  | Var (id, exp, seq) =>
    let val l = stringOfExpression_internal tab form tex typ exp in
	[(stringOfTab tab) ^ (kwd tex "var") ^ " " ^ (stringOfIdent id) ^ " := "
	 ^ (List.hd l) ^ (if List.length l = 1 then ";" else "")]
	@ (if List.length l > 1 then List.tl l else [])
	@ (stringOfSequence_internal tab form tex typ seq)
    end

  | SeqRegion (s, _) => stringOfSequence_internal tab form tex typ s

(**
 * converts expression into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ e
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param e the expression.
 *
 * @return the string representation of the expression.
 *)
and stringOfExpression_internal tab form tex typ = fn
    Id (id) => [stringOfIdent id]

  | Val (v) => stringOfValue_internal tab form tex typ v

  | Cast (e,t) =>
    [(List.hd (stringOfExpression_internal tab form tex typ e))
     ^ (if typ then " :> " ^ (stringOfType form tex false t) else "")]

  | ExpRegion (e, _) => stringOfExpression_internal tab form tex typ e

(**
 * converts expression list into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ el
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param el the expression list.
 *
 * @return the string representation of the expression list.
 *)
and stringOfExpressionList tab form tex typ exp_list =
    Utils.listConcatWith [", "] (List.map (stringOfExpression_internal tab form tex typ) exp_list)

(**
 * converts value into string accordingly to the formula/tex demand.
 *
 * @params tab form tex typ v
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param typ flag indicating whether types have to be displayed.
 * @param v the value.
 *
 * @return the string representation of the value.
 *)
and stringOfValue_internal tab form tex typ = fn
    Star => [(kwdStar tex)]

  | Int (n) => [Int.toString n]

  | Proc (i, in_list, j, out_list, blc) =>
    ((kwd tex "proc") ^ "(" ^ (stringOfParamList form tex typ i (kwd tex "in") in_list)
     ^ "; " ^ (stringOfParamList form tex typ j (kwd tex "out") out_list)
     ^ ") {")
    :: (stringOfBlock_internal tab form tex typ false blc)

  | ValRegion (v, _) => stringOfValue_internal tab form tex typ v

(**
 * converts value into string accordingly to the formula/tex demand.
 *
 * @params form tex v
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param v the value.
 *
 * @return the string representation of v.
 *)
fun stringOfValue form tex v = stringOfValue_internal 0 form tex true v

(**
 * converts expression into string accordingly to the formula/tex demand.
 *
 * @params form tex e
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param e the expression.
 *
 * @return the string representation of e.
 *)
fun stringOfExpression form tex e = stringOfExpression_internal 0 form tex true e

(**
 * converts command into string accordingly to the formula/tex demand.
 *
 * @params form tex c
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param c the command.
 *
 * @return the string representation of c.
 *)
fun stringOfCommand form tex c = stringOfCommand_internal 0 form tex true c

(**
 * converts sequence into string accordingly to the formula/tex demand.
 *
 * @params form tex s
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param s the sequence.
 *
 * @return the string representation of s.
 *)
fun stringOfSequence form tex s = stringOfSequence_internal 0 form tex true s

(**
 * converts program into string accordingly to the formula/tex demand.
 *
 * @params form tex p
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param p the program.
 *
 * @return the string representation of p.
 *)
fun stringOfProgram form tex p = stringOfSequence_internal 0 form tex true p

(**
 * converts program into string accordingly to the formula/tex demand,
 * without any type information.
 *
 * @params form tex p
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param p the program.
 *
 * @return the string representation of p.
 *)
fun stringOfProgram_Untyped form tex p = stringOfSequence_internal 0 form tex false p

end

