(**
 * The First-Order LoopW Type Inference main program.
 *
 * The main program parses the command-line arguments and calls the compiler.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure Main = struct

(** Current compiler version number. *)
val compVersion = "loopW internal compiler version 0.3"

(**
 * displays the compiler command usage and exits.
 *)
fun usage () = let
in
	print ("usage: loopW [-version] [-v123] [-uprint] [-print] [-pprint] [-form] [-tex] [-ott] file.loop\n"
			^ "  -version print version and exit\n"
			^ "  -v       Verbosity level (1 : low, 2 : medium, 3 : high)\n"
			^ "  -uprint  Erase all type information and pretty print (file.cs)\n"
			^ "  -print   Pretty print (file.rev.loop)\n"
			^ "  -pprint  Complete infered proof pretty printing (file.proof)\n"
			^ "  -form    Display formula-like types instead of imperative types\n"
			^ "  -tex     Tex output\n"
			^ "  -ott     Ott/twelf complete infered proof pretty printing (file.loop.ott)\n");
	OS.Process.exit OS.Process.failure
end

(**
 * gets the value in an Option, calling usage() if it is NONE.
 *
 * @params x
 * @param x the Option value.
 *
 * @return the inner value.
 *)
fun someOf x =
	case x of
		NONE => usage ()
      | SOME x => x

(**
 * parses the argument list to build the tuple of options.
 *
 * @params l
 * @param l the string list of arguments.
 *
 * @return the tuple (version, verbosity, printUntyped, printTyped, printProof,
 * formula, tex, funTranslate, ott, fileName).
 *)
 fun parseArgs l =
 	if List.length l = 0 then usage () else
        (case List.find (fn s => s = "-version") l of
             NONE => false
           | SOME _ => true,
         case List.find (fn s => (s <> "-version") andalso ((String.substring (s,0,2)) = "-v")) l of
             NONE => 0
           | SOME s =>
           	let val n = someOf (Int.fromString (String.extract (s, 2, SOME 1)))
            in if (n < 1) orelse (n > 3) then usage() else n end,
         case List.find (fn s => s = "-uprint") l of
             NONE => false
           | SOME _ => true,
         case List.find (fn s => s = "-print") l of
             NONE => false
           | SOME _ => true,
         case List.find (fn s => s = "-pprint") l of
             NONE => false
           | SOME _ => true,
         case List.find (fn s => s = "-form") l of
             NONE => false
           | SOME _ => true,
         case List.find (fn s => s = "-tex") l of
             NONE => false
           | SOME _ => true,
         case List.find (fn s => s = "-ott") l of
             NONE => false
           | SOME _ => true,
         List.last l)

(**
 * builds the tuple of options and calls the compiler with it.
 *)
fun run () =
	let val (version, verb, printuprog, printprog, printproof, form, tex, ott, file) =
			parseArgs (CommandLine.arguments ())
		val name =
			String.concatWith "."
				(List.rev (List.tl (List.rev (String.tokens (fn c => c = #".") file))))
		val (params : Compiler.compilerParams) =
			{ sourceFile = file, fileNamePrefix = name, verbLevel = verb,
			  printUntyped = printuprog, printTyped = printprog, printProof = printproof,
			  formula = form, tex = tex, ott = ott}
	in if version
		then print (compVersion ^ "\n")
		else Compiler.compile params
    end

(**
 * calls run() and wait for any exception in order to display the exception message.
 *)
fun main _ =
	((run () handle e =>
		(TextIO.output (TextIO.stdErr,
						((General.exnMessage e) ^ "\n"));
		 OS.Process.exit OS.Process.failure));
      OS.Process.success)

end

