(**
 * The First-Order LoopW Type Inference proof derivation pretty-printer.
 *
 * The proof derivation pretty-printer converts proof abstract syntax into string
 * containing both program and typing derivation.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ProofDerivPrettyPrinter :> PROOF_DERIV_PRETTY_PRINTER = struct

open PrettyPrinterKeywords
open BasePrettyPrinter
open ProofAst
open CoreAstUtils

(** Lemma counter *)
val (init, next) =
	let val c = ref 0 in
		(fn () => c := 0, fn () => (c := !c + 1; !c))
	end

(**
 * builds a string with | and spaces for typing tabular display.
 *
 * @params n
 * @param tex the tex flag.
 * @param n the number of tabs.
 *
 * @return the string tab.
 *)
fun stringOfTab_T tex n =
	if n = 0 then ""
	else (kwdVBar tex) ^ "   " ^ (stringOfTab_T tex (n - 1))

(**
 * appends a lemma number reference to a string list representing a program.
 * If the program string list is empty, it returns a new line ; else it appends
 * the lemma number reference to the last line.
 *
 * @params tab l s
 * @param tab the number of tabs.
 * @param l the string list representing the program.
 * @param s the string representation of the lemma number.
 *
 * @return the string representation of the program with lemma reference.
 *)
fun append_with_lemma tab tex l s =
	case l of
		[] => [(stringOfTab_T tex tab) ^ s]
	  | [s'] =>
		if String.isPrefix "    by #" s'
		then [(stringOfTab_T tex tab) ^ s ^ s']
		else s'::[(stringOfTab_T tex tab) ^ s]
	  | s'::q => s'::(append_with_lemma tab tex q s)

(**
 * concatenates two string; if they represent a lemma number reference, it
 * glues them together.
 *
 * @params s1 s2
 * @param s1 the first string.
 * @param s2 the second string.
 *
 * @return the concatenation of s1 and s2.
 *)
fun lemma_fusion s1 s2 =
	if (String.isPrefix "    by #" s1)
		andalso (String.isPrefix "    by #" s2)
	then s1 ^ ", " ^ (String.extract (s2, 7, NONE))
	else s1 ^ s2

(**
 * converts command typing derivation into string accordingly to the formula/tex demand.
 *
 * @params tab form tex c
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param c the command.
 *
 * @return the string representation of the command typing derivation.
 *)
fun stringOfCommand tab form tex = fn
	Block (s, (j, tidl)) =>
	let val (po, ty) = stringOfSequence (tab + 1) form tex s
	in
		(po,
		 ((stringOfTab_T tex tab) ^ "-/"  ^ (stringOfIdentList (List.map #1 tidl)) ^ "/-")
		 :: (append_with_lemma tab tex ty
		 		("[" ^ (stringOfTypedIdentList form tex false tidl) ^ "]")))
	end

  | For (id, lid, exp, etyp, (s, (j, tidl))) =>
	let val (po1, ty1) = stringOfExpression tab form tex exp
		val (po2, ty2) = stringOfSequence (tab + 1) form tex s
		val ty1' = String.concatWith " | " ty1
		val out_l = Terms.substituteAll lid
						(case etyp of TNat(n) => n
							| _ => raise Fail "ProofPrettyPrinter: not a TNat") tidl
	in
		(po1 @ po2,
		 ((stringOfTab_T tex tab) ^ "-"
		 ^ "(" ^ (stringOfTypedIdent form tex false (id, TNat (FId (lid)))) ^ ")"
		 ^ "[" ^ (stringOfTypedIdentList form tex false tidl) ^ "]")
		 :: (append_with_lemma tab tex ty2
		 		("[" ^ (stringOfTypedIdentList form tex false out_l) ^ "]")))
	end

  | Affect (id, etyp, exp) =>
	let val (po, ty) = stringOfExpression tab form tex exp
	in
		(po,
		 append_with_lemma tab tex ty
			("[" ^ (stringOfTypedIdent form tex false (id, etyp)) ^ "]"))
	end

  | Inc (id, etyp) =>
	let val t = (case etyp of TNat(n) => TNat(s(n))
					| _ => raise Fail "ProofPrettyPrinter: not a Nat")
	in
		([],
		 [(stringOfTab_T tex tab) ^ "[" ^ (stringOfTypedIdent form tex false (id, t)) ^ "]"])
	end

  | Dec (id, etyp) =>
	let val t = (case etyp of TNat(n) => TNat(p(n))
					| _ => raise Fail "ProofPrettyPrinter: not a Nat")
	in
		([],
		 [(stringOfTab_T tex tab) ^ "[" ^ (stringOfTypedIdent form tex false (id, t)) ^ "]"])
	end

  | ProcCall (exp, etyp, exp_typ_list, subst, typid_list) =>
	let val (po1, ty1) = stringOfExpression tab form tex exp
		val (po2, ty2) = stringOfExpressionList tab form tex exp_typ_list
		val ty1' = String.concatWith " | " ty1
		val ty2' = String.concatWith " | " ty2
	in
		(po1 @ po2,
		 [(stringOfTab_T tex tab) ^ "[" ^ (stringOfTypedIdentList form tex false typid_list) ^ "]"
		 ^ (lemma_fusion ty1' ty2')])
	end

  | Label (id, etyp, (s, (j, tidl))) =>
	let val (po, ty) = stringOfSequence (tab + 1) form tex s
	in
		(po,
		 ((stringOfTab_T tex tab) ^ "-/" ^ (stringOfIdentList (List.map #1 tidl)) ^ "/-"
		 ^ "(" ^ (stringOfTypedIdent form tex false (id, etyp)) ^ ")")
		 :: (append_with_lemma tab tex ty
				("[" ^ (stringOfTypedIdentList form tex false tidl) ^ "]")))
	end

  | Jump (exp, etyp, exp_typ_list, subst, typid_list) =>
	let val (po1, ty1) = stringOfExpression tab form tex exp
		val (po2, ty2) = stringOfExpressionList tab form tex exp_typ_list
		val ty1' = String.concatWith " | " ty1
		val ty2' = String.concatWith " | " ty2
	in
		(po1 @ po2,
		 [(stringOfTab_T tex tab) ^ "["
		 ^ (stringOfTypedIdentList form tex false typid_list) ^ "]"
		 ^ (lemma_fusion ty1' ty2')])
	end

(**
 * converts sequence typing derivation into string accordingly to the formula/tex demand.
 *
 * @params tab form tex s
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param s the sequence.
 *
 * @return the string representation of the sequence typing derivation.
 *)
and stringOfSequence tab form tex (s, (j, omega)) =
	case s of
		Empty (subst) => ([], [])

	  | Comm (c, seq) =>
		let val (po1, ty1) = stringOfCommand tab form tex c
			val (po2, ty2) = stringOfSequence tab form tex seq
		in
			(po1 @ po2, ty1 @ ty2)
		end

	  | Cst (id, t, exp, seq) =>
		let val (po1, ty1) = stringOfExpression tab form tex exp
			val (po2, ty2) = stringOfSequence tab form tex seq
		in
			(po1 @ po2,
			 (append_with_lemma tab tex ty1
				("(" ^ (stringOfIdent id)
				 ^ ":" ^ (stringOfType form tex false t) ^ ")"))
			 @ ty2)
		end

	  | Var (id, t, exp, seq) =>
		let val (po1, ty1) = stringOfExpression tab form tex exp
			val (po2, ty2) = stringOfSequence tab form tex seq
		in
			(po1 @ po2,
			 (append_with_lemma tab tex ty1
			 	("[" ^ (stringOfIdent id)
			 	 ^ ":" ^ (stringOfType form tex false t) ^ "]"))
			 @ ty2)
		end

	  | SSubst (seq, omega_i, id, exp, t) =>
	  	let val (po1, ty1) = stringOfSequence tab form tex seq
			val (po2, ty2) = stringOfExpression tab form tex exp
		in
			(po1 @ po2, ty1 @ ty2)
		end

(**
 * converts expression typing derivation into string accordingly to the formula/tex demand.
 *
 * @params tab form tex e
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param e the expression.
 *
 * @return the string representation of the expression typing derivation.
 *)
and stringOfExpression tab form tex (e, tau) =
	case e of
		Id (id) => ([], [])

	  | Val (v) => stringOfValue tab form tex v

	  | ESubst (e, t_i, id, exp, neqm) =>
	  	let val (po1, ty1) = stringOfExpression tab form tex e
			val (po2, ty2) = stringOfExpression tab form tex exp
		in
			(po1 @ po2, ty1 @ ty2)
		end

	  | Lemma (tl, t) =>
	  	let val n = Int.toString (next ())
	  	in
			([n ^ ": "
			 ^ (String.concatWith ", " (List.map (stringOfType form tex false) tl))
			 ^ (kwdVdash tex) ^ " " ^ (stringOfType form tex false t)],
			 ["    by #" ^ n])
		end

(**
 * converts expression list into string accordingly to the formula/tex demand.
 *
 * @params tab form tex el
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param el the expression list.
 *
 * @return the string representation of the expression list typing derivation.
 *)
and stringOfExpressionList tab form tex exp_list =
	case exp_list of
		[] => ([], [])
	  | [(e,t)] => stringOfExpression tab form tex e
	  | (e,t)::q =>
	  	let val (po1, ty1) = stringOfExpression tab form tex e
			val (po2, ty2) = stringOfExpressionList tab form tex q
			val ty1' = String.concatWith " | " ty1
			val ty2' = String.concatWith " | " ty2
		in
			(po1 @ po2, [(lemma_fusion ty1' ty2')])
		end

(**
 * converts value typing derivation into string accordingly to the formula/tex demand.
 *
 * @params tab form tex v
 * @param tab the tabulation level.
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param v the value.
 *
 * @return the string representation of the value typing derivation.
 *)
and stringOfValue tab form tex (v, tau) =
	case v of
		Star =>
		(case tau of
			TEqual (n, m) =>
			if n <> m
			then
				let val n = Int.toString (next ())
				in
					([n ^ ": " ^ (kwdVdash tex) ^ (stringOfType form tex false tau)],
					 ["    by #" ^ n])
				end
			else ([], [])
		  | _ => ([], []))

	  | Int (n) => ([], [])

	  | Proc (i, in_list, j, out_list, (s, (k, tidl))) =>
	  	let val (po, ty) = stringOfSequence (tab + 1) form tex s
	  	in
			(po,
			 ((stringOfTab_T tex tab) ^ "-"
			 ^ (if List.length in_list > 0
			 	then "(" ^ (stringOfTypedIdentList form tex false in_list) ^ ")"
			 	else "")
			 ^ (if List.length out_list > 0
			 	then "[" ^ (stringOfTypedIdentList form tex false
				 (List.map (fn (x,t) => (x, TEqual(FZero, FZero))) out_list)) ^ "]"
				else ""))
			:: ty)
		end

(**
 * calculates the longest string in a string list.
 *
 * @params sl
 * @param sl the string list.
 *
 * @return the size of the longest string in sl.
 *)
fun max_width sl =
	List.foldl (fn (s,l) =>
		   let val m = String.size s
		   in if m > l then m else l end) 0 sl

(**
 * builds a string containing the given number of white spaces.
 *
 * @params n
 * @param n the number of white spaces.
 *
 * @return the string with n spaces.
 *)
fun make_spc n = if n <= 0 then "" else " " ^ (make_spc (n - 1))

(**
 * appends at the end of each string of a string list an amount of white
 * spaces in order to have exactly string of the same given size.
 *
 * @params max
 * @param max the size each string of sl must be.
 * @param sl the string list.
 *
 * @return sl with each string augmented to size max.
 *)
fun tabulate max = fn
	[] => []
  | s::q => (s ^ (make_spc (max - (String.size s))))::(tabulate max q)

(**
 * completes to string list in order to have the same list size.
 *
 * @params (l1, l2)
 * @param l1 the first string list.
 * @param l2 the second string list.
 *
 * @return the pair of lists having with the shortest one completed with
 * empty strings.
 *)
fun complete (l1, l2) =
	case (l1, l2) of
		([], []) => ([], [])
	  | (a::b, x::y) => let val (b', y') = complete (b, y) in (a::b', x::y') end
	  | (a::b, []) => let val (b', y') = complete (b, []) in (a::b', ""::y') end
	  | ([], x::y) => let val (b', y') = complete ([], y) in (""::b', x::y') end

(**
 * converts program with typing derivation into string
 * accordingly to the formula/tex demand.
 *
 * @params form tex p
 * @param form the formula-display (instead of imperative types) flag.
 * @param tex the tex flag.
 * @param pr the string representation of the untyped program.
 * @param p the program.
 *
 * @return the string representation of p along with its partial typing derivation.
 *)
fun stringOfTyping form tex pr p =
	let val () = init ()
		val (po, ty) = stringOfSequence 0 form tex p
		val pr' = if tex then pr else tabulate (max_width pr) pr
		val (pr1, ty1) = complete (pr', ty)
		val s =
			if tex
			then
				(ListPair.map (fn (x,y) =>
					(if String.isPrefix "\\mathbf{cst} " (String.extract(x, 2, NONE))
					 then "\n\n" else "")
					^ (implode [#"\034"]) ^ "\\parbox[t]{0.45\\textwidth}{" ^ (implode [#"\034"])
					^ x ^ (implode [#"\034"]) ^ "}" ^ (implode [#"\034"])
					^ " & "
					^ (implode [#"\034"]) ^ "\\parbox[t]{0.45\\textwidth}{" ^ (implode [#"\034"])
					^ y ^ (implode [#"\034"]) ^ "}" ^ (implode [#"\034"]))
					(pr1, ty1))
			else
				(ListPair.map (fn (x,y) => x ^ "   " ^ y) (pr1, ty1))
	in
		s @ ("" :: po)
	end

end

