(**
 * The First-Order LoopW Type Inference proof inference.
 *
 * The proof inference builds a complete proof structure from a
 * source code.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure ProofInfer :> PROOF_INFER = struct

open CoreAst
open CoreAstUtils
open Terms
open TermMatchUnif
open ProofEnvironment

(** Infix operation on environment *)
infix +! ++ -! --

(**
 * result type, either OK or ERROR with information enough to display an error message.
 *)
datatype result = OK | ERROR of env * env * (ident list * env) option * string * Error.region option

(**
 * builds a string with an error message, depending on the exception.
 *
 * @params ex
 * @param ex the exception.
 *
 * @return the string representation of the exception.
 *)
fun getMsg ex =
	case ex of
	MatchErrorTypes (tau, tau') =>
    "Cannot match\n  " ^ (BasePrettyPrinter.stringOfType false false false tau)
    ^ "\nwith\n  " ^ (BasePrettyPrinter.stringOfType false false false tau')
      | MatchErrorTypeList (tl, tl') =>
    "Cannot match\n  " ^ (String.concatWith " *** " (List.map (BasePrettyPrinter.stringOfType false false false) tl))
    ^ "\nwith\n  " ^ (String.concatWith " *** " (List.map (BasePrettyPrinter.stringOfType false false false) tl'))
      | MatchErrorTerms (t, t') =>
    "Cannot match\n  " ^ (BasePrettyPrinter.stringOfTerm false t)
    ^ "\nwith\n  " ^ (BasePrettyPrinter.stringOfTerm false t')
      | MatchUnifErrorTypes (tau, tau') =>
    "Cannot match+unif\n  " ^ (BasePrettyPrinter.stringOfType false false false tau)
    ^ "\nwith\n  " ^ (BasePrettyPrinter.stringOfType false false false tau')
      | MatchUnifErrorTypeList (tl, tl') =>
    "Cannot match+unif\n  " ^ (String.concatWith " *** " (List.map (BasePrettyPrinter.stringOfType false false false) tl))
    ^ "\nwith\n  " ^ (String.concatWith " *** " (List.map (BasePrettyPrinter.stringOfType false false false) tl'))
      | MatchUnifErrorTerms (t, t') =>
    "Cannot match+unif\n  " ^ (BasePrettyPrinter.stringOfTerm false t)
    ^ "\nwith\n  " ^ (BasePrettyPrinter.stringOfTerm false t')
      | NotUnifyableTypes (tau, tau') =>
    "Cannot unify\n  " ^ (BasePrettyPrinter.stringOfType false false false tau)
    ^ "\nand\n  " ^ (BasePrettyPrinter.stringOfType false false false tau')
      | NotUnifyableTerms (t, t') =>
    "Cannot unify\n  " ^ (BasePrettyPrinter.stringOfTerm false t)
    ^ "\nand\n  " ^ (BasePrettyPrinter.stringOfTerm false t')
	  | AbsentIdent (x, xl) =>
	(BasePrettyPrinter.stringOfIdent x) ^ " not in "
	^ (BasePrettyPrinter.stringOfTypedIdentList false false false xl)
	  | LengthMismatch => "List.length mismatch"
	  | excep => raise excep

(**
 * collects the equalities in two environment, discarding identities.
 *
 * @params gamma omega
 * @param gamma the first environment.
 * @param omega the second environment.
 *
 * @return the equalities (n = m) where (n <> m).
 *)
fun getEqualities gamma omega =
	let val getE = fn init => fn env =>
			  List.foldl
				(fn ((_, ProofAst.TEqual (n, m)), l)
						=> if (not (Utils.listMember (ProofAst.TEqual (n, m)) l))
									andalso (n <> m)
								then (ProofAst.TEqual (n, m))::l else l
				   | (_, l) => l)
				 init env
	in
		getE (getE [] gamma) omega
	end

(**
 * builds a substitution derivation for a sequence.
 *
 * @params gamma omega seq (j, phi_i) eql
 * @param gamma the first typing environment.
 * @param omega the second typing environment.
 * @param seq the sequence.
 * @param j the list of identifier existentially quantified.
 * @param phi_i the formula phi[i] in which i will be substituted by the terms of the equation.
 * @param eql the equation list [(i, m = n)].
 *
 * @return the sequence derivation with the substitution rule.
 *)
fun mk_seq_subst_lemma gamma omega seq (j, phi_i) eql = (* phi[i] and [(i,m,n)] *)
	let val eq = getEqualities gamma omega (* collect the equalities to build the lemma *)
		val rec aux = (* auxiliary function taking the final inner sequence
						 and the pair of equation list (already-done, to-be-done) *)
		fn s => fn
		  (l, []) => s (* if there is no further equations, return the sequence *)
		| (l, (i, m, n)::q) =>
		let val phi_n_j = (* build phi[n/j] for j in the already-done equation list *)
				substituteAllList (List.map (fn (i', m', n') => (i', n')) l) phi_i
			val phi_n_j_m_k = (* phi[n/j][m/k] for k in the to-be-done equation list *)
				substituteAllList (List.map (fn (i', m', n') => (i', m')) l) phi_n_j
			val phi_n_m = (* phi[n/j][n/i][m/k] for the current equation *)
				substituteAll i n phi_n_j_m_k
			val lemma = (* build the lemma expression *)
				(ProofAst.Lemma (eq, ProofAst.TEqual (m, n)), ProofAst.TEqual (m, n))
		in
			aux (* add the substitution rule to the inner sequence *)
				(ProofAst.SSubst (s, phi_n_j_m_k, i, lemma, ProofAst.TEqual (m, n)), (j, phi_n_m))
				(* and move the equation to the already-done part *)
				((i, m, n)::l, q)
		end
	in aux seq ([], eql) end

(**
 * builds a substitution derivation for an expression.
 *
 * @params gamma omega exp phi_i eql
 * @param gamma the first typing environment.
 * @param omega the second typing environment.
 * @param exp the expression.
 * @param phi_i the formula phi[i] in which i will be substituted by the terms of the equation.
 * @param eql the equation list (i, m = n).
 *
 * @return the expression derivation with the substitution rule.
 *)
fun mk_exp_subst_lemma gamma omega exp phi_i eql = (* phi[i] and (i,m,n) *)
	let val eq = getEqualities gamma omega (* collect the equalities to build the lemma *)
		val rec aux = (* auxiliary function taking the final inner expression
						 and the pair of equation list (already-done, to-be-done) *)
		fn e => fn
		  (l, []) => e (* if there is no further equations, return the expression *)
		| (l, (i, m, n)::q) =>
		let val phi_n_j = (* build phi[n/j] for j in the already-done equation list *)
				substituteList (List.map (fn (i', m', n') => (i', n')) l) phi_i
			val phi_n_j_m_k = (* phi[n/j][m/k] for k in the to-be-done equation list *)
				substituteList (List.map (fn (i', m', n') => (i', m')) q) phi_n_j
			val phi_n_m = (* phi[n/j][n/i][m/k] for the current equation *)
				substitute i n phi_n_j_m_k
			val lemma = (* build the lemma expression *)
				(ProofAst.Lemma (eq, ProofAst.TEqual (m, n)), ProofAst.TEqual (m, n))
		in
			aux (* add the substitution rule to the inner expression *)
				(ProofAst.ESubst (e, phi_n_j_m_k, i, lemma, ProofAst.TEqual (m, n)), phi_n_m)
				(* and move the equation to the already-done part *)
				((i, n, m)::l, q)
		end
	in aux exp ([], eql) end

(**
 * infer proof for sequences according to the loopW dependent type system.
 *
 * @params gamma omega (j, omega') reg sequence
 * @param gamma the constant identifier input typing environment.
 * @param omega the variable identifier input typing environment.
 * @param j the identifier list of existential quantification of output environment.
 * @param omega' the variable identifier output typing environment (existentialy quantified over j).
 * @param reg the region of the source code containing this sequence (for error messages).
 * @param sequence the sequence on which to perform proof inference.
 *
 * @return (seq', result) where seq' is the complete proof representation of sequence
 * and result is either OK if proof inference suceeded or ERROR with the error data.
 *)
fun inferSeq gamma omega (j, omega') reg sequence =
	let val void = (ProofAst.Empty [], (j, omega')) in (* empty sequence used to display error messages *)
	case sequence of
		(* Rule T.EMPTY *)
		Empty => (* G; W |- e |> {j}W' *)
		(* We first need to remove from omega the variables that are not in omega' *)
		(let val delta = List.filter (fn (x,t) => ListAssoc.memberKey x omega') omega
			(* Unification and matching to get the types (omega''[j][k], omega''[u/j][k]), (j,m,n), (i,u) *)
			val (typl, eql, subst) = matchUnifList j (omega', delta) (* (omega''[j][n/k], omega''[u/j][m/k]) *)
			val eql = List.map (fn (i,n,m) => (i,m,n)) eql (* we must permute equation in order to have (before, after) *)
			val omega_j_k = List.map (fn (x, t1, t2) => (x, t1)) typl (* omega''[j][k] *)
			val omega_j_m = substituteAllList (List.map (fn (k, m, n) => (k, m)) eql) omega_j_k (* omega''[j][m/k] *)
			(* We build the derivation using the substitution rule and lemmas *)
			val proof_seq = mk_seq_subst_lemma gamma omega (ProofAst.Empty (subst), (j, omega_j_m)) (j, omega_j_k) eql
		 in
			(proof_seq, OK)
		 end
		 handle excep => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.EMPTY: " ^ (getMsg excep), reg)))

	  (* Rule T.CST *)
	  | Cst (id, exp, seq) => (* G; W |- cst id = exp; seq |> {j}W' *)
		let val (proof_exp, res1) = (* G; W |- exp : tau ( = #2 proof_exp) *)
				inferExp gamma omega reg exp
			val (proof_seq, res2) = (* G, id : tau; W |- seq |> {j}W' *)
				inferSeq (gamma +! (id, #2 proof_exp)) omega (j, omega') reg seq
		in
			if res1 <> OK (* Typing error in exp *)
			then ((ProofAst.Cst (id, #2 proof_exp, proof_exp, void), (j, omega')), res1)
			else ((ProofAst.Cst (id, #2 proof_exp, proof_exp, proof_seq), (j, omega')), res2)
		end

	  (* Rule T.VAR *)
	  | Var (id, exp, seq) => (* G; W |- var id := exp; seq |> {j}W' *)
		let val (proof_exp, res1) = (* G; W |- exp : tau ( = #2 proof_exp) *)
				inferExp gamma omega reg exp
			val (proof_seq, res2) = (* G; W, id : tau |- seq |> {j}W' *)
				inferSeq gamma (omega +! (id, #2 proof_exp)) (j, omega') reg seq
		in
			if res1 <> OK (* Typing error in exp *)
			then ((ProofAst.Var (id, #2 proof_exp, proof_exp, void), (j, omega')), res1)
			else if res2 <> OK (* Typing error in seq *)
				then ((ProofAst.Var (id, #2 proof_exp, proof_exp, proof_seq), (j, omega')), res2)
				else
					((ProofAst.Var (id, #2 proof_exp, proof_exp, proof_seq), (j, omega')), res2)
		end

	  (* Rule T.BLOCK *)
	  | Comm (Block (seq1, (k, x_tau_list)), seq2) => (* G; W |- { seq1 }{k}x:tau; seq2 |> {j}W' *)
		(case split omega (dom x_tau_list) of (* We try to split omega according to the out list x_tau_list *)
			  NONE => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.BLOCK: undefined variables", reg))
			| SOME (x_sigma_list, omega'') =>
			let val (proof_seq1, res1) = (* G; x : sigma |- seq1 |> {k}x:tau *)
					inferSeq gamma x_sigma_list (k, x_tau_list) reg seq1
				val (proof_seq2, res2) = (* G; W, x : tau |- seq2 |> {j}W' *)
					inferSeq gamma (x_tau_list ++ omega'') (j, omega') reg seq2
			in (* We check the side condition about k that must not be free in surrouding types *)
				if not (List.all (fn id => not (Utils.listMember id ((fv_formulas (img gamma)) (* k not-in FV(G,W) *)
																	@ (fv_formulas (img omega))))
											andalso ((Utils.listMember id j) (* k in j or k not-in FV(W')*)
													orelse (not (Utils.listMember id (fv_formulas (img omega'))))))
								k)
				then (void, ERROR (gamma, omega, SOME (j, omega'),
							"ProofInfer: T.BLOCK: " ^ (BasePrettyPrinter.stringOfIdentList k) ^ " in FV(G,W,\n"
							^ (BasePrettyPrinter.stringOfTypedIdentList false false false omega'), reg))
				else
					if res1 <> OK (* Typing error in seq1 *)
					then ((ProofAst.Comm (ProofAst.Block (proof_seq1, (k, x_tau_list)), void), (j, omega')), res1)
					else ((ProofAst.Comm (ProofAst.Block (proof_seq1, (k, x_tau_list)), proof_seq2), (j, omega')), res2)
			end)

	  (* Rule T.INC *)
	  | Comm (Inc (id), seq) => (* G; W, id : nat(n) |- inc(id); seq |> {j}W' *)
		(case getVarType id omega of
		  SOME (ProofAst.TNat (n)) => (* id must be declared and of type TNat *)
		  let val (proof_seq, res) = (* G; W, id : nat(s(n)) |- seq |> {j}W' *)
				inferSeq gamma (omega +! (id, ProofAst.TNat (s n))) (j, omega') reg seq
		  in
			((ProofAst.Comm (ProofAst.Inc (id, ProofAst.TNat (n)), proof_seq), (j, omega')), res)
		  end
		| _ => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.INC: undefined variable"
									^ (BasePrettyPrinter.stringOfIdent id), reg)))

	  (* Rule T.DEC *)
	  | Comm (Dec (id), seq) => (* G; W, id : nat(n) |- dec(id); seq |> {j}W' *)
		(case getVarType id omega of
		  SOME (ProofAst.TNat (n)) => (* id must be declared and of type TNat *)
		  let val (proof_seq, res) = (* G; W, id : nat(p(n)) |- seq |> {j}W' *)
				inferSeq gamma (omega +! (id, ProofAst.TNat (p n))) (j, omega') reg seq
		  in
			((ProofAst.Comm (ProofAst.Dec (id, ProofAst.TNat (n)), proof_seq), (j, omega')), res)
		  end
		| _ => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.DEC: undefined variable"
									^ (BasePrettyPrinter.stringOfIdent id), reg)))

	  (* Rule T.ASSIGN *)
	  | Comm (Affect (id, exp), seq) => (* G; W, id : tau |- id := exp; seq |> {j}W' *)
		(case getVarType id omega of
		  SOME _ => (* id must be declared of any type *)
		  let val (proof_exp, res1) = (* G; W, id : tau |- exp : sigma ( = #2 proof_exp) *)
					inferExp gamma omega reg exp
			val (proof_seq, res2) = (* G; W, id : sigma |- seq |> {j}W' *)
					inferSeq gamma (omega +! (id, #2 proof_exp)) (j, omega') reg seq
		  in
			if res1 <> OK (* Typing error in exp *)
			then ((ProofAst.Comm (ProofAst.Affect (id, #2 proof_exp, proof_exp), void), (j, omega')), res1)
			else ((ProofAst.Comm (ProofAst.Affect (id, #2 proof_exp, proof_exp), proof_seq), (j, omega')), res2)
		  end
		| _ => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.ASSIGN: undefined variable"
									^ (BasePrettyPrinter.stringOfIdent id), reg)))

	  (* Rule T.FOR *)
	  | Comm (For (id, exp, (seq1, (k, x_sigma_list))), seq2) => (* G; W, x : sigma[0] |- for id := 0 until exp { seq1 }x:sigma[id]; seq2 |> {j}W' *)
		let val (proof_exp, res1) = (* G; W, id : tau |- exp : tau ( = #2 proof_exp) *)
				inferExp gamma omega reg exp
		in
			if res1 <> OK (* Typing error in exp *)
			then ((ProofAst.Comm (ProofAst.For (id, id, proof_exp, #2 proof_exp, (void, (j, omega'))), void), (j, omega')), res1)
			else if List.length k <> 0 (* The user tries to existentially quantifiate over loop invariant *)
				then (void, ERROR (gamma, omega, SOME (j, omega'),
						"ProofInfer: T.FOR: no existential quantification allowed", reg))
				else
				(case #2 proof_exp of (* check the type of exp *)
					ProofAst.TNat(n) => (* OK we have a type nat(n) *)
					let val sigma_0 = substituteAll id ProofAst.FZero x_sigma_list (* build sigma[0] *)
						val sigma_S = substituteAll id (s (ProofAst.FId id)) x_sigma_list (* build sigma[s(id)] *)
						val sigma_N = substituteAll id n x_sigma_list (* build sigma[n] *)
					in
					(case split omega (dom x_sigma_list) of (* We try to split omega according to the out list x_sigma_list *)
						NONE => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.FOR: undefined variables", reg))
					  | SOME (sigma_0', omega'') =>
						if not (envEquals sigma_0 sigma_0') (* We check whether sigma_0 is defined *)
						then (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.FOR: missing sigma[0]", reg))
						else
							let val (proof_seq1, res1) = (* G, id : nat(id); x : sigma[id] |- seq1 |> x : sigma[s(id)] *)
									inferSeq (gamma +! (id, ProofAst.TNat (ProofAst.FId (id)))) x_sigma_list (k, sigma_S) reg seq1
								val (proof_seq2, res2) = (* G; W, x : sigma[N] |- seq2 |> {j}W' *)
									inferSeq gamma (sigma_N ++ omega'') (j, omega') reg seq2
							in (* We check the side condition about id that must not be free in gamma *)
								if Utils.listMember id (fv_formulas (img gamma))
								then (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.FOR: id in gamma", reg))
								else if res1 <> OK (* Typing error in seq1 *)
									then ((ProofAst.Comm (ProofAst.For (id, id, proof_exp, ProofAst.TNat(n),
											(proof_seq1, (k, x_sigma_list))), void),
											(j, omega')), res1)
									else ((ProofAst.Comm (ProofAst.For (id, id, proof_exp, ProofAst.TNat(n),
											(proof_seq1, (k, x_sigma_list))), proof_seq2),
											(j, omega')), res2)
							end)
					end
				  | _ => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.FOR: exp not of type nat", reg)))
		end

	  (* Rule T.CALL *)
	  | Comm (ProcCall (exp, exp_list, id_list), seq) => (* G; W, idl : rho |- exp(expl, idl); seq |> {j}W' *)
		(let val (proof_exp, res) = (* G; W, idl : rho |- exp : proc({i} in tau; {k} out sigma) (= #2 proof_exp) *)
				inferExp gamma omega reg exp
			val (proof_exp_list, res_list) = (* G; W, idl : rho |- expl : tau[N] (= #2 proof_exp_list) *)
				ListPair.unzip (List.map (inferExp gamma omega reg) exp_list)
			val tau_N = List.map #2 proof_exp_list
		 in
			if (res <> OK) orelse List.exists (fn x => x <> OK) res_list (* Typing error in exp or exp_list *)
			then ((ProofAst.Comm (ProofAst.ProcCall (proof_exp, #2 proof_exp,
							  ListPair.zip (proof_exp_list, (List.map #2 proof_exp_list)),
							  [], []),
					   void), (j, omega')),
					if res <> OK then res else valOf (List.find (fn x => x <> OK) res_list))
			else
			(case #2 proof_exp of
				ProofAst.TProc(i, tau, k, sigma) => (* OK we have a proc type *)
				let val subst = matchOrderedList i (tau, tau_N) (* We match tau[i] with tau[N] and build the substitution *)
					val sigma_N = List.map (substituteList subst) sigma (* We build sigma[N] *)
					val out_r = (ListPair.zip (id_list, sigma_N)) (* We build id_list:sigma[N] *)
				in
					if (List.length id_list) = (List.length sigma)
					then (* If we have as many out variables as requiered *)
					(case split omega id_list of (* We try to split omega according to the out list *)
						NONE => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.CALL: undefined out variables", reg))
					  | SOME (_, omega'') =>
						let val (proof_seq, res) = (* G; W, idl : sigma[N] |- seq |> {j}W' *)
								inferSeq gamma (out_r ++ omega'') (j, omega') reg seq
						in (* We check the side condition about k that must not be free in surrouding types *)
							if not (List.all (fn id => not (Utils.listMember id ((fv_formulas (img gamma)) (* k not-in FV(G,W) *)
																				@ (fv_formulas (img omega))))
														andalso ((Utils.listMember id j) (* k in j or k not-in FV(W')*)
																orelse (not (Utils.listMember id (fv_formulas (img omega'))))))
											k)
							then (void, ERROR (gamma, omega, SOME (j, omega'),
										"ProofInfer: T.CALL: " ^ (BasePrettyPrinter.stringOfIdentList k) ^ " in FV(G,W,\n"
										^ (BasePrettyPrinter.stringOfTypedIdentList false false false omega'), reg))
							else ((ProofAst.Comm (ProofAst.ProcCall (proof_exp, #2 proof_exp,
																	 ListPair.zip (proof_exp_list, (List.map #2 proof_exp_list)),
																	 subst, out_r),
												  proof_seq), (j, omega')), res)
						end)
					else (void, ERROR (gamma, omega, SOME (j, omega'),
										"ProofInfer: T.CALL: incorrect number of out variables", reg))
				end
			  | _ => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.CALL: exp not of type proc", reg)))
		 end
		 handle excep => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.CALL: " ^ (getMsg excep), reg)))

	  (* Rule T.LABEL *)
	  | Comm (Label (id, (seq1, (k, x_tau_list))), seq2) => (* G; W |- id:{ seq1 }{k}x:tau; seq2 |> {j}W' *)
		let val neg_tau = ProofAst.TProc (k, List.map #2 x_tau_list, [], [ProofAst.TBot]) (* the negation of the out type *)
		in
		(case split omega (dom x_tau_list) of (* We try to split omega according to the out list x_tau_list *)
			NONE => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.LABEL: undefined variables", reg))
		  | SOME (x_sigma_list, omega'') =>
			let val (proof_seq1, res1) = (* G, id : neg_tau; x : sigma |- seq1 |> {k}x:tau *)
					inferSeq (gamma +! (id, neg_tau)) x_sigma_list (k, x_tau_list) reg seq1
				val (proof_seq2, res2) = (* G; W, x : tau |- seq2 |> {j}W' *)
					inferSeq gamma (x_tau_list ++ omega'') (j, omega') reg seq2
			in (* We check the side condition about k that must not be free in surrouding types *)
				if not (List.all (fn id => not (Utils.listMember id ((fv_formulas (img gamma)) (* k not-in FV(G,W) *)
																	@ (fv_formulas (img omega))))
											andalso ((Utils.listMember id j) (* k in j or k not-in FV(W')*)
													orelse (not (Utils.listMember id (fv_formulas (img omega'))))))
						k)
				then (void, ERROR (gamma, omega, SOME (j, omega'),
									"ProofInfer: T.LABEL: " ^ (BasePrettyPrinter.stringOfIdentList k) ^ " in FV(G,W,\n"
									^ (BasePrettyPrinter.stringOfTypedIdentList false false false omega'), reg))
				else
					if res1 <> OK (* Typing error in seq1 *)
					then ((ProofAst.Comm (ProofAst.Label (id, neg_tau, (proof_seq1, (k, x_tau_list))),
											void), (j, omega')), res1)
					else ((ProofAst.Comm (ProofAst.Label (id, neg_tau, (proof_seq1, (k, x_tau_list))),
											proof_seq2), (j, omega')), res2)
			end)
		end

	  (* Rule T.JUMP *)
	  | Comm (Jump (exp, exp_list, typid_list), seq) => (* G; W, idl : rho |- jump(exp, expl, idl:sigma); seq2 |> {j}W' *)
		(let val (proof_exp, res) = (* G; W, idl : rho |- exp : {i}~tau[i] *)
				inferExp gamma omega reg exp
			val (proof_exp_list, res_list) = (* G; W, idl : rho |- expl : tau[N] (= #2 proof_exp_list) *)
				ListPair.unzip (List.map (inferExp gamma omega reg) exp_list)
			val tau_N = List.map #2 proof_exp_list
		 in
			if (res <> OK) orelse List.exists (fn x => x <> OK) res_list (* Typing error in exp or exp_list *)
			then ((ProofAst.Comm (ProofAst.Jump (proof_exp, #2 proof_exp,
												ListPair.zip (proof_exp_list, (List.map #2 proof_exp_list)),
												[], []),
									void), (j, omega')),
				if res <> OK then res else valOf (List.find (fn x => x <> OK) res_list))
			else
			(case #2 proof_exp of
				ProofAst.TProc(i, tau, [], [ProofAst.TBot]) => (* OK we have a neg type *)
				let val subst = matchOrderedList i (tau, tau_N) (* We match tau[i] with tau[N] and build the substitution *)
				in
				(case split omega (List.map #1 typid_list) of (* We try to split omega according to the out list *)
					NONE => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.JUMP: undefined out variables", reg))
				  | SOME (_, omega'') =>
					let val (proof_seq, res) = (* G; W, idl : sigma |- jump(exp, expl, idl); seq2 |> {j}W' *)
							inferSeq gamma (typid_list ++ omega'') (j, omega') reg seq
					in
					((ProofAst.Comm (ProofAst.Jump (proof_exp, #2 proof_exp,
													ListPair.zip (proof_exp_list, List.map #2 proof_exp_list),
													subst, typid_list),
									proof_seq), (j, omega')), res)
					end)
				end
			  | _ => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.JUMP: exp not of type neg", reg)))
		 end
		 handle excep => (void, ERROR (gamma, omega, SOME (j, omega'), "ProofInfer: T.JUMP: " ^ (getMsg excep), reg)))

	  | SeqRegion (s, r) => inferSeq gamma omega (j, omega') (SOME r) s
	  | Comm (CommRegion (c, r), s) => inferSeq gamma omega (j, omega') (SOME r) (Comm (c, s))

	end

(**
 * infer proof for expressions according to the loopW dependent type system.
 *
 * @params gamma omega reg exp
 * @param gamma the constant identifier input typing environment.
 * @param omega the variable identifier input typing environment.
 * @param reg the region of the source code containing this expression (for error messages).
 * @param exp the expression on which to perform proof inference.
 *
 * @return (exp', result) where exp' is the complete proof representation of exp
 * and result is either OK if proof inference suceeded or ERROR with the error data.
 *)
and inferExp gamma omega reg = fn
	(* Rule T.ENV *)
	Id (id) => (* G; W |- id : tau , if id:tau is in G, W *)
	(case getVarType id (gamma ++ omega) of
		NONE => ((ProofAst.Id ("ERR", 0), ProofAst.TVar ("ERR", 0)),
					ERROR (gamma, omega, NONE, "ProofInfer: T.ENV: Undefined identifier "
												^ (BasePrettyPrinter.stringOfIdent id), reg))
	  | SOME tau => ((ProofAst.Id (id), tau), OK))

  | Val (v) =>
	let val (proof_val, res) = inferVal gamma omega reg v
	in
		((ProofAst.Val (proof_val), #2 proof_val), res)
	end

  (* Casts in the source code must be converted in applications of the substitution rule *)
  | Cast (e, t) => (* G; W |- e :> tau : tau *)
	((case (CoreAstUtils.removeRegionExpression e, t) of
		(Val (Star), ProofAst.TEqual (n,m)) => (* If we cast * into n = m, then we don't need unification *)
		((ProofAst.Val (ProofAst.Star, ProofAst.TEqual (n,m)), ProofAst.TEqual (n,m)), OK)
	  | _ =>
	  	let val (proof_exp, res) = (* G; W |- e : sigma *)
	  			inferExp gamma omega reg e
	  	in
	  		if res <> OK then ((ProofAst.Id ("ERR", 0), ProofAst.TVar ("ERR", 0)), res) (* Typing error in e *)
	  		else (* We unify sigma with tau and obtain tau[i] and an equation list eql *)
	  			let val (t_i, eql) = unify (#2 proof_exp, t)
	  				val proof_exp' = (* with use tau[i] and eql to build the substitution rules *)
	  					mk_exp_subst_lemma gamma omega proof_exp t_i eql
	  			in
	  				(proof_exp', OK)
	  			end
	  	end)
	 handle excep => ((ProofAst.Id ("ERR", 0), ProofAst.TVar ("ERR", 0)),
			  ERROR (gamma, omega, NONE, "ProofInfer: CAST: " ^ (getMsg excep), reg)))

  | ExpRegion (e, r) => inferExp gamma omega (SOME r) e

(**
 * infer proof for values according to the loopW dependent type system.
 *
 * @params gamma omega reg val
 * @param gamma the constant identifier input typing environment.
 * @param omega the variable identifier input typing environment.
 * @param reg the region of the source code containing this value (for error messages).
 * @param val the value on which to perform proof inference.
 *
 * @return (val', result) where val' is the complete proof representation of val
 * and result is either OK if proof inference suceeded or ERROR with the error data.
 *)
and inferVal gamma omega reg = fn
	(* Rule T.UNIT *)
	Star => (* G; W |- * : T *)
	((ProofAst.Star, top), OK)

  (* Rule T.NUM *)
  | Int (n) => (* G; W |- n : nat(n) *)
	let val tn = ProofAst.TNat (CoreAstUtils.mkNat (n)) (* We build the nat type term *)
	in
		((ProofAst.Int (n), tn), OK)
	end

  (* Rule T.PROC *)
  | Proc (i, in_list, j, out_list, (seq, _)) =>
  	(* G; W |- proc({i} in x:tau[i]; {j} out y:sigma[i][j]){ seq }_ : proc({i} in tau[i]; {j} out sigma[i][j]) *)
	if List.length out_list = 0
	then ((ProofAst.Star, ProofAst.TVar ("ERR", 0)),
		ERROR (gamma, omega, NONE, "ProofInfer: T.PROC: no output parameters", reg))
	else
		let val (proof_seq, res) = (* G, ; W, idl : rho |- jump(exp, expl, idl:sigma); seq2 |> {j}W' *)
				inferSeq (in_list ++ gamma) (List.map (fn (x,t) => (x, top)) out_list) (j, out_list) reg seq
		val proc_type = ProofAst.TProc (i, List.map #2 in_list, j, List.map #2 out_list)
	in (* We check the side condition about i that must not be free in gamma *)
		if not (List.all (fn id => not (Utils.listMember id (fv_formulas (img gamma)))) i)
		then ((ProofAst.Star, ProofAst.TVar ("ERR", 0)),
				ERROR (gamma, omega, NONE, "ProofInfer: T.PROC: " ^ (BasePrettyPrinter.stringOfIdentList i)
											^ " in FV(G)", reg))
		else
			((ProofAst.Proc (i, in_list, j, out_list, (proof_seq, (j, out_list))), proc_type), res)
	end

  | ValRegion (v, r) => inferVal gamma omega (SOME r) v

(**
 * infer proof for programs according to the loopW dependent type system.
 *
 * @params gamma omega (j, omega') p
 * @param gamma the constant identifier input typing environment.
 * @param omega the variable identifier input typing environment.
 * @param j the identifier list of existential quantification of output environment.
 * @param omega' the variable identifier output typing environment (existentialy quantified over j).
 * @param p the program on which to perform proof inference.
 *
 * @throws Fail (via Error.fail) if proof inference did not succed.
 *
 * @return the complete proof representation of p.
 *)
fun inferProg gamma omega (j, omega') (p : program) =
	case inferSeq gamma omega (j, omega') NONE p of
		(seq, OK) => seq
	  | (seq, ERROR (gamma, omega, omega', msg, reg)) =>
	  	(print "\n";
	  	print (String.concatWith
	  			"\n" (ProofDerivPrettyPrinter.stringOfTyping
				  false false (CorePrettyPrinter.stringOfProgram_Untyped false false p) seq));
		print "\n";
		print ("\nGamma: [" ^ (BasePrettyPrinter.stringOfTypedIdentList false false false gamma) ^ "]\nOmega: ["
				^ (BasePrettyPrinter.stringOfTypedIdentList false false false omega) ^ "]\n"
				^ (case omega' of
			   NONE => ""
			 | SOME (j, omega') => "j: {" ^ (BasePrettyPrinter.stringOfIdentList j) ^ "}\nOmega': ["
					   ^ (BasePrettyPrinter.stringOfTypedIdentList false false false omega') ^ "]\n"));
		case reg of
		 SOME r => Error.fail r msg
	   | NONE => Error.failm msg)

(**
 * returns the complete proof.
 *
 * @params p
 * @param p the source program.
 *
 * @return the complete proof representation of p.
 *)
fun proofInfer (p : program) =
	inferProg empty empty ([], empty) p

end

