(**
 * The First-Order LoopW Type Inference formula/term matching and unification.
 *
 * Provides functions in order to perform pattern-matching and unification.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
signature TERM_MATCH_UNIF = sig

(** matching error for types *)
exception MatchErrorTypes of (ProofAst.formula * ProofAst.formula)
(** matching error for types lists *)
exception MatchErrorTypeList of (ProofAst.formula list * ProofAst.formula list)
(** matching error for terms *)
exception MatchErrorTerms of (ProofAst.term * ProofAst.term)

(** matching+unification error for types *)
exception MatchUnifErrorTypes of (ProofAst.formula * ProofAst.formula)
(** matching+unification error for types lists *)
exception MatchUnifErrorTypeList of (ProofAst.formula list * ProofAst.formula list)
(** matching+unification error for terms *)
exception MatchUnifErrorTerms of (ProofAst.term * ProofAst.term)

(** unification error for types *)
exception NotUnifyableTypes of (ProofAst.formula * ProofAst.formula)
(** unification error for terms *)
exception NotUnifyableTerms of (ProofAst.term * ProofAst.term)

(** no identifier found error *)
exception AbsentIdent of (ProofAst.ident * (ProofAst.ident * ProofAst.formula) list)

(** length mismatch beetwen lists error *)
exception LengthMismatch

(**
 * performs the pattern-matching where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (p, t)
 * @param idl the identifier list of pattern meta-variables.
 * @param p the formula pattern.
 * @param t the formula to match.
 *
 * @throws MatchErrorTypes if the pattern cannot match the formula.
 *
 * @return the substitutions list that matches.
 *)
val match
	: ProofAst.ident list -> (ProofAst.formula * ProofAst.formula)
	-> (ProofAst.ident * ProofAst.term) list

(**
 * performs the pattern-matching of a formulas list with a patterns list
 * where the pattern meta-variables belongs to an identifiers list. The pattern-matching is performed
 * all along the lists, using the same order.
 *
 * @params idl (pl, fl)
 * @param idl the identifier list of pattern meta-variables.
 * @param pl the formulas list pattern.
 * @param fl the formulas list to match.
 *
 * @throws MatchErrorTypeList if the pattern cannot match the formula.
 * @throws LengthMismatch if the length of the lists differs.
 *
 * @return the substitutions list that matches.
 *)
val matchOrderedList
	: ProofAst.ident list -> (ProofAst.formula list * ProofAst.formula list)
	-> (ProofAst.ident * ProofAst.term) list

(**
 * performs the pattern-matching of a formulas list with a patterns list
 * where the pattern meta-variables belongs to an identifiers list. The pattern-matching is performed
 * according to the identifier associated with the patterns and formulas elements.
 *
 * @params idl (pl, fl)
 * @param idl the identifier list of pattern meta-variables.
 * @param pl the (identifiers, formulas) list pattern.
 * @param fl the (identifiers, formulas) list to match.
 *
 * @throws MatchErrorTypeList if the pattern cannot match the formula.
 * @throws LengthMismatch if the length of the lists differs.
 * @throws AbsentIdent if an identifier belongs to pl and not to fl.
 *
 * @return the substitutions list that matches.
 *)
val matchList
	: ProofAst.ident list -> ((ProofAst.ident * ProofAst.formula) list
								* (ProofAst.ident * ProofAst.formula) list)
	-> (ProofAst.ident * ProofAst.term) list

(**
 * performs the pattern-matching and unification where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (pl, fl)
 * @param idl the identifier list of pattern meta-variables {i}.
 * @param pl the (identifier, formula) list pattern (x:f)[i][n/j].
 * @param fl the (identifier, formula) list to match (x:f)[u/i][m/j].
 *
 * @throws MatchUnifErrorTypes if the pattern cannot match the formula.
 *
 * @return the formulas with unification variables, the unification variables associated with equations
 * and the substitutions list that matches (((x:f)[i][j], (x:f)[u/i][j]), (j,n,m), (i,u)).
 *)
val matchUnifList
	: ProofAst.ident list -> ((ProofAst.ident * ProofAst.formula) list
								* (ProofAst.ident * ProofAst.formula) list)
	-> ((ProofAst.ident * ProofAst.formula * ProofAst.formula) list
		* (ProofAst.ident * ProofAst.term * ProofAst.term) list
		* (ProofAst.ident * ProofAst.term) list)

(**
 * performs the unification of two formulas on the term-level.
 *
 * @params (a, b)
 * @param a the first formula to unify f[n/i].
 * @param b the second formula to unify f[m/i].
 *
 * @throws NotUnifyableTypes when unification is impossible (formulas structure differs).
 *
 * @return the formula with meta-variables f[i] and the equations list (i, n = m).
 *)
val unify
	: (ProofAst.formula * ProofAst.formula)
	-> ProofAst.formula
		* (ProofAst.ident * ProofAst.term * ProofAst.term) list

(**
 * performs the unification of two formulas list on the term-level. The unification is
 * performed all along the lists, using to the same order.
 *
 * @params (l1, l2)
 * @param l1 the first formula to unify l1[n/i].
 * @param l2 the second formula to unify l2[m/i].
 *
 * @throws LengthMismatch if lists size differs.
 *
 * @return the list of formula with meta-variables f[i] associated with the equations list (i, n = m).
 *)
val unifyOrderedList
	: (ProofAst.formula list * ProofAst.formula list)
	-> (ProofAst.formula
		* (ProofAst.ident * ProofAst.term * ProofAst.term) list) list

(**
 * performs the unification of two (identifiers, formulas) list on the term-level.
 * The unification is performed
 * according to the identifier associated with the formulas.
 *
 * @params (xl1, xl2)
 * @param xl1 the first formula to unify xl1[n/i].
 * @param xl2 the second formula to unify xl2[m/i].
 *
 * @throws LengthMismatch if lists size differs.
 * @throws AbsentIdent if lists differs for some identifier.
 *
 * @return the list of formula with meta-variables f[i] associated with the equations list (i, n = m).
 *)
val unifyList
	: ((ProofAst.ident * ProofAst.formula) list
		* (ProofAst.ident * ProofAst.formula) list)
	-> ((ProofAst.ident * ProofAst.formula) list
		* (ProofAst.ident * ProofAst.term * ProofAst.term) list)

end

