(**
 * The First-Order LoopW Type Inference formula/term matching and unification.
 *
 * Provides functions in order to perform pattern-matching and unification.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure TermMatchUnif :> TERM_MATCH_UNIF = struct

open ProofAst
open Terms

(** matching error for types *)
exception MatchErrorTypes of (ProofAst.formula * ProofAst.formula)
(** matching error for types lists *)
exception MatchErrorTypeList of (ProofAst.formula list * ProofAst.formula list)
(** matching error for terms *)
exception MatchErrorTerms of (ProofAst.term * ProofAst.term)

(** matching+unification error for types *)
exception MatchUnifErrorTypes of (ProofAst.formula * ProofAst.formula)
(** matching+unification error for types lists *)
exception MatchUnifErrorTypeList of (ProofAst.formula list * ProofAst.formula list)
(** matching+unification error for terms *)
exception MatchUnifErrorTerms of (ProofAst.term * ProofAst.term)

(** unification error for types *)
exception NotUnifyableTypes of (ProofAst.formula * ProofAst.formula)
(** unification error for terms *)
exception NotUnifyableTerms of (ProofAst.term * ProofAst.term)

(** no identifier found error *)
exception AbsentIdent of (ProofAst.ident * (ProofAst.ident * ProofAst.formula) list)

(** length mismatch beetwen lists error *)
exception LengthMismatch

(**
 * generates a new fresh variable.
 *
 * @return a fresh variable "unif".
 *)
val fresh = fn _ => alpha ("var", 0)

(**
 * performs the pattern-matching of terms where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (p, t)
 * @param idl the identifier list of pattern meta-variables.
 * @param p the term pattern.
 * @param t the term to match.
 *
 * @throws MatchErrorTerms if the pattern cannot match the term.
 *
 * @return the substitutions list that matches.
 *)
fun term_match xl = fn
    (FId (id1), FId (id2)) =>
    if Utils.listMember id1 xl then [(id1, FId (id2))] (* a match is found, adding it to the substitution list *)
    else if id1 = id2 then [] (* the identifiers must be equals *)
    	else raise MatchErrorTerms (FId (id1), FId (id2)) (* else it is a matching error *)
    	
  | (FId (id1), t) =>
    if Utils.listMember id1 xl
    then [(id1, t)] (* a match is found, adding it to the substitution list *)
    else raise MatchErrorTerms (FId (id1), t) (* else it is a matching error *)
    
  | (FZero, FZero) => [] (* matching *)
  
  | (FFun (s, tl), FFun (s', tl')) => (* matching a function *)
    if (s <> s') orelse (List.length tl <> List.length tl') (* the arity must be the same (what about the function name ???) *)
    then raise MatchErrorTerms (FFun (s, tl), FFun (s', tl')) (* else it is a matching error *)
    else
		(* (if List.exists (fn x => x = s) xl
			then [(s, FId (s'))]
			else if s = s' then [] else raise MatchErrorTerms (FFun (s, tl), FFun (s', tl')))@ *)
		(case List.foldl (* union of substitutions *)
			(fn (x, SOME l) => ListAssoc.union x l | _ => NONE) (SOME [])
			(List.map (term_match xl) (ListPair.zip (tl, tl'))) of (* recursive call on argument list *)
			SOME l => l
		  | NONE => raise MatchErrorTerms (FFun (s, tl), FFun (s', tl'))) (* if some union is impossible, it is a matching error *)
		  
  | (a, b) => raise MatchErrorTerms (a, b) (* else it is a matching error *)

(**
 * performs the pattern-matching of formulas where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (p, f)
 * @param idl the identifier list of pattern meta-variables.
 * @param p the formula pattern.
 * @param f the formula to match.
 *
 * @throws MatchErrorTypes if the pattern cannot match the formula.
 *
 * @return the substitutions list that matches.
 *)
fun match xl = fn
    (TNat (n), TNat (m)) => term_match xl (n, m) (* calls matching on terms *)
    
  | (TEqual (n1, m1), TEqual (n2, m2)) => (* calls matching on the subterms and try union *)
    (case ListAssoc.union (term_match xl (n1, n2)) (term_match xl (m1, m2)) of
    	SOME l => l
      | NONE => raise MatchErrorTypes (TEqual (n1, m1), TEqual (n2, m2))) (* if the union is impossible, it is a matching error *)
       
  | (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2)) =>
    if (List.length sigma1 <> List.length sigma2) orelse
       (List.length tau1 <> List.length tau2) (* if *)
    then raise MatchErrorTypes (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2))
    else
    	let val ((i1', sigma1', j1', tau1'), (i2', sigma2', j2', tau2')) =
    		(* alpha-convert the two types to identify the bound variables *)
			alphaConvProc ((i1, sigma1, j1, tau1), (i2, sigma2, j2, tau2))
		in  (* recursive call upon every subtype *)
		    (case (ListAssoc.unionFlat (List.map (match xl) (ListPair.zip (tau1', tau2'))),
		    		ListAssoc.unionFlat (List.map (match xl) (ListPair.zip (sigma1', sigma2')))) of
		    		(SOME l1, SOME l2) =>
		    		(case ListAssoc.union l1 l2 of
		    			NONE => raise MatchErrorTypes (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2))
		    		  | SOME l => l)
		    	  | _ => raise MatchErrorTypes (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2)))
		end
	
  | (TVar (x), TVar (y)) =>
  	if x = y then [] (* variables must be identical (no first-order matching/unification) *)
  	else raise MatchErrorTypes (TVar (x), TVar (y))
  	
  | (TBot, TBot) => [] (* matching *)
  
  | (a, b) => raise MatchErrorTypes (a, b) (* else it is a matching error *)

(**
 * performs the pattern-matching of a formulas list with a patterns list
 * where the pattern meta-variables belongs to an identifiers list. The pattern-matching is performed
 * all along the lists, using the same order.
 *
 * @params idl (pl, fl)
 * @param idl the identifier list of pattern meta-variables.
 * @param pl the formulas list pattern.
 * @param fl the formulas list to match.
 *
 * @throws MatchErrorTypeList if the pattern cannot match the formula.
 * @throws LengthMismatch if the length of the lists differs.
 *
 * @return the substitutions list that matches.
 *)
fun matchOrderedList xl (l1, l2) =
    if List.length l1 <> List.length l2 then raise LengthMismatch
    else
	(case ListAssoc.unionFlat (List.map (match xl) (ListPair.zip (l1, l2))) of
	     SOME l => l
	   | NONE => raise MatchErrorTypeList (l1, l2))

(**
 * performs the pattern-matching of a formulas list with a patterns list
 * where the pattern meta-variables belongs to an identifiers list. The pattern-matching is performed
 * according to the identifier associated with the patterns and formulas elements.
 *
 * @params idl (pl, fl)
 * @param idl the identifier list of pattern meta-variables.
 * @param pl the (identifiers, formulas) list pattern.
 * @param fl the (identifiers, formulas) list to match.
 *
 * @throws MatchErrorTypeList if the pattern cannot match the formula.
 * @throws LengthMismatch if the length of the lists differs.
 * @throws AbsentIdent if an identifier belongs to pl and not to fl.
 *
 * @return the substitutions list that matches.
 *)
fun matchList xl (xl1, xl2) =
    if List.length xl1 <> List.length xl2 then raise LengthMismatch
    else
	let val rec aux = fn
		[] => []
	  | (x,t)::q =>
		(case List.find (fn (y,t') => y = x) xl2 of
		     SOME (y, t') =>
		     (case ListAssoc.union (match xl (t, t')) (aux q) of
		     	SOME l => l
		      | NONE => raise MatchErrorTypeList (List.map #2 xl1, List.map #2 xl2))
		   | _ => raise AbsentIdent (x, xl2))
	in
	    aux xl1
	end

(**
 * builds a new unification equation with a fresh identifier.
 *
 * @params (a, b) eql
 * @param a the first term of the equation.
 * @param b the second term of the equation.
 * @param eql the equation list to which add the new equation.
 *
 * @return the fresh identifier and the new equation list.
 *)
fun new_term_subst_eq (a, b) eql = let val i = fresh() in ((FId (i), FId (i)), (i, a, b)::eql, []) end

(**
 * performs the pattern-matching and unification where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (p, f)
 * @param idl the identifier list of pattern meta-variables {i}.
 * @param p the term pattern t[i][n/j].
 * @param f the term to match t[u/i][m/j].
 *
 * @throws MatchUnifErrorTerms if the pattern cannot match the term.
 *
 * @return the terms with unification variables, the unification variables associated with equations
 * and the substitutions list that matches ((t[i][j], t[u/i][j]), (j,n,m), (i,u)).
 *)
fun term_matchUnif xl (a, b) =
	(* auxiliary function called only if matching is requiered (no unification on terms) *)
    let val rec aux = fn xl => fn
	    (FId (id1), FId (id2)) =>
	    if Utils.listMember id1 xl
	    then ((FId (id1), FId (id2)), [], [(id1, FId (id2))]) (* a match is found, adding it to the substitution list *)
	    else if id1 = id2
	    		then ((FId (id1), FId (id2)), [], []) (* no equation *)
	    		else new_term_subst_eq (FId (id1), FId (id2)) [] (* new equation *)
	    
	  | (FId (id1), t) =>
	    if Utils.listMember id1 xl
	    then ((FId (id1), t), [], [(id1, t)]) (* a match is found, adding it to the substitution list *)
	    else new_term_subst_eq (FId (id1), t) [] (* new equation *)
	    
	  | (FZero, FZero) => ((FZero, FZero), [], []) (* no equation *)
	  
	  | (FFun (s, tl), FFun (s', tl')) =>
	    if (s <> s') orelse (List.length tl <> List.length tl') (* the arity must be the same (what about the function name ???) *)
	    then (* here, we need to unify, but only if no matching meta-variables are in subterms *)
			if List.all (fn id => not (Utils.listMember id xl)) (List.concat (List.map fv_term tl))
			then new_term_subst_eq (FFun (s, tl), FFun (s', tl')) [] (* ok => new equation *)
			else raise MatchUnifErrorTerms (FFun (s, tl), FFun (s', tl')) (* else => error *)
	    else (* first calling term_matchUnif recursively on subterms *)
			let val t_eql_substl_list = List.map (term_matchUnif xl) (ListPair.zip (tl, tl'))
		    	val tl1 = List.map (fn t => #1(#1(t))) t_eql_substl_list (* getting the new list tl[i][j] *)
		    	val tl2 = List.map (fn t => #2(#1(t))) t_eql_substl_list (* getting the new list tl'[u/i][j] *)
		    	val eql = List.concat (List.map #2 t_eql_substl_list) (* building the complete equation list *)
		    	val substl = (* consolidating the substitution list (to check that same variables are substituted by same terms *)
		    		List.foldl (fn (s1, s2) =>
		    						case ListAssoc.union s1 s2 of
		    						  NONE => raise MatchUnifErrorTerms (FFun (s, tl), FFun (s', tl'))
		    						| SOME s'' => s'')
		    					[] (List.map #3 t_eql_substl_list)
		in
		    ((FFun (s, tl1), FFun (s, tl2)), eql, substl)
		end
		
	  | (a, b) => raise MatchUnifErrorTerms (a, b) (* an error since matching was requiered... new_term_subst_eq (a, b) [] *)
    in
	if List.exists (fn x => Utils.listMember x (fv_term a)) xl (* checks whether a match is required *)
	then aux xl (a,b) (* if yes, then we proceed to it *)
	else
	    if a = b then ((a, b), [], []) (* else if a = b then nothing needs to be done *)
	    else new_term_subst_eq (a, b) [] (* else we simply create the equality *)
    end

(**
 * performs the pattern-matching and unification where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (p, f)
 * @param idl the identifier list of pattern meta-variables {i}.
 * @param p the formula pattern f[i][n/j].
 * @param f the formula to match f[u/i][m/j].
 *
 * @throws MatchUnifErrorTypes if the pattern cannot match the formula.
 *
 * @return the formulas with unification variables, the unification variables associated with equations
 * and the substitutions list that matches ((f[i][j], f[u/i][j]), (j,n,m), (i,u)).
 *)
fun matchUnif xl = fn
    (TNat (n), TNat (m)) => (* calls matchUnif on terms *)
    let val ((n', m'), eql, substl) = term_matchUnif xl (n, m)
    in
		((TNat (n'), TNat (m')), eql, substl)
    end
    
  | (TEqual (n1, m1), TEqual (n2, m2)) => (* calls matchUnif on subterms *)
    let val ((n1', n2'), eql, substl) = term_matchUnif xl (n1, n2)
		val ((m1', m2'), eql', substl') = term_matchUnif xl (m1, m2)
		val eql'' = eql @ eql'
		val substl'' = (* try union of substitutions *)
			(case ListAssoc.union substl substl' of
			    NONE => raise MatchUnifErrorTypes (TEqual (n1, m1), TEqual (n2, m2))
			  | SOME s => s)
    in
		((TEqual (n1', m1'), TEqual (n2', m2')), eql'', substl'')
    end
    
  | (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2)) =>
    if (List.length sigma1 <> List.length sigma2) orelse
       (List.length tau1 <> List.length tau2)
    then raise MatchUnifErrorTypes (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2))
    else
	let val ((i1', sigma1', j1', tau1'), (i2', sigma2', j2', tau2')) =
    		(* alpha-convert the two types to identify the bound variables *)
			alphaConvProc ((i1, sigma1, j1, tau1), (i2, sigma2, j2, tau2))
			(* recursive call upon every subtype *)
		val l1 = List.map (matchUnif xl) (ListPair.zip (sigma1', sigma2'))
	    val l2 = List.map (matchUnif xl) (ListPair.zip (tau1', tau2'))
	    val sigma1'' = List.map (fn t => #1(#1(t))) l1
	    val sigma2'' = List.map (fn t => #2(#1(t))) l1
	    val tau1'' = List.map (fn t => #1(#1(t))) l2
	    val tau2'' = List.map (fn t => #2(#1(t))) l2
	    val eql = (List.concat (List.map #2 l1)) @ (List.concat (List.map #2 l2))
	    val substl = (* try union of substitutions *)
	    	List.foldl
	    		(fn (s1, s2) =>
					case ListAssoc.union s1 s2 of
					    NONE =>
					    raise MatchErrorTypes (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2))
					  | SOME s'' => s'')
				[] ((List.map #3 l1) @ (List.map #3 l2))
	in
	    ((TProc (i1', sigma1'', j1', tau1''), TProc (i2', sigma2'', j2', tau2'')), eql, substl)
	end
	
  | (TVar (x), TVar (y)) => (* type variables must be identical *)
  if x = y then ((TVar (x), TVar (y)), [], []) (* no equation *)
  else raise MatchUnifErrorTypes (TVar (x), TVar (y)) (* error *)
  
  | (TBot, TBot) => ((TBot, TBot), [], []) (* no equation *)
  
  | (a,b) => raise MatchUnifErrorTypes (a,b) (* error *)

(**
 * performs the pattern-matching and unification where the pattern meta-variables
 * belongs to an identifiers list.
 *
 * @params idl (pl, fl)
 * @param idl the identifier list of pattern meta-variables {i}.
 * @param pl the (identifier, formula) list pattern (x:f)[i][n/j].
 * @param fl the (identifier, formula) list to match (x:f)[u/i][m/j].
 *
 * @throws MatchUnifErrorTypes if the pattern cannot match the formula.
 *
 * @return the formulas with unification variables, the unification variables associated with equations
 * and the substitutions list that matches (((x:f)[i][j], (x:f)[u/i][j]), (j,n,m), (i,u)).
 *)
fun matchUnifList xl (xl1, xl2) =
    if List.length xl1 <> List.length xl2 then raise LengthMismatch
    else
	let val rec aux = fn
		[] => ([], [], [])
	  | (x,t)::q =>
		(case (List.find (fn (y,t') => y = x) xl2, aux q) of
		     (SOME (y, t'), (xfl, l, s)) =>
		     let val ((f1, f2), eql, subst) = matchUnif xl (t, t') in
			 (case ListAssoc.union subst s of
			      NONE => raise MatchUnifErrorTypeList (List.map #2 xl1, List.map #2 xl2)
			    | SOME s' => ((x, f1, f2)::xfl, eql @ l, s'))
		     end
		   | _ => raise AbsentIdent (x, xl2))
	    val (phi_i, eql, subst) = aux xl1
	in
	    (* We must permute the equations in eql in order to have (in -> out) *)
	    (phi_i, (* List.map (fn (i,n,m) => (i,m,n)) *) eql, subst)
	end

(**
 * builds a new unification equation with a fresh identifier.
 *
 * @params (a, b) eql
 * @param a the first term of the equation.
 * @param b the second term of the equation.
 * @param eql the equation list to which add the new equation.
 *
 * @return the fresh identifier and the new equation list.
 *)
fun new_term_subst (a, b) eql = let val i = fresh() in (FId (i), (i, a, b)::eql) end

(**
 * performs the unification of two terms.
 *
 * @params (a, b)
 * @param a the first term to unify t[n/i].
 * @param b the second term to unify t[m/i].
 *
 * @return the term with meta-variables t[i] and the equations list (i, n = m).
 *)
fun term_unify (a, b) =
    case (a, b) of
		(FId (id1), FId (id2)) =>
		if id1 = id2 then (FId (id1), []) (* if identifiers are identical : no equation, *)
		else new_term_subst (a,b) [] (* new equation *)
		
      | (FZero, FZero) => (FZero, []) (* no equation *)
      
      | (FFun (s, tl), FFun (s', tl')) =>
      	if (s <> s') orelse (List.length tl <> List.length tl')
      	then new_term_subst (a,b) [] (* if function differs in name or arity : new equation *)
      	else (* recursivly builds equations *)
	    let val teql = List.map term_unify (ListPair.zip (tl, tl')) in
			(FFun (s, List.map #1 teql), List.concat (List.map #2 teql))
	    end
	    
      | (a, b) => new_term_subst (a,b) [] (* new equation *)

(**
 * performs the unification of two formulas on the term-level.
 *
 * @params (a, b)
 * @param a the first formula to unify f[n/i].
 * @param b the second formula to unify f[m/i].
 *
 * @throws NotUnifyableTypes when unification is impossible (formulas structure differs).
 *
 * @return the formula with meta-variables f[i] and the equations list (i, n = m).
 *)
fun unify (a,b) =
    case (a,b) of
		(TNat (n), TNat (m)) =>
		let val (n', eql) =
			if n = m then (n, []) (* no equation *)
			else new_term_subst (n, m) [] (* new equation *) (* Danger : don't perform term unification : term_unify (n, m)*)
		in (TNat (n'), eql) end
		
      | (TEqual (n1, m1), TEqual (n2, m2)) =>
      	(* unifying subterms *)
      	let val (n', eql) = if n1 = n2 then (n1, []) else new_term_subst (n1, n2) []
	    	val (m', eql) = if m1 = m2 then (m1, eql) else new_term_subst (m1, m2) eql
	    in (TEqual (n', m'), eql) end
	    (* Danger : don't perform term unification : (term_unify (n1, n2)) @ (term_unify (m1, m2))*)
	    
      | (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2)) =>
      	if (List.length sigma1 <> List.length sigma2) orelse
      		(List.length tau1 <> List.length tau2)
      	then raise NotUnifyableTypes (a,b) (* arity must be the same *)
      	else
      	let val ((i1', sigma1', j1', tau1'), (i2', sigma2', j2', tau2')) =
    		(* alpha-convert the two types to identify the bound variables *)
			alphaConvProc ((i1, sigma1, j1, tau1), (i2, sigma2, j2, tau2))
			(* unify sigmas *)
			val sigma_eql = List.map unify (ListPair.zip (sigma1', sigma2'))
			(* unify taus *)
			val tau_eql = List.map unify (ListPair.zip (tau1', tau2'))
	    in
			(TProc (i1', List.map #1 sigma_eql, j1', List.map #1 tau_eql),
			 (List.concat (List.map #2 sigma_eql)) @ (List.concat (List.map #2 sigma_eql)))
	    end
	    
      | (TVar (x), TVar (y)) =>
      if x = y then (TVar (x), []) (* type variables must be identical... ??? more unification ??? *)
      else raise NotUnifyableTypes (a,b)
      
      | (TBot, TBot) => (TBot, []) (* no equation *)

      | (a, b) => raise NotUnifyableTypes (a,b) (* can't unify structuraly different formulas *)

(**
 * performs the unification of two formulas list on the term-level. The unification is
 * performed all along the lists, using to the same order.
 *
 * @params (l1, l2)
 * @param l1 the first formula to unify l1[n/i].
 * @param l2 the second formula to unify l2[m/i].
 *
 * @throws LengthMismatch if lists size differs.
 *
 * @return the list of formula with meta-variables f[i] associated with the equations list (i, n = m).
 *)
fun unifyOrderedList (l1, l2) =
    if List.length l1 <> List.length l2 then raise LengthMismatch
    else List.map unify (ListPair.zip (l1, l2))

(**
 * performs the unification of two (identifiers, formulas) list on the term-level.
 * The unification is performed
 * according to the identifier associated with the formulas.
 *
 * @params (xl1, xl2)
 * @param xl1 the first formula to unify xl1[n/i].
 * @param xl2 the second formula to unify xl2[m/i].
 *
 * @throws LengthMismatch if lists size differs.
 * @throws AbsentIdent if lists differs for some identifier.
 *
 * @return the list of formula with meta-variables f[i] associated with the equations list (i, n = m).
 *)
fun unifyList (xl1, xl2) =
    if List.length xl1 <> List.length xl2 then raise LengthMismatch
    else
	let val rec aux = fn
		[] => []
	      | (x,t)::q =>
		(case List.find (fn (y,t') => y = x) xl2 of
		     SOME (y, t') => (x, unify (t, t')) :: (aux q)
		   | _ => raise AbsentIdent (x, xl2))
	    val l = List.map (fn (x, (t, l)) => ((x,t), l)) (aux xl1)
	in
	    (List.map #1 l, List.concat (List.map #2 l))
	end

end

