(**
 * The First-Order LoopW Type Inference formula/term substitution and alpha-equivalence.
 *
 * Provides functions in order to calculate free variables, perform term substitution
 * into formulas/terms, and identify formulas up to alpha-equivalence.
 *
 * @author Tristan Crolard <crolard@univ-paris12.fr>
 * @author Emmanuel Polonowski <polonowski@univ-paris12.fr>
 * @version 0.2.4
 *)
structure Terms :> TERMS = struct

open ProofAst

(**
 * gives a new fresh variable with the given identifier.
 *
 * @params i
 * @param i the identifier.
 *
 * @return a fresh identifier.
 *)
val alpha = let val n = ref 0 in fn (s : string, x : int) => (n := !n + 1; (s, !n)) end

(**
 * calculates the free variables of a term.
 *
 * @params t
 * @param t the term.
 *
 * @return the list of free identifiers.
 *)
fun fv_term (FId (id)) = [id]
  | fv_term (FZero) = []
  | fv_term (FFun (s, tl)) = List.concat (List.map fv_term tl)

(**
 * calculates the free variables of a formula.
 *
 * @params f
 * @param f the formula.
 *
 * @return the list of free identifiers.
 *)
fun fv_formula (TBot) = []
  | fv_formula (TNat (n)) = fv_term n
  | fv_formula (TEqual (n, m)) = Utils.listFusion (fv_term n) (fv_term m)
  | fv_formula (TVar (id)) = []
  | fv_formula (TProc (i, sigma, j, tau)) =
	let val fv_sigma = List.foldr Utils.listFusionPair [] (List.map fv_formula sigma)
		val fv_sigma' = List.filter (fn id => not (Utils.listMember id i)) fv_sigma
		val fv_tau = List.foldr Utils.listFusionPair [] (List.map fv_formula tau)
		val fv_tau' = List.filter (fn id => not (Utils.listMember id (i @ j))) fv_tau
	in Utils.listFusion fv_sigma' fv_tau'
	end

(**
 * calculates the free variables of a formula list.
 *
 * @params fl
 * @param fl the formula list.
 *
 * @return the list of free identifiers.
 *)
fun fv_formulas fl =
	List.foldr Utils.listFusionPair [] (List.map fv_formula fl)

(**
 * performs the substitution of a term into a term.
 *
 * @params x t tt
 * @param x the identifier to be substituted.
 * @param t the term to substitute to x.
 * @param tt the term in which to perform the substitution.
 *
 * @return the new formula tt[t/x].
 *)
fun term_substitute i t = fn
	FId (id) => if i = id then t else FId (id)
  | FZero => FZero
  | FFun (id, tl) =>
	FFun (case t of FId(nid) => if (i = id) then nid else id
				  | _ => id, List.map (term_substitute i t) tl)

(**
 * performs the capture-avoiding substitution of a term into a formula.
 *
 * @params x t f
 * @param x the identifier to be substituted.
 * @param t the term to substitute to x.
 * @param f the formula in which to perform the substitution.
 *
 * @return the new formula f[t/x].
 *)
fun substitute x t = fn
	TNat (n) => TNat (term_substitute x t n)
	
  | TEqual (n, m) => TEqual (term_substitute x t n, term_substitute x t m)
  
  | TProc (i, sigma, j, tau) =>
	 (* if x is in i, no substitution will occur in the formula *)
	if Utils.listMember x i then TProc (i, sigma, j, tau)
	else 
	let val fv_term_t = fv_term t (* we need the free variables of t in order to perform only needed alpha-conversion *)
		(* parsing the list of fv(t) to get those matching i, replacing them in the list by fresh variables *)
		val i_alpha = List.map (fn id => if Utils.listMember id fv_term_t then alpha id else id) i
		(* zip the old i and the new i_alpha, hence providing pairs either (x, x) of (x, fresh) *)
		val isubst = ListPair.zip (i, i_alpha)
		val sigma_alpha = (* performs alpha-conversion of sigma according to isubst *)
			List.foldl (fn ((id, id'), s) =>
				   if id = id' then s (* no alpha-conversion requiered *)
				   else List.map (substitute id (FId (id'))) s) sigma isubst (* alpha-conversion *)
		val (j_alpha, tau_alpha) = (* we compute here the alpha-conversion of tau *)
			if Utils.listMember x j then (j, tau) (* if x is in j, no substitution will occur hence no alhpa-conversion is needed *)
			else (* parsing the list of fv(t) to get those matching j, replacing them in the list by fresh variables *)
				let val j_alpha = List.map (fn id => if Utils.listMember id fv_term_t then alpha id else id) j
					(* zip the old j and the new j_alpha, hence providing pairs either (x, x) of (x, fresh) *)
					val jsubst = (ListPair.zip (j, j_alpha))
					val tau_alpha = (* performs alpha-conversion of tau according to jsubst *)
						List.foldl (fn ((id, id'), s) =>
							if id = id' then s (* no alpha-conversion requiered *)
							else List.map (substitute id (FId (id'))) s) tau (jsubst @ isubst) (* alpha-conversion *)
				in (j_alpha, tau_alpha) end
	in
		TProc (i_alpha, List.map (substitute x t) sigma_alpha,
				j_alpha, if Utils.listMember x j then tau_alpha
							else List.map (substitute x t) tau_alpha)
	end
	
  | TVar (id) => TVar (id)
  
  | TBot => TBot

(**
 * performs a list of substitutions into a formula.
 *
 * @params xtlist f
 * @param xtlist the list of substitutions to perform.
 * @param f the formula in which to perform the substitution.
 *
 * @return the new formula f[t1/x1, ..., tn/xn].
 *)
fun substituteList sl t =
	case sl of
		[] => t
	  | (x,u)::q => substitute x u (substituteList q t)

(**
 * performs a substitution into a named formula list.
 *
 * @params x t yflist
 * @param x the identifier to be substituted.
 * @param t the term to substitute to x.
 * @param yflist the named formula list in which to perform the substitution.
 *
 * @return the new formula [y,f][t/x].
 *)
fun substituteAll x t l = List.map (fn (id, f) => (id, substitute x t f)) l

(**
 * performs a list of substitutions into a named formula list.
 *
 * @params xtlist yflist
 * @param xtlist the list of substitutions to perform.
 * @param yflist the named formula list in which to perform the substitution.
 *
 * @return the new formula [y,f][t1/x1, ..., tn/xn].
 *)
fun substituteAllList sl l = List.map (fn (id, f) => (id, substituteList sl f)) l

(**
 * converts the proc types in order to have the same bound variables.
 *
 * @params (f1, f2)
 * @param f1 the first formula proc({i} sigma; out {j} tau).
 * @param f2 the second formula proc({i'} sigma'; out {j'} tau').
 *
 * @return (f1', f2') where f1 = proc({i''} sigma[i''/i]; out {j''} tau[j''/j]),
 * and f1 = proc({i''} sigma'[i''/i']; out {j''} tau'[j''/j']).
 *)
fun alphaConvProc ((i1, sigma1, j1, tau1), (i2, sigma2, j2, tau2)) =
	let val (j, tau1'', tau2'') = (* converts tau1 and tau2 using the same identifier list *)
		if j1 = j2 then (j1, tau1, tau2) (* No alpha-conversion required *)
		else
			let val j_alpha = List.map (fn id => alpha id) j1 (* Builds the list of fresh variables*)
				val j_subst1 = ListPair.zip (j1, j_alpha) (* Substitution for tau1 *)
				val j_subst2 = ListPair.zip (j2, j_alpha) (* Substitution for tau2 *)
				val tau1_subst = (* performs the substitution in tau1 *)
					List.foldl (fn ((id, id'), s) =>
						List.map (substitute id (FId (id'))) s) tau1 j_subst1
				val tau2_subst = (* performs the substitution in tau2 *)
					List.foldl (fn ((id, id'), s) =>
						List.map (substitute id (FId (id'))) s) tau2 j_subst2
			in (j_alpha, tau1_subst, tau2_subst) end
			
		val (i, sigma1', sigma2', tau1', tau2') = (* converts sigma1 and sigma2 (and tau1'& and tau2'') using the same identifier list *)
			if i1 = i2	then (i1, sigma1, sigma2, tau1'', tau2'') (* No alpha-conversion required *)
			else
				let val i_alpha = List.map (fn id => alpha id) i1 (* Builds the list of fresh variables*)
					val i_subst1 = ListPair.zip (i1, i_alpha) (* Substitution for sigma1 and tau1 *)
					val i_subst2 = ListPair.zip (i2, i_alpha) (* Substitution for sigma2 and tau2 *)
					val tau1_subst = (* performs the substitution in tau1'' *)
						List.foldl (fn ((id, id'), s) =>
							List.map (substitute id (FId (id'))) s) tau1'' i_subst1
					val tau2_subst = (* performs the substitution in tau2'' *)
						List.foldl (fn ((id, id'), s) =>
							List.map (substitute id (FId (id'))) s) tau2'' i_subst2
					val sigma1_subst = (* performs the substitution in sigma1 *)
						List.foldl (fn ((id, id'), s) =>
							List.map (substitute id (FId (id'))) s) sigma1 i_subst1
					val sigma2_subst = (* performs the substitution in sigma2 *)
						List.foldl (fn ((id, id'), s) =>
							List.map (substitute id (FId (id'))) s) sigma2 i_subst2
				in (i_alpha, sigma1_subst, sigma2_subst, tau1_subst, tau2_subst) end
		in
			((i, sigma1', j, tau1'), (i, sigma2', j, tau2'))
		end

(**
 * check the equality of a pair of formulas up to alpha-equivalence.
 *
 * @params (f1, f2)
 * @param f1 the first formula to compare.
 * @param f2 the second formula to compare.
 *
 * @return true if and only if f1 = f2 up to alpha-equivalence.
 *)
fun alphaEqual (t1, t2) =
	case (t1, t2) of
		(TNat (n), TNat (m)) => n = m
		
	  | (TEqual (n1, m1), TEqual (n2, m2)) => (n1 = n2) andalso (m1 = m2)
	  
	  | (TVar (x), TVar (y)) => x = y
	  
	  | (TBot, TBot) => true
	  
	  | (TProc (i1, sigma1, j1, tau1), TProc (i2, sigma2, j2, tau2)) =>
		if (List.length sigma1 <> List.length sigma2) orelse
			(List.length tau1 <> List.length tau2)
		then false
		else
			let val ((i1', sigma1', j1', tau1'), (i2', sigma2', j2', tau2')) =
				alphaConvProc ((i1, sigma1, j1, tau1), (i2, sigma2, j2, tau2))
			in (* recursive call of alpha-equivalence *)
				(List.all alphaEqual (ListPair.zip (tau1', tau2')))
				andalso (List.all alphaEqual (ListPair.zip (sigma1', sigma2')))
			end
			
	  | _ => false

end

